# BE CAREFUL: Transfering numbers in cython to C might result in
# problems, since they might be interpreted as ints instead of doubles
# (e.g. 1 -> 1.0)!!!
#
# Note that numpy advanaced matrix indexing does not work. Copy via np.asarray(), see generateHardDistribution().

from adaboost import AdaBoostClassifierMod
import supervised
import numpy as np
cimport numpy as np
import sklearn
from sklearn import metrics
import utils
import globs
import os
import cython
from sklearn.utils.validation import check_is_fitted
from libc.stdlib cimport malloc, free
from sklearn.datasets import fetch_mldata
from sklearn.datasets import load_svmlight_file
# from sklearn.datasets import dump_svmlight_file
from sklearn import preprocessing
from sklearn.ensemble import AdaBoostClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.tree import DecisionTreeClassifier
import os

# from numpy import empty
import random
import pickle
import gzip
import math
# from sklearn import tree  # for graphviz output

import matplotlib.pyplot as plt
from sklearn import datasets
from sklearn.utils import shuffle
from votingClass import VotingClassifier

from cpython cimport PyObject

#cdef extern from "numpy/arrayobject.h":
#    void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)


# Data structure to exchange instance data with c
#    dim      -- dimension (number of features)
#    npoints  -- number of points
#    features -- features as an (npoints * dim array in row order)
#    labels   -- labels
cdef public struct classificationdata:
    int dim
    int npoints
    double* features
    double* labels

ctypedef public classificationdata ClassificationData


# Data structure to return scores
#    trainingscore -- score on trainging set
#    testingscore  -- score on testing set
#    nnonz         -- number of nonzeros in solution (if applicable)
cdef public struct scoreresult:
    double trainingscore
    double testingscore
    int nnonz

ctypedef public scoreresult ScoreResult


# Data structure to exchange results of boosting algorithms with c
#    dim      -- dimension (number of features)
#    npoints  -- number of points
#    nlearner -- number of base learners
#    learners -- learners as an (nlearner * npoints array in row order)
#    margin   -- margin on training set
#
# The learners are given by their predictions (can be used to reconstruct)
cdef public struct boosterResult:
    int dim
    int npoints
    int nlearner
    double* learners
    double* predict
    double* weights
    double margin
    double trainingscore

ctypedef public boosterResult BoosterResult


# print str depending on debug flag
cdef printFunc(str):
    if globs.debugoutput == 1:
        print(str)

# generate hard instance according to "Random noise defeats convex boosting"
cdef generateHardDistribution(gamma, examples, noise, rstate):
    baseD = [[1.0, 0.0, 1.0], [gamma, -gamma, 1.0], [gamma, -gamma, 1.0], [gamma, 5.0*gamma, 1.0]]
    np.random.seed(rstate)
    index = np.random.choice(np.arange(0, len(baseD)), size=examples, replace=True)
    npBaseD = np.matrix(baseD)
    upsample = npBaseD[index]
    features = upsample[:,[0,1]]
    labels = upsample[:,2]
    labels = np.squeeze(np.asarray(labels))
    features = np.asarray(features)
    for i in range(labels.shape[0]):
        if np.random.random() < noise:
            labels[i] = (-1.0 * labels[i] + 1.0)/2.0
        else:
            labels[i] = (labels[i] + 1.0)/2.0
    return features, labels

# generate hard instance according to "Random noise defeats convex boosting"
cdef generateHardDistribution2(examples, noise, rstate):
    cdef dim = 21
    cdef nsize = int(examples/4)
    cdef dimtwo = int(dim/2)
    cdef dimfour = int(dim/4)
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([examples, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty([examples], np.double)

    np.random.seed(rstate)

    # determine labels
    for i in range(labels.shape[0]):
        if np.random.random() < 0.5:
            labels[i] = 1
        else:
            labels[i] = -1

    # treat large margin
    for i in range(nsize):
        for j in range(dim):
            features[i][j] = labels[i]

    # treat pullers
    for i in range(nsize):
        row = nsize + i
        for j in range(dim):
            if j < dimtwo + 1:
                features[row][j] = labels[row]
            else:
                features[row][j] = -labels[row]

    # treat penalizers
    for i in range(2*nsize):
        row = 2 * nsize + i

        # init with the negative labels
        for j in range(dim):
            features[row][j] = -labels[row]

        index = np.random.choice(dimtwo+1, size=dimfour)
        for j in index:
            features[row][j] = labels[row]

        index = np.random.choice(dimtwo, size=dimfour+1)
        for j in index:
            features[row][dimtwo + 1 + j] = labels[row]

    # randomly change labels
    for i in range(examples):
        if np.random.random() < noise:
            labels[i] = -labels[i]

    assert features.shape[0] == examples
    assert features.shape[1] == dim
    assert labels.shape[0] == examples

    return features, labels

# load data
cdef loadLearnerData(probname, rstate):
    if probname.startswith('hard_'):
        args = probname.split('_')
        if len(args) != 4:
            print("Deteced incomplete hard instance name. Usage: hard_[gamma]_[npoints]_[noise].\n")
            os._exit(0)

        gamma = float(args[1])
        npoints = int(args[2])
        noise = float(args[3])

        if npoints < 0:
            print("Negative number of points in hard instance.\n")
            os._exit(0)

        if gamma <= 0.0 or gamma > 1.0/6.0:
            print("Gamma for hard instance needs to be in (0,1/6).\n")
            os._exit(0)

        if noise <= 0.0 or noise > 1.0:
            print("Noise for hard instance should be in (0,1).\n")
            os._exit(0)

        print('Detected hard instance with gamma=%g, %d points, and noise=%g.\n' % (gamma,npoints,noise))

        features, labels = generateHardDistribution(gamma, npoints, noise, rstate)
        # dump_svmlight_file(features, labels, 'hard__%f_%d_%f' % (gamma,npoints,noise))

    elif probname.startswith('hard2_'):

        args = probname.split('_')
        if len(args) != 3:
            print("Deteced incomplete hard2 instance name. Usage: hard2_[npoints]_[gamma].\n")
            os._exit(0)

        npoints = int(args[1])
        gamma = float(args[2])

        if npoints < 0:
            print("Negative number of points in hard2 instance.\n")
            os._exit(0)

        if gamma <= 0.0 or gamma >= 1.0:
            print("Gamma in hard2 instance needs to be in (0,1).\n")
            os._exit(0)

        print('Detected hard2 instance with %d points and gamma=%g.\n' % (npoints,gamma))

        features, labels = generateHardDistribution2(npoints, gamma, rstate)
        # dump_svmlight_file(features, labels, 'hard2_%d_%f' % (npoints,gamma))

    elif probname == "digits":
        features, labels = datasets.load_digits(10, True)
    elif probname == "iris":
        features, labels = datasets.load_iris(True)
    elif probname == "breast_cancer":
        features, labels = datasets.load_breast_cancer(True)
    elif probname == "rand":
        features, labels = sklearn.datasets.make_classification(n_samples=20000, flip_y=0.2, n_redundant=25, n_features=50, random_state=rstate)
        # features, labels = sklearn.datasets.make_classification(n_samples=1000)
    elif probname == "rand2000":
        features, labels = sklearn.datasets.make_classification(n_samples=2000, flip_y=0.2, n_redundant=25, n_features=50, random_state=rstate)
        # features, labels = sklearn.datasets.make_classification(n_samples=1000)
    elif probname == "vote":
        mldata = fetch_mldata("datasets-UCI vote")
        # need some mapping from strings to numbers
        features, labels = mldata.data, mldata.target
    elif probname == "huheart":
        mldata = fetch_mldata("datasets-UCI heart-h")
        # need some mapping from strings to numbers
        features, labels = mldata.data, mldata.target
    elif probname == "clvheart":
        mldata = fetch_mldata("uci-20070111 heart-c")
        # need some mapping from strings to numbers
        features, labels = mldata.data, mldata.target
    elif probname == "sonar":
        mldata = fetch_mldata("datasets-UCI sonar")
        # need some mapping from strings to numbers
        features, labels = mldata.data, mldata.target
    elif probname == "mnist":
        mldata = fetch_mldata("MNIST original")
        features, labels = mldata.data / 255., mldata.target
        # create problem from digit 0 vs. the others
        for i in range(len(labels)):
            if labels[i] > 1:
                labels[i] = 1
    else:
        # allow to load svmlight/libsvm data sets, see
        # http://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/
        try:
            features, labels = load_svmlight_file(probname, dtype='double')
        except:
            print("File '" + probname + "' does not exist.")
            os._exit(0)
        # translate scipy matrix to numpy
        features = features.toarray()

# old code:
#        printFunc("--> Loading data from " + probname + ".dtz ...")
#        f = gzip.open(probname + '.dtz', 'rb')
#        features, labels = pickle.load(f)
#        f.close()
#        printFunc("Data points loaded: " + str(len(labels)))

    # adjust labels to be in {0,1} (for two classes)
    le = preprocessing.LabelEncoder()
    labels = le.fit_transform(labels)
    assert len(le.classes_) == 2

    return features, labels


# generate data from problem name
cdef public ClassificationData* generateLearnerData(const char* pname, int rstate):

    # convert c string to python
    cdef unicode probname = pname.decode('UTF-8')

    features, labels = loadLearnerData(probname, rstate)

    labels = labels.astype("double", order='C', casting='safe', subok=False, copy=True)

    # store data
    cdef ClassificationData* data = <ClassificationData*> malloc(sizeof(ClassificationData))
    if not data:
        raise MemoryError()

    data.npoints = features.shape[0]
    data.dim = features.shape[1]
    s = data.npoints * data.dim

    assert labels.shape[0] == data.npoints

    printFunc("Using problem:      " + '{:2s}'.format(probname))
    printFunc("Number of points:   " + '{:3d}'.format(data.npoints))
    printFunc("Number of features: " + '{:3d}'.format(data.dim))
    printFunc("")

    # allocate storage
    cdef double* featuresc = <double*> malloc(s * sizeof(double))
    if not featuresc:
        raise MemoryError()
    cdef double* labelsc = <double*> malloc(data.npoints * sizeof(double))
    if not labelsc:
        raise MemoryError()

    # copy data
    cdef k = 0
    for i in range(data.npoints):
        for j in range(data.dim):
            featuresc[k] = features[i][j]
            k = k + 1

    for i in range(data.npoints):
        labelsc[i] = labels[i]

    data.features = featuresc
    data.labels = labelsc

    return data


# generate testing data from file
cdef public ClassificationData* generateTestingData(const char* tname):

    # convert c string to python
    cdef unicode testname = tname.decode('UTF-8')

    features, labels = load_svmlight_file(testname, dtype='double')

    # translate scipy matrix to numpy
    features = features.toarray()

    # adjust labels to be in {0,1} (for two classes)
    le = preprocessing.LabelEncoder()
    labels = le.fit_transform(labels)
    assert len(le.classes_) <= 2

    labels = labels.astype("double", order='C', casting='safe', subok=False, copy=True)

    # store data
    cdef ClassificationData* testingdata = <ClassificationData*> malloc(sizeof(ClassificationData))
    if not testingdata:
        raise MemoryError()

    testingdata.npoints = features.shape[0]
    testingdata.dim = features.shape[1]
    s = testingdata.npoints * testingdata.dim

    assert labels.shape[0] == testingdata.npoints

    # allocate storage
    cdef double* featuresc = <double*> malloc(s * sizeof(double))
    if not featuresc:
        raise MemoryError()
    cdef double* labelsc = <double*> malloc(testingdata.npoints * sizeof(double))
    if not labelsc:
        raise MemoryError()

    # copy data
    cdef k = 0
    for i in range(testingdata.npoints):
        for j in range(testingdata.dim):
            featuresc[k] = features[i][j]
            k = k + 1

    for i in range(testingdata.npoints):
        labelsc[i] = labels[i]

    testingdata.features = featuresc
    testingdata.labels = labelsc

    return testingdata

# --------------------------------------------------------------------------------------------------------
# create predictions in {0,1}, [-1,1] or [-C, C], normalized by labels
cdef createPred(clf, features, labels, predtype):

    # hard code cutoff for argument of logarithm (if predtype 2)
    cdef double M = 1000.0

    if predtype == 0:
        # compute prediction in {0,1}
        pred = clf.predict(features)

        # normalize with labels, now {-1,1}
        pred = [ (2.0 * x - 1.0) * (2.0 * y - 1.0) for (x,y) in zip(pred,labels)]

    elif predtype == 1:
        # compute probabilities in [0,1]
        proba = clf.predict_proba(features)

        # normalize with labels, now [-1,1]
        pred = [ (2.0 * x - 1.0) * (2.0 * y - 1.0) for (x,y) in zip(proba[:,1],labels)]

    elif predtype == 2:
        # compute probabilities
        proba = clf.predict_proba(features)
        pred = np.empty([proba.shape[0]], np.double)
        for i in range (proba.shape[0]):
            p = proba[i][1]
            if p > M * (1.0 - p):
                val = M
            elif p < 1.0/M * (1.0 - p):
                val = 1.0/M
            else:
                val = p / (1.0 - p)
            assert val >= 1.0/M - np.finfo(proba.dtype).eps
            assert val <= M + np.finfo(proba.dtype).eps

            pred[i] = math.log(val)

        # normalize with labels: (0.5 * log(val)) * (2.0 * label[i] - 1.0)
        pred = [ 0.5 * x * (2.0 * y - 1.0) for (x,y) in zip(pred,labels)]

    return pred


# generate a column of the MIP
cdef public double* generateCol(int dim, int npoints, double* distc, double* featuresc, double* labelsc, int maxdepth, int rstate, int predtype):
    """Generate a column of the MIP.

    Keyword arguments:
    dim       -- dimension/number of features
    npoints   -- number of points
    distc     -- dual values of LP relaxation (weights for weak learner)
    featuresc -- given features (npoints * dim array in row order)
    labelsc   -- given labels (npoints)
    maxdepth  -- maximal depth of decision trees used for column generation
    rstate    -- random seed
    predtype  -- predtype (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)
    """

    # we have to translate c data into python data structures:
    # cdef np.ndarray[np.double_t, ndim=2, mode="c"] features = np.empty([npoints, dim], dtype=np.double)
    # cdef np.ndarray[np.double_t, ndim=1, mode="c"] labels = np.empty([npoints], dtype=np.double)
    # cdef np.ndarray[np.double_t, ndim=1, mode="c"] dist = np.empty([npoints], dtype=np.double)
    #
    # The conservatie solution would be:
    # copy in data
    #    cdef k = 0
    #    for i in range(npoints):
    #        dist[i] = distc[i]
    #        labels[i] = labelsc[i]
    #        for j in range(dim):
    #            features[i][j] = featuresc[k]
    #            k = k + 1

    # we view the data coming as a contiguous data as vectors and matrix
    cdef np.ndarray[np.double_t, ndim=1, mode="c"] labels = np.asarray(<np.double_t[:npoints]> labelsc)
    cdef np.ndarray[np.double_t, ndim=1, mode="c"] dist = np.asarray(<np.double_t[:npoints]> distc)
    cdef np.ndarray[np.double_t, ndim=2, mode="c"] features = np.asarray(<np.double_t[:npoints, :dim]> featuresc)

    # for debugging
    # cdef k = 0
    # for i in range(npoints):
    #    assert abs(dist[i] - distc[i]) <= 1e-6, "dist[%d] != distc[%d]" % (i,i)
    #    assert abs(labels[i] - labelsc[i]) <= 1e-6, "labels[%d] = %f != labelsc[%d] = %f" % (i,labels[i],i,labelsc[i])
    #    for j in range(dim):
    #        assert abs(features[i][j] - featuresc[k]) <= 1e-6, "features[%d][%d] != featuresc[%d]" % (i,j,k)
    #        k = k + 1

    # create copy since array might be changed in weakLearner
    cdef np.ndarray[np.double_t, ndim=1, mode="c"] d = np.copy(dist)

    # compute weak learner
    cRF = weakLearner(features, labels, d, globs.tLearn, maxdepth, rstate)

#    dotfile = open("dtree2.dot", 'w')
#    tree.export_graphviz(cRF, out_file = dotfile)
#    dotfile.close()

    # create predictions
    pred = createPred(cRF, features, labels, predtype)

    # set up return vector for c
    cdef double* predc = <double*> malloc(npoints * sizeof(double))
    if not predc:
        raise MemoryError()
    for i in range(npoints):
        predc[i] = pred[i]

    return predc


# compute weak learner
cdef weakLearner(features, labels, dist, cType, maxdepth, rstate):

    assert labels.shape[0] == features.shape[0]
    assert dist.shape[0] == features.shape[0]

    globs.compAuC = 0
    globs.plotROC = 0
    globs.genClassReport = 0

    cRF = supervised.trainClassifierWeighted(features, labels, dist, cType, maxdepth, rstate)

    return cRF

# --------------------------------------------------------------------------------------------------------
# simple boosting (for benchmarking)
cdef public ScoreResult SimpleBoost(int dim, int npoints, double* featuresc, double* labelsc, int T, int maxdepth, int rstate, int ntestpoints, double* testfeaturesc, double* testlabelsc, int predtype, int votertype):
    """Run and evaluate simple boosting.

    Keyword arguments:
    dim       -- dimension/number of features
    npoints   -- number of points
    featuresc -- training features (npoints * dim array in row order)
    labelsc   -- training labels (npoints)
    T         -- number of rounds
    maxdepth  -- maximal depth of decision trees
    ntestpoints   -- number of testing pointe
    testfeaturesc -- testing features (ntestpoints * dim array in row order)
    testlabelsc   -- testing labels (ntestpoints)
    predtype  -- predtype (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)
    votertype -- type of voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)
    """

    # we have to translate c data into python data structures:
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([npoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty(npoints, np.double)

    cdef np.ndarray[double, ndim=2, mode="c"] testfeatures = np.empty([ntestpoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] testlabels = np.empty([ntestpoints], np.double)

    # for running the algorithm
    cdef np.ndarray[double, ndim=1, mode="c"] dist = np.empty(npoints, np.double)

    # copy in data
    cdef k = 0
    for i in range(npoints):
        dist[i] = 1.0/npoints
        labels[i] = labelsc[i]
        for j in range(dim):
            features[i][j] = featuresc[k]
            k = k + 1

    # start with a uniform prior
    eps = 0.01

    # determine learning rate
    eta = math.sqrt(2.0/T * math.log(1.0/eps))
    estimators = []

    for t in range(T):
        printFunc("\n---> Boosting iteration: " + str(t))

        # find base learner
        cRF = weakLearner(features, labels, dist, globs.tLearn, maxdepth, rstate)

        # compute normalized predictions
        pred = createPred(cRF, features, labels, predtype)

        s = 0.0
        for i in range(npoints):
            # compute new values using the flipped predictions
            if predtype == 2:
                dist[i] = dist[i] * math.exp(-eta * pred[i])
            else:
                dist[i] = dist[i] * math.exp(eta * (1.0 - pred[i]/2.0))
            s = s + dist[i]
        for i in range(npoints):
            dist[i] = dist[i] / s
        estimators.append(["est" + str(t), cRF])

    # copy in testing data
    k = 0
    for i in range(ntestpoints):
        testlabels[i] = testlabelsc[i]
        for j in range(dim):
            testfeatures[i][j] = testfeaturesc[k]
            k = k + 1

    # evaluate score on testing data
    cdef ScoreResult result
    result.nnonz = len(estimators)
    if votertype == 0:
        clf = VotingClassifier(estimators=estimators, voting='hard')
        result.trainingscore = clf.score(features, labels)
        result.testingscore = clf.score(testfeatures, testlabels)
    elif votertype == 1:
        clf = VotingClassifier(estimators=estimators, voting='soft')
        result.trainingscore = clf.score(features, labels)
        result.testingscore = clf.score(testfeatures, testlabels)
    else:
        assert votertype == 2
        clf = VotingClassifier(estimators=estimators, voting='soft')
        result.trainingscore = metrics.accuracy_score(labels, clf.predictAlternative(features))
        result.testingscore = metrics.accuracy_score(testlabels, clf.predictAlternative(testfeatures))

    return result


# --------------------------------------------------------------------------------------------------------
# simple boosting version that returns the weights for reconstructing the solution
cdef public BoosterResult* SimpleBooster(int dim, int npoints, double* featuresc, double* labelsc, int T, int maxdepth, int rstate, int predtype, int votertype):
    """Retrun result of simple boosting.

    Keyword arguments:
    dim       -- dimension/number of features
    npoints   -- number of points
    featuresc -- training features (npoints * dim array in row order)
    labelsc   -- training labels (npoints)
    T         -- number of rounds
    maxdepth  -- maximal depth of decision trees
    predtype  -- predtype (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)
    votertype -- type of voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)
    """

    # we have to translate c data into python data structures:
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([npoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty(npoints, np.double)

    # for running the algorithm
    cdef np.ndarray[double, ndim=1, mode="c"] dist = np.empty(npoints, np.double)

    # copy in data
    cdef k = 0
    for i in range(npoints):
        dist[i] = 1.0/npoints
        labels[i] = labelsc[i]
        for j in range(dim):
            features[i][j] = featuresc[k]
            k = k + 1

    # prepare result
    cdef BoosterResult* result = <BoosterResult*> malloc(sizeof(BoosterResult))
    if not result:
        raise MemoryError()

    result.dim = dim
    result.npoints = npoints
    result.nlearner = T
    result.learners = <double*> malloc(sizeof(double) * T * npoints)
    result.predict = <double*> malloc(sizeof(double) * T * npoints)
    result.weights = <double*> malloc(sizeof(double) * T)

    # start with a uniform prior
    eps = 0.01

    # determine learning rate
    eta = math.sqrt(2.0/T * math.log(1.0/eps))
    estimators = []
    cdef cnt = 0
    for t in range(T):
        printFunc("\n---> Boosting iteration: " + str(t))

        # compute base learner
        cRF = weakLearner(features, labels, dist, globs.tLearn, maxdepth, rstate)

        # compute normalized predictions
        pred = createPred(cRF, features, labels, predtype)

        result.weights[t] = 1.0
        s = 0.0
        for i in range(npoints):
            # store weights and prediction (take old dist!)
            result.learners[cnt] = dist[i]

            # return predictions
            result.predict[cnt] = pred[i]
            cnt = cnt + 1

            # compute new values using the flipped predictions
            if predtype == 2:
                dist[i] = dist[i] * math.exp(-eta * pred[i])
            else:
                assert pred[i] == -1.0 or pred[i] == 1.0
                dist[i] = dist[i] * math.exp(eta * (1.0 - pred[i]/2.0))
            s = s + dist[i]
        for i in range(npoints):
            dist[i] = dist[i] / s
        estimators.append(["est" + str(t), cRF])

    assert cnt == T * npoints

    # evaluate score on testing data
    if votertype == 0:
        clf = VotingClassifier(estimators=estimators, voting='hard')
        result.trainingscore = clf.score(features, labels)
    elif votertype == 1:
        clf = VotingClassifier(estimators=estimators, voting='soft')
        result.trainingscore = clf.score(features, labels)
    else:
        assert votertype == 2
        clf = VotingClassifier(estimators=estimators, voting='soft')
        result.trainingscore = metrics.accuracy_score(labels, clf.predictAlternative(features))

    # get class probabilities and compute margin for training data
    result.margin = clf.margin(features, labels)

    return result


# run AdaBoost (for benchmarking)
cdef public ScoreResult AdaBoost(int dim, int npoints, double* featuresc, double* labelsc, int T, int maxdepth, int rstate, int type, int ntestpoints, double* testfeaturesc, double* testlabelsc, double* trainingmargin, double* testingmargin):
    """Run and evaluate AdaBoost with decision trees as base learners

    Keyword arguments:
    dim            -- dimension/number of features
    npoints        -- number of points
    featuresc      -- training features (npoints * dim array in row order)
    labelsc        -- training labels (npoints)
    T              -- number of rounds
    maxdepth       -- maximal depth of decision trees
    rstate         -- random seed
    type           -- type of predicition mapping (0: SAMME, 1: SAMME.R)
    ntestpoints    -- number of testing pointe
    testfeaturesc  -- testing features (ntestpoints * dim array in row order)
    testlabelsc    -- testing labels (ntestpoints)
    trainingmargin -- pointer to store minimal margin on training data (or NULL)
    testingmargin  -- pointer to store minimal margin on testing data (or NULL)
    """

    # we have to translate c data into python data structures:
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([npoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty(npoints, np.double)

    cdef np.ndarray[double, ndim=2, mode="c"] testfeatures = np.empty([ntestpoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] testlabels = np.empty([ntestpoints], np.double)

    # copy in data
    cdef k = 0
    for i in range(npoints):
        labels[i] = labelsc[i]
        for j in range(dim):
            features[i][j] = featuresc[k]
            k = k + 1

    # explicitly setup base classifier in order to control maximal depth
    base_est = DecisionTreeClassifier(max_depth=maxdepth, min_samples_split=globs.minSplit,
                                      criterion=globs.crit, max_features=globs.maxFeat, min_samples_leaf=globs.minSamp,
                                      random_state=rstate)

    if type == 0:
        clf = AdaBoostClassifierMod(base_estimator=base_est, n_estimators=T, algorithm='SAMME')
    else:
        clf = AdaBoostClassifierMod(base_estimator=base_est, n_estimators=T, algorithm='SAMME.R')
    clf.fit(features, labels)

    # output predictions of estimators
    # np.set_printoptions(threshold=np.nan)
    # for x in clf.estimators_:
    #     pred = x.predict(features)
    #     print(pred)

    cdef ScoreResult result

    # count number of weak learners without repetition: we use the basic signiture of the trees to compare
    result.nnonz = 0
    for e in range(clf.n_estimators):
        tree_e = clf.estimators_[e].tree_
        treesig_e = list(zip(tree_e.feature, tree_e.threshold, tree_e.children_left, tree_e.children_right))
        eq = 0
        for f in range(e-1):
            tree_f = clf.estimators_[f].tree_
            treesig_f = list(zip(tree_f.feature, tree_f.threshold, tree_f.children_left, tree_f.children_right))
            if treesig_e == treesig_f:
#                print("Equal trees: %d and %d" % (e, f))
#                print(treesig_e)
#                print(treesig_f)
                eq = 1
                break
        if eq == 0:
            result.nnonz += 1

    result.trainingscore = clf.score(features, labels)

    # get class probabilities and compute margin for training data
    m = 1.0
    p = clf.predict_proba(features)
    for x in np.nditer(p):
        a = abs(x - 0.5)
        if ( a < m ):
            m = a
    printFunc("AdaBoost training set margin: " + '{:f}'.format(m))
    if trainingmargin != NULL:
        trainingmargin[0] = m    # use ...[0] to assign to pointer

    # copy in testing data
    k = 0
    for i in range(ntestpoints):
        testlabels[i] = testlabelsc[i]
        for j in range(dim):
            testfeatures[i][j] = testfeaturesc[k]
            k = k + 1

    # get class probabilities and compute margin for testing data
    m = 1.0
    p = clf.predict_proba(testfeatures)
    for x in np.nditer(p):
        a = abs(x - 0.5)
        if ( a < m ):
            m = a
    printFunc("AdaBoost test set margin: " + '{:f}'.format(m))
    if testingmargin != NULL:
        testingmargin[0] = m

    result.testingscore = clf.score(testfeatures, testlabels)

    return result


# AdaBoost version that returns the weights for reconstructing the solution
cdef public BoosterResult* AdaBooster(int dim, int npoints, double* featuresc, double* labelsc, int T, int maxdepth, int rstate, int predtype):
    """Return the result of AdaBoost

    Keyword arguments:
    dim            -- dimension/number of features
    npoints        -- number of points
    featuresc      -- training features (npoints * dim array in row order)
    labelsc        -- training labels (npoints)
    T              -- number of rounds
    maxdepth       -- maximal depth of decision trees
    rstate         -- random seed
    predtype       -- predtype (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)
    """

    # we have to translate c data into python data structures:
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([npoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty(npoints, np.double)

    # copy in data
    cdef k = 0
    for i in range(npoints):
        labels[i] = labelsc[i]
        for j in range(dim):
            features[i][j] = featuresc[k]
            k = k + 1

    # prepare result
    cdef BoosterResult* result = <BoosterResult*> malloc(sizeof(BoosterResult))
    if not result:
        raise MemoryError()

    result.dim = dim
    result.npoints = npoints
    result.learners = <double*> malloc(sizeof(double) * T * npoints)
    result.predict = <double*> malloc(sizeof(double) * T * npoints)
    result.weights = <double*> malloc(sizeof(double) * T)

    # explicitly setup base classifier in order to control maximal depth
    base_est = DecisionTreeClassifier(max_depth=maxdepth, min_samples_split=globs.minSplit,
                                      criterion=globs.crit, max_features=globs.maxFeat, min_samples_leaf=globs.minSamp,
                                      random_state=rstate)

    clf = AdaBoostClassifierMod(base_estimator=base_est, n_estimators=T)
    clf.fit(features, labels)

    # ----------------------------------------------------------------------------
    # obtain base learners and store the corresponding weights and predictions
    assert clf.estimators_sample_weights_.shape[0] <= T
    assert clf.estimators_sample_weights_.shape[1] == npoints

    cdef int nestimators = clf.estimators_sample_weights_.shape[0]
    assert len(clf.estimators_) == nestimators
    assert clf.n_estimators == nestimators
    assert clf.estimators_sample_weights_.shape[1] == npoints

    result.nlearner = nestimators
    cdef int cnt = 0
    for e in range(nestimators):
        # store weight
        result.weights[e] = clf.estimator_weights_[e]

        # recover predictions
        assert clf.estimators_sample_weights_[e].shape[0] == npoints

        pred = createPred(clf.estimators_[e], features, labels, predtype)

        # store weights and prediction
        for i in range(npoints):
            result.learners[cnt] = clf.estimators_sample_weights_[e][i]
            result.predict[cnt] = pred[i]
            cnt = cnt + 1

    assert cnt == npoints * nestimators


    # ----------------------------------------------------------------------------------------------------------------------------
    # # build voting classifier out of AdaBoost result
    # estimators = []
    # for e in range(nestimators):
    #     estimators.append(["est" + str(e), clf.estimators_[e]])
    # voter = VotingClassifier(estimators=estimators, voting='soft', weights=clf.estimator_weights_)
    # print("score of AdaBoost with voting classifier 1: " + str(voter.score(features, labels)))
    # print("score of AdaBoost with voting classifier 2: " + str(metrics.accuracy_score(labels, voter.predictAlternative(features))))
    #
    # # for comparison try to reconstruct base learners
    # estimators = []
    # for e in range(nestimators):
    #     cRF = supervised.trainClassifierWeighted(features, labels, clf.estimators_sample_weights_[e], globs.tLearn, maxdepth)
    #     estimators.append(["est" + str(e) , cRF])
    # voter2 = VotingClassifier(estimators=estimators, voting='soft', weights=clf.estimator_weights_)
    # print("score of AdaBoost with voting classifier 3 " + str(voter2.score(features, labels)))
    # print("score of AdaBoost with voting classifier 4: " + str(metrics.accuracy_score(labels, voter2.predictAlternative(features))))
    # ----------------------------------------------------------------------------------------------------------------------------

    # store training score
    result.trainingscore = clf.score(features, labels)

    # get class probabilities and compute margin for training data
    m = 1.0
    p = clf.predict_proba(features)
    for x in np.nditer(p):
        a = abs(x - 0.5)
        if ( a < m ):
            m = a
    printFunc("AdaBoost training set margin: " + '{:f}'.format(m))
    result.margin = m

    return result

# run decision tree
cdef public ScoreResult GradientBoosting(int dim, int npoints, double* featuresc, double* labelsc, int maxdepth, int rstate, int ntestpoints, double* testfeaturesc, double* testlabelsc):
    """Run and evaluate gradient boosting

    Keyword arguments:
    dim       -- dimension/number of features
    npoints   -- number of points
    featuresc -- training features (npoints * dim array in row order)
    labelsc   -- training labels (npoints)
    maxdepth  -- maximal depth of decision trees
    rstate    -- random seed
    ntestpoints   -- number of testing pointe
    testfeaturesc -- testing features (ntestpoints * dim array in row order)
    testlabelsc   -- testing labels (ntestpoints)
    """

    # we have to translate c data into python data structures:
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([npoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty(npoints, np.double)

    cdef np.ndarray[double, ndim=2, mode="c"] testfeatures = np.empty([ntestpoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] testlabels = np.empty([ntestpoints], np.double)

    # copy in data
    cdef k = 0
    for i in range(npoints):
        labels[i] = labelsc[i]
        for j in range(dim):
            features[i][j] = featuresc[k]
            k = k + 1

    clf = GradientBoostingClassifier(loss='exponential', max_depth=maxdepth, random_state=rstate)
    clf.fit(features, labels)

    cdef ScoreResult result
    result.nnonz = clf.n_estimators_
    result.trainingscore = clf.score(features, labels)

    # copy in testing data
    k = 0
    for i in range(ntestpoints):
        testlabels[i] = testlabelsc[i]
        for j in range(dim):
            testfeatures[i][j] = testfeaturesc[k]
            k = k + 1

    result.testingscore = clf.score(testfeatures, testlabels)

    return result


# run random forest tree
cdef public ScoreResult RandomForest(int dim, int npoints, double* featuresc, double* labelsc, int maxdepth, int rstate, int ntestpoints, double* testfeaturesc, double* testlabelsc):
    """Run and evaluate random forests

    Keyword arguments:
    dim       -- dimension/number of features
    npoints   -- number of points
    featuresc -- training features (npoints * dim array in row order)
    labelsc   -- training labels (npoints)
    maxdepth  -- maximal depth of decision trees
    rstate    -- random seed
    ntestpoints   -- number of testing pointe
    testfeaturesc -- testing features (ntestpoints * dim array in row order)
    testlabelsc   -- testing labels (ntestpoints)
    """

    # we have to translate c data into python data structures:
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([npoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty(npoints, np.double)

    cdef np.ndarray[double, ndim=2, mode="c"] testfeatures = np.empty([ntestpoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] testlabels = np.empty([ntestpoints], np.double)

    # copy in data
    cdef k = 0
    for i in range(npoints):
        labels[i] = labelsc[i]
        for j in range(dim):
            features[i][j] = featuresc[k]
            k = k + 1

    if ( maxdepth > 0 ):
        clf = RandomForestClassifier(n_estimators=globs.numTrees, max_depth=maxdepth, min_samples_split=globs.minSplit,
                                     criterion=globs.crit, max_features=globs.maxFeat, min_samples_leaf=globs.minSamp,
                                     n_jobs=-1, random_state=rstate)
    else:
        clf = RandomForestClassifier(n_estimators=globs.numTrees, min_samples_split=globs.minSplit,
                                     criterion=globs.crit, max_features=globs.maxFeat, min_samples_leaf=globs.minSamp,
                                     n_jobs=-1, random_state=rstate)
    clf.fit(features, labels)

    cdef ScoreResult result
    result.trainingscore = clf.score(features, labels)

    # collect total number of nodes
    result.nnonz = 0
    for estimator in clf.estimators_:
        result.nnonz += estimator.tree_.node_count

    # copy in testing data
    k = 0
    for i in range(ntestpoints):
        testlabels[i] = testlabelsc[i]
        for j in range(dim):
            testfeatures[i][j] = testfeaturesc[k]
            k = k + 1

    result.testingscore = clf.score(testfeatures, testlabels)

    return result


# run decision tree
cdef public ScoreResult DecTree(int dim, int npoints, double* featuresc, double* labelsc, int maxdepth, int rstate, int ntestpoints, double* testfeaturesc, double* testlabelsc):
    """Run and evaluate Decision Trees

    Keyword arguments:
    dim       -- dimension/number of features
    npoints   -- number of points
    featuresc -- training features (npoints * dim array in row order)
    labelsc   -- training labels (npoints)
    rstate    -- random seed
    ntestpoints   -- number of testing pointe
    testfeaturesc -- testing features (ntestpoints * dim array in row order)
    testlabelsc   -- testing labels (ntestpoints)
    """

    # we have to translate c data into python data structures:
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([npoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty(npoints, np.double)

    cdef np.ndarray[double, ndim=2, mode="c"] testfeatures = np.empty([ntestpoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] testlabels = np.empty([ntestpoints], np.double)

    # copy in data
    cdef k = 0
    for i in range(npoints):
        labels[i] = labelsc[i]
        for j in range(dim):
            features[i][j] = featuresc[k]
            k = k + 1

    if ( maxdepth > 0 ):
        clf = DecisionTreeClassifier(random_state=rstate, max_depth=maxdepth)
    else:
        clf = DecisionTreeClassifier(random_state=rstate)
    clf.fit(features, labels)

    cdef ScoreResult result
    result.nnonz = clf.tree_.node_count
    result.trainingscore = clf.score(features, labels)

    # copy in testing data
    k = 0
    for i in range(ntestpoints):
        testlabels[i] = testlabelsc[i]
        for j in range(dim):
            testfeatures[i][j] = testfeaturesc[k]
            k = k + 1

    result.testingscore = clf.score(testfeatures, testlabels)

    return result


# --------------------------------------------------------------------------------------------------------
# Evaluate strong learner.
cdef public ScoreResult evalStrongLearner(int dim, int npoints, double* featuresc, double* labelsc, int nlearner, double* dualvalsc, double* weightsc, int ntestpoints, double* testfeaturesc, double* testlabelsc, int maxdepth, int rstate, int votertype):
    """Evaluate strong learner.

    Keyword arguments:
    dim       -- dimension/number of features
    npoints   -- number of points
    featuresc -- given features (npoints * dim array in row order)
    labelsc   -- given labels (npoints)
    nlearner  -- number of weak learners
    dualvalsc -- dual values used to generate weak learners (nlearner * npoints array in row order)
    weightsc  -- weights of MIP solution for each learner (nlearner)
    maxdepth  -- maximal depth of decision trees
    rstate    -- random seed
    votertype -- type of voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)
    """

    # we have to translate c data into python data structures:
    cdef np.ndarray[double, ndim=2, mode="c"] dualvals = np.empty([nlearner, npoints], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] weights = np.empty(nlearner, np.double)
    cdef np.ndarray[double, ndim=2, mode="c"] features = np.empty([npoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] labels = np.empty(npoints, np.double)

    cdef np.ndarray[double, ndim=2, mode="c"] testfeatures = np.empty([ntestpoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] testlabels = np.empty(ntestpoints, np.double)

    # copy in data
    cdef k = 0
    for l in range(nlearner):
        weights[l] = weightsc[l]
        for i in range(npoints):
            dualvals[l][i] = dualvalsc[k]
            k = k + 1

    cdef weightsum = 0.0
    for i in range(nlearner):
        weightsum = weightsum + weightsc[i]
    for i in range(nlearner):
        weights[i] = weightsc[i] / weightsum

    k = 0
    for i in range(npoints):
        labels[i] = labelsc[i]
        for j in range(dim):
            features[i][j] = featuresc[k]
            k = k + 1

    # collect base lerners
    estimators = []
    for i, dist in enumerate(dualvals):
        cRF = supervised.trainClassifierWeighted(features, labels, dist, globs.tLearn, maxdepth, rstate)
        estimators.append(["est" + str(i), cRF])

    # copy data
    k = 0
    for i in range(ntestpoints):
        testlabels[i] = testlabelsc[i]
        for j in range(dim):
            testfeatures[i][j] = testfeaturesc[k]
            k = k + 1

    # evaluate score on testing data
    cdef ScoreResult result
    result.nnonz = len(estimators)
    if votertype == 0:
        clf = VotingClassifier(estimators=estimators, voting='hard', weights=weights)
        result.trainingscore = clf.score(features, labels)
        result.testingscore = clf.score(testfeatures, testlabels)
    elif votertype == 1:
        clf = VotingClassifier(estimators=estimators, voting='soft', weights=weights)
        result.trainingscore = clf.score(features, labels)
        result.testingscore = clf.score(testfeatures, testlabels)
    else:
        assert votertype == 2
        clf = VotingClassifier(estimators=estimators, voting='soft', weights=weights)
        result.trainingscore = metrics.accuracy_score(labels, clf.predictAlternative(features))
        result.testingscore = metrics.accuracy_score(testlabels, clf.predictAlternative(testfeatures))

    return result


# Evaluate the sum of a bag of strong learners.
cdef public ScoreResult evalBagStrongLearners(int num, int dim, int* nbagpoints, int* nbaglearner, double* featuresc, double* labelsc, double* dualvalsc, double* weightsc, int ntestpoints, double* testfeaturesc, double* testlabelsc, int maxdepth, int rstate, int votertype):
    """Evaluate strong learner.

    Keyword arguments:
    num          -- total number of strong learners (bags)
    dim          -- dimension/number of features
    nbagpoints   -- number of points for each strong learner
    nbaglearner  -- number of weak learners in each strong learner
    featuresc    -- given training bag features (for each bag b: nbagpoints[b] * dim array in row order)
    labelsc      -- given training bag labels (for each bag b: nbagpoints[b])
    dualvalsc    -- dual values used to generate weak learners (nlearner * npoints array in row order)
    weightsc     -- weights of MIP solution for each learner (nlearner)
    ntestpoints  -- number of test points
    testfeaturesc-- test points in row format (ntestpoints * dim)
    testlabelsc  -- test labels (ntestpoints)
    maxdepth     -- maximal depth
    rstate       -- random seed
    votertype    -- type of voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)
    """

    cdef nfeatures = 0
    cdef nlabels = 0
    cdef ndualvals = 0
    cdef nweights = 0
    cdef ntrainpoints = 0    # total number of trainging points

    cdef np.ndarray[double, ndim=2, mode="c"] dualvals
    cdef np.ndarray[double, ndim=2, mode="c"] features
    cdef np.ndarray[double, ndim=1, mode="c"] labels

    estimators = []

    globs.compAuC = 0
    globs.plotROC = 0
    globs.genClassReport = 0

    # loop through bags and create weak learners
    for s in range(num):

        nlearner = nbaglearner[s]
        npoints = nbagpoints[s]

        nweights = nweights + nlearner
        ntrainpoints = ntrainpoints + npoints

        # we have to translate c data into python data structures:
        dualvals = np.empty([nlearner, npoints], np.double)
        features = np.empty([npoints, dim], np.double)
        labels = np.empty(npoints, np.double)

        # copy in data
        for l in range(nlearner):
            for i in range(npoints):
                dualvals[l][i] = dualvalsc[ndualvals]
                ndualvals = ndualvals + 1

        for i in range(npoints):
            labels[i] = labelsc[nlabels]
            nlabels = nlabels + 1
            for j in range(dim):
                features[i][j] = featuresc[nfeatures]
                nfeatures = nfeatures + 1

        for i, dist in enumerate(dualvals):
            cRF = supervised.trainClassifierWeighted(features, labels, dist, globs.tLearn, maxdepth, rstate)
            estimators.append(["est" + str(s) + ":" + str(i) , cRF])

    # prepare weights
    cdef np.ndarray[double, ndim=1, mode="c"] weights = np.empty(nweights, np.double)
    cdef weightsum = 0.0
    for i in range(nweights):
        weightsum = weightsum + weightsc[i]
    for i in range(nweights):
        weights[i] = weightsc[i] / weightsum

    cdef np.ndarray[double, ndim=2, mode="c"] trainfeatures = np.empty([ntrainpoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] trainlabels = np.empty(ntrainpoints, np.double)

    # copy training data
    k = 0
    for i in range(ntrainpoints):
        trainlabels[i] = labelsc[i]
        for j in range(dim):
            trainfeatures[i][j] = featuresc[k]
            k = k + 1

    cdef np.ndarray[double, ndim=2, mode="c"] testfeatures = np.empty([ntestpoints, dim], np.double)
    cdef np.ndarray[double, ndim=1, mode="c"] testlabels = np.empty(ntestpoints, np.double)

    # copy test data
    k = 0
    for i in range(ntestpoints):
        testlabels[i] = testlabelsc[i]
        for j in range(dim):
            testfeatures[i][j] = testfeaturesc[k]
            k = k + 1

    # evaluate score on testing data
    cdef ScoreResult result
    result.nnonz = len(estimators)
    if votertype == 0:
        clf = VotingClassifier(estimators=estimators, voting='hard', weights=weights)
        result.trainingscore = clf.score(features, labels)
        result.testingscore = clf.score(testfeatures, testlabels)
    elif votertype == 1:
        clf = VotingClassifier(estimators=estimators, voting='soft', weights=weights)
        result.trainingscore = clf.score(features, labels)
        result.testingscore = clf.score(testfeatures, testlabels)
    else:
        assert votertype == 2
        clf = VotingClassifier(estimators=estimators, voting='soft', weights=weights)
        result.trainingscore = metrics.accuracy_score(labels, clf.predictAlternative(features))
        result.testingscore = metrics.accuracy_score(testlabels, clf.predictAlternative(testfeatures))

    return result
