# -*- coding: utf-8 -*-

import pickle
import gzip

from sklearn.model_selection import cross_val_score
from sklearn import model_selection
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor, AdaBoostClassifier, GradientBoostingClassifier, ExtraTreesClassifier
from sklearn.multiclass import OneVsRestClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn import svm, metrics
import numpy as np
from scipy import interp
from sklearn.metrics import roc_curve, auc
from sklearn.metrics import classification_report
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB


# own imports
import globs
import utils


def printFunc(str):
    if globs.debugoutput == 1:
        print(str)

def validateSavedRandomForest(labels, features, factorName=''):
    '''
    very simple test of saved random forest classifier
    :param labels:
    :param features:
    :param factorName:
    :return:
    '''
    printFunc("")
    printFunc("Validation of saved random forest classifier.")

    clfRF = loadClassifier(factorName)

    printFunc("Cross validating and benchmarking classifier...")

    score = clfRF.score(features,labels)
    printFunc("Accuracy: %0.3f" % (score))

    if globs.genClassReport == 1:
        # generate additional classification report
        ypred = clfRF.predict(features)
        printFunc("Test statistics:")
        printFunc(classification_report(labels, ypred))
        utils.plotFeatureImportance(clfRF, globs.featLabel, factorName)


    if globs.genClassReport == 2:
        ypred = clfRF.predict(features)

        printFunc()
        printFunc("Combined train-test statistics:")
        printFunc()
        printFunc(classification_report(labels, ypred))

    printFunc("-> done.")

    return clfRF


# deterministic training using weights
def trainClassifierWeighted(features, labels, weights, cType, maxdepth, rstate):
    '''
    train random forest on labels and features
    :param labels: labels
    :param features: features
    :param weights: weights for each data point
    :param cType: type of classifier
    :param maxdepth: maximal depth of decision trees
    :param rstate: random seed
    :return:
    '''
    assert labels.shape[0] == features.shape[0]
    assert weights.shape[0] == features.shape[0]

    if cType == globs.tDecTree:
        clf = DecisionTreeClassifier(max_depth=maxdepth, min_samples_split=globs.minSplit,
                                     criterion=globs.crit, max_features=globs.maxFeat, min_samples_leaf=globs.minSamp,
                                     random_state=rstate)
        factorName = 'DecisionTree'

    elif cType == globs.tRF:
        clf = RandomForestClassifier(n_estimators=globs.numTrees, max_depth=maxdepth, min_samples_split=globs.minSplit,
                                     criterion=globs.crit, max_features=globs.maxFeat, min_samples_leaf=globs.minSamp,
                                     n_jobs=-1, random_state=rstate)
        factorName = 'RandomForest'

    elif cType == globs.tExtraTrees:
        clf = ExtraTreesClassifier(n_estimators=globs.numTrees, max_depth=maxdepth, min_samples_split=globs.minSplit,
                                   criterion=globs.crit, max_features=globs.maxFeat, min_samples_leaf=globs.minSamp,
                                   n_jobs=-1, random_state=rstate)
        factorName = 'ExtraTrees'

    elif cType == globs.tBoostTrees:
        clf = GradientBoostingClassifier(n_estimators=globs.numTrees, max_depth=maxdepth, min_samples_split=globs.minSplit,
                                         max_features=globs.maxFeat, min_samples_leaf=globs.minSamp,
                                         random_state=rstate)
        factorName = 'BoostedTrees'

    elif cType == globs.tSVM:
        clf = svm.SVC(gamma=0.001, probability=True)
        factorName = 'SVM'

    elif cType == globs.tRF1v:
        clf = OneVsRestClassifier(RandomForestClassifier(n_estimators=globs.numTrees, max_depth=None, min_samples_split=globs.minSplit,
                                 criterion=globs.crit, random_state=rstate))
        factorName = 'RandomForest1vsRest'

    elif cType == globs.tSVM1v:
        clf = OneVsRestClassifier(svm.SVC(gamma=0.001, probability=True))
        factorName = 'SVM1vsRest'

    elif cType == globs.tLogRegL2:
        clf = LogisticRegression(C=globs.C, penalty='l2')
        factorName = 'LogisticRegressionL2'

    elif cType == globs.tLogRegL1:
        clf = LogisticRegression(C=globs.C, penalty='l1')
        factorName = 'LogisticRegressionL1'

    elif cType == globs.tAdaBoost:
        clf = AdaBoostClassifier()
        factorName = 'AdaBoost'

    elif cType == globs.tGaussNB:
        clf = GaussianNB()
        factorName = 'GaussianNB'

    # normalize weights and make sure that they are not all 0
    s = weights.sum(dtype=np.double)
    if s == 0.0:
        weights[:] = 1.0 / weights.shape[0]
    else:
        weights = weights / s

    printFunc("Supervised training with weights: " + factorName + ".")

    clfRF = clf.fit(features, labels, sample_weight=weights)

    if globs.genClassReport == 2:
        ypred = clfRF.predict(features)

        printFunc()
        printFunc("Combined train-test statistics:")
        printFunc()
        printFunc(classification_report(labels, ypred))

    return clfRF


def trainRandomForestReg(labels, features):
    printFunc("")
    printFunc("Supervised training: random forest regressor.")

    # standard
    lnt = 100
    lmd = None
    lmss = 1

    # # optimized for mean-squared error
    # lnt = 10
    # lmd = 5
    # lmss = 10

    # # optimized for median absolute error
    # lnt = 10
    # lmd = 10
    # lmss = 10


    clf = RandomForestRegressor(n_estimators=lnt, max_depth=lmd, min_samples_split=lmss, random_state=globs.rstate)

    printFunc("Cross validating and benchmarking regressor...")

    if globs.cvExplicit == 1:
        Xtrain, Xtest, ytrain, ytest = model_selection.train_test_split(features, labels, test_size=0.5, random_state=globs.rstate)
        clfRF = clf.fit(Xtrain, ytrain)
        val = clfRF.score(Xtest, ytest)
        printFunc("Held out cross validation (CV) accuracy: %0.3f" % val)
    else:
        score = cross_val_score(clf, features, labels, cv=globs.numCvs)
        val = [score.mean(), score.std() * 2]
        printFunc("Accuracy: %0.3f (+/- %0.3f)" % (val[0], val[1]))

    Xtrain, Xtest, ytrain, ytest = model_selection.train_test_split(features, labels, test_size=0.5, random_state=globs.rstate)
    clfRF = clf.fit(Xtrain, ytrain)
    if globs.genClassReport == 1:
        utils.plotRelError(ytest, clfRF.predict(Xtest))

    clfRF = clf.fit(features, labels)
    printFunc("Saving classifier to file...")
    pickle.dump(clfRF, open(globs.fClassifier, "wb"))
    printFunc("-> done.")
    return clfRF


def computeAuCScore(clf, labels, features, pos_label=''):
    """
    this function computes area-under-curve (AuC) and accuracy for potential black box optimization
    :param clf: classifier to test
    :param labels: list of labels
    :param features: list of features
    :return:
    """
    aFeatures = np.asarray(features)
    aLabels= np.asarray(labels)
    # generate the folds for cross validation in the AuC test
    cv = model_selection.KFold(n_splits=globs.numCvs)
    # cv.get_n_splits(aFeatures)

    meanTpr = 0.0
    meanFpr = np.linspace(0, 1, 100)
    scores = []

    for i, (train, test) in enumerate(cv.split(aFeatures)):
        probas = clf.fit(aFeatures[train], aLabels[train]).predict_proba(aFeatures[test])
        if pos_label != '':
            fpr, tpr, thresholds = roc_curve(aLabels[test], probas[:, 1], pos_label)
        else:
            fpr, tpr, thresholds = roc_curve(aLabels[test], probas[:, 1])
        meanTpr += interp(meanFpr, fpr, tpr)
        meanTpr[0] = 0.0
        scores.append(clf.score(aFeatures[test], aLabels[test]))

    meanTpr /= globs.numCvs
    # len(cv)
    meanTpr[-1] = 1.0
    meanAuc = auc(meanFpr, meanTpr)
    meanScore = float(sum(scores)) / globs.numCvs

    return [meanAuc, meanScore]


def loadClassifier(factorName=''):
    if factorName=='':
        f = gzip.open(globs.fClassifier + '.clf', 'rb')
        clfRF = pickle.load(f)
        f.close()
    else:
        f = gzip.open(globs.fClassifier +'_' + factorName + '.clf', 'rb')
        clfRF = pickle.load(f)
        f.close()
    return(clfRF)


def saveClassifier(clfRF, factorName=''):
    if factorName=='':
        f = gzip.open(globs.fClassifier + '.clf', 'wb')
        pickle.dump(clfRF, f)
        f.close()
    else:
        f = gzip.open(globs.fClassifier +'_' + factorName + '.clf', 'wb')
        pickle.dump(clfRF, f)
        f.close()
    return(clfRF)
