# -*- coding: utf-8 -*-

import os
import logging

import numpy as np
from scipy import interp
import matplotlib.pyplot as plt
from sklearn import model_selection
from sklearn.metrics import roc_curve, auc
import sklearn

import pickle
import gzip

import globs

logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)

#########################
# VISUALIZATION
#########################

def plotROCCurve(clf, labels, features, factorName='',pos_label=''):
    # from scikit-learn documentation for quick comparison
    aFeatures = np.asarray(features)
    aLabels= np.asarray(labels)
    cv = model_selection.KFold(n_splits=globs.numCvs)
    mean_tpr = 0.0
    mean_fpr = np.linspace(0, 1, 100)

    for i, (train, test) in enumerate(cv.split(aFeatures)):
        probas = clf.fit(aFeatures[train], aLabels[train]).predict_proba(aFeatures[test])
        # Compute ROC curve and area the curve
        if pos_label != '':
            fpr, tpr, thresholds = roc_curve(aLabels[test], probas[:, 1],pos_label)
        else:
            fpr, tpr, thresholds = roc_curve(aLabels[test], probas[:, 1])
        mean_tpr += interp(mean_fpr, fpr, tpr)
        mean_tpr[0] = 0.0
        roc_auc = auc(fpr, tpr)
        plt.plot(fpr, tpr, lw=1, label='ROC fold %d (area = %0.2f)' % (i, roc_auc))

    plt.plot([0, 1], [0, 1], '--', color=(0.6, 0.6, 0.6), label='Random guessing')

    mean_tpr /= globs.numCvs
    # len(cv)
    mean_tpr[-1] = 1.0
    mean_auc = auc(mean_fpr, mean_tpr)
    plt.plot(mean_fpr, mean_tpr, 'k--',
             label='Mean ROC (area = %0.2f)' % mean_auc, lw=2)

    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver Operating Characteristic')
    plt.legend(loc="lower right")
    if factorName == '':
        plt.savefig(os.getcwd() + '/graphs/rfROC.pdf')
    else:
        plt.savefig(os.getcwd() + '/graphs/rfROC_' + factorName + '.pdf')
    plt.close()


def plotFeatureImportance(clf, featNames, factorName=''):
    # scikit-learn standard source
    numFeat = len(globs.featLabel)
    importances = clf.feature_importances_
    std = np.std([tree.feature_importances_ for tree in clf.estimators_],
                 axis=0)
    indices = np.argsort(importances)[::-1]
    # print(indices)

    # Print the feature ranking
    print("\n***\n")
    print("Feature ranking report:\n")

    for f in range(numFeat):
        print("%d. Feature: %s (%f)" % (f + 1, featNames[indices[f]], importances[indices[f]]))

    # Plot the feature importances of the forest
    plt.figure()
    plt.title("Feature importances")
    indices = indices[range(numFeat)]
    plt.bar(range(numFeat), importances[indices],
           color="yellow", yerr=std[indices], align="center")
    plt.xticks(range(numFeat), indices)
    plt.xlim([-1, numFeat])
    plt.ylim([0, 1])
    if factorName == '':
        plt.savefig(os.getcwd() + '/graphs/featureImportance.pdf')
    else:
        plt.savefig(os.getcwd() + '/graphs/featureImportance_' + factorName + '.pdf')
    plt.close()
    print("\n*** \n")


def plotRelError(ytrue, ypred):
    diff = [abs(a-b) / 10 for a,b in zip(ytrue,ypred)]
    importances = np.asarray(diff)
    indices = np.argsort(importances)[::1]
    plt.figure()
    plt.title("Relative score error in %")
    # indices =
    plt.plot(range(len(indices)), importances[indices],
           color="blue")
    # plt.plot(range(len(indices)), importances[indices],
    #        color="yellow", align="center")
    # plt.xticks(range(len(indices)), indices)
    plt.xlim([0, len(indices)])
    # plt.ylim([0, 1])
    plt.savefig(os.getcwd() + '/graphs/scoreRelError.pdf')


#########################
# END VISUALIZATION
#########################

def generateDataStore():
    # TODO: load actual data
    features, labels = sklearn.datasets.make_classification(n_samples=1000,flip_y=0,n_features=50,n_informative=5)
    f = gzip.open('test.dtz', 'wb')
    pickle.dump([features, labels], f)
    f.close()
