import numpy as np
from sklearn.utils.validation import check_is_fitted
from sklearn.metrics import accuracy_score
import math

class VotingClassifier(object):
    """Stripped-down version of VotingClassifier that uses prefit estimators"""

    def __init__(self, estimators, voting='hard', weights=None):
        self.estimators = [e[1] for e in estimators]
        self.named_estimators = dict(estimators)
        self.voting = voting
        self.weights = weights

    def fit(self, X, y, sample_weight=None):
        raise NotImplementedError

    def predict(self, X):
        """ Predict class labels for X.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
        Returns
        ----------
        maj : array-like, shape = [n_samples]
            Predicted class labels.
        """

        check_is_fitted(self, 'estimators')
        if self.voting == 'soft':
            maj = np.argmax(self.predict_proba(X), axis=1)

        else:  # 'hard' voting
            predictions = self._predict(X)
            maj = np.apply_along_axis(lambda x:
                                      np.argmax(np.bincount(x, weights=self.weights)),
                                      axis=1,
                                      arr=predictions.astype('int'))
            # equivalent code:
            # maj = np.empty(X.shape[0], np.int)
            # nlearner = predictions.shape[0]
            # margin = 1e10
            # for i in range(X.shape[0]):
            #    sumzeros = 0
            #    sumones = 0
            #    k = 0
            #    for r in predictions[i]:
            #        if self.weights is None:
            #            if ( r > 0.5 ):
            #                sumones += 1.0/nlearner
            #            else:
            #                sumzeros += 1.0/nlearner
            #        else:
            #            if ( r > 0.5 ):
            #                sumones += self.weights[k]
            #            else:
            #                sumzeros += self.weights[k]
            #        k += 1
            #    if ( sumones >= sumzeros ):
            #        maj[i] = 1
            #    else:
            #        maj[i] = 0
            #    d = abs(sumones - sumzeros);
            #    if ( d < margin ):
            #        margin = d
            # print("Minimal margin: %f\n" % margin)
        return maj

    def _collect_probas(self, X):
        """Collect results from clf.predict calls. """
        return np.asarray([clf.predict_proba(X) for clf in self.estimators])

    def _predict_proba(self, X):
        """Predict class probabilities for X in 'soft' voting """
        if self.voting == 'hard':
            raise AttributeError("predict_proba is not available when"
                                 " voting=%r" % self.voting)
        check_is_fitted(self, 'estimators')
        avg = np.average(self._collect_probas(X), axis=0, weights=self.weights)
        return avg

    def margin(self, X, y):
        """Compute margin of preditions for X.
       Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
        y : array-like, shape = (n_samples) or (n_samples, n_outputs)
            True labels for X.
        Returns
        ----------
        avg : array-like, shape = [n_samples, n_classes]
            Weighted average probability for each class per sample.
        """
        predictions = self._predict(X)
        predictions = predictions.astype('int')
        predictions = predictions.tolist()
        bc = [np.bincount(x, weights=self.weights).tolist() for x in predictions]
        bc = [x if len(x) == 2 else [x[0], 0] for x in bc]
        M = [abs(x[0] - x[1])/(2.0 * (x[0] + x[1])) for x in bc]
        return min(M)

    @property
    def predict_proba(self):
        """Compute probabilities of possible outcomes for samples in X.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
        Returns
        ----------
        avg : array-like, shape = [n_samples, n_classes]
            Weighted average probability for each class per sample.
        """
        return self._predict_proba

    def transform(self, X):
        """Return class labels or probabilities for X for each estimator.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.
        Returns
        -------
        If `voting='soft'`:
          array-like = [n_classifiers, n_samples, n_classes]
            Class probabilities calculated by each classifier.
        If `voting='hard'`:
          array-like = [n_samples, n_classifiers]
            Class labels predicted by each classifier.
        """
        check_is_fitted(self, 'estimators')
        if self.voting == 'soft':
            return self._collect_probas(X)
        else:
            return self._predict(X)

    def _predict(self, X):
        """Collect results from clf.predict calls. """
        return np.asarray([clf.predict(X) for clf in self.estimators]).T


    def score(self, X, y, sample_weight=None):
        """Returns the mean accuracy on the given test data and labels.

        In multi-label classification, this is the subset accuracy
        which is a harsh metric since you require for each sample that
        each label set be correctly predicted.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        y : array-like, shape = (n_samples) or (n_samples, n_outputs)
            True labels for X.

        sample_weight : array-like, shape = [n_samples], optional
            Sample weights.

        Returns
        -------
        score : float
            Mean accuracy of self.predict(X) wrt. y.

        """
        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)

    def _samme_proba(self, estimator, X):
        """Calculate geometric mean of probabilities
        """

        # compute class probabilities [n_samples, 2]
        proba = estimator.predict_proba(X)

        # Displace zero probabilities so the log is defined.
        # Also fix negative elements which may occur with
        # negative sample weights.
        proba[proba < np.finfo(proba.dtype).eps] = np.finfo(proba.dtype).eps
        log_proba = np.log(proba)

        return (log_proba - 0.5 * log_proba.sum(axis=1)[:, np.newaxis])

    def decision_function(self, X):
        """Compute the decision function of ``X``.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape = [n_samples, n_features]

        Returns
        -------
        score : array, shape = [n_samples]
            The decision function of the input samples. The order of
            outputs is the same of that of X.
        """
        check_is_fitted(self, "estimators")

        # apply SAMME.R
        if self.weights is None:
            pred = sum(self._samme_proba(estimator, X) for estimator in self.estimators)
            pred /= X.shape[0]
        else:
            pred = sum(self._samme_proba(estimator, X) * w for estimator, w in zip(self.estimators, self.weights))
            pred /= self.weights.sum()

        pred[:, 0] *= -1
        return pred.sum(axis=1)

    def predictAlternative(self, X):
        """Predict classes for X.

        The prediction of an input sample is computed as the weighted mean
        prediction of the classifiers in the ensemble.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape = [n_samples, n_features]

        Returns
        -------
        y : array of shape = [n_samples]
            The predicted classes.
        """

        check_is_fitted(self, 'estimators')
        pred = self.decision_function(X)

        y = [1 if x > 0 else 0 for x in pred]
        return y
