/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of the program ipboost                               */
/*                                                                           */
/*    an implementation of a boosting learning algorithm.                    */
/*                                                                           */
/*    Copyright (C) 2016-                                                    */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   handlepython.c
 * @brief  handle python interface
 */

#include "Python.h"

#include "handle_python.h"

#include "generateColumn.h"

#include <scip/pub_misc.h>

/** initialize python interface */
SCIP_RETCODE initPythonInterface(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   char cmd[SCIP_MAXSTRLEN];

   SCIPinfoMessage(scip, NULL, "initializing python interface ...\n");

   /* starting with Python 3, we have to use the inittab mechanism */
#if ( PY_MAJOR_VERSION >= 3 && CYTHON_PEP489_MULTI_PHASE_INIT > 0 )
   if ( PyImport_AppendInittab("generateColumn", PyInit_libgenerateColumn) == -1 )
   {
      SCIPerrorMessage("Cannot use inittab.\n");
      return SCIP_ERROR;
   }
   Py_Initialize();

   /* test whether module is there */
   if ( PyImport_ImportModule("generateColumn") == NULL )
   {
      SCIPerrorMessage("Cannot find python module.\n");
      Py_Finalize();
      return SCIP_ERROR;
   }
#else
   /* init python 2.x - has to always be done first */
   Py_Initialize();
   if ( PyErr_Occurred() )
   {
      PyErr_Print();
      return SCIP_ERROR;
   }
#endif

   if ( ! Py_IsInitialized() )
   {
      SCIPerrorMessage("Python interface not initialized.\n");
      return SCIP_ERROR;
   }

   SCIPinfoMessage(scip, NULL, "using python version: %s.\n", Py_GetVersion());

   SCIPdebugMessage("IPBOOST_LEARNER_DIR: %s\n", IPBOOST_LEARNER_DIR);
#ifdef IPBOOST_LEARNER_DIR
   /* modify path to find our own code (IPBOOST_LEARNER_DIR should be defined externally) */
   (void) SCIPsnprintf(cmd, SCIP_MAXSTRLEN, "import sys\nsys.path.append('%s')\n", IPBOOST_LEARNER_DIR);
   /* SCIPdebugMessage("Running python program: <%s>\n", cmd); */
   if ( PyRun_SimpleString(cmd) < 0 )
   {
      SCIPerrorMessage("Could not execute setting of import path.\n");
      return SCIP_ERROR;
   }
   if ( PyErr_Occurred() )
   {
      PyErr_Print();
      return SCIP_ERROR;
   }
   (void) SCIPsnprintf(cmd, SCIP_MAXSTRLEN, "print(sys.path)\n");
#ifdef SCIP_DEBUG
   printf("Python path:\n");
   PyRun_SimpleString(cmd);
#endif
#endif

   SCIPdebugMessage("using python prefix path: %ls\n", Py_GetPrefix());
   SCIPdebugMessage("using python exec path: %ls\n", Py_GetExecPrefix());
   SCIPdebugMessage("using python path: %ls\n", Py_GetPath());

   /* initialize own code */
#if PY_MAJOR_VERSION < 3
   (void) PyInit_generateColumn();
#endif
#if CYTHON_PEP489_MULTI_PHASE_INIT == 0
   (void) PyInit_libgenerateColumn();
#endif
   if ( PyErr_Occurred() )
   {
      PyErr_Print();
      return SCIP_ERROR;
   }
   SCIPdebugMessage("Initialized generate columns.\n");

   SCIPinfoMessage(scip, NULL, "Python interface initialized.\n\n");

   return SCIP_OKAY;
}

/** free python interface */
SCIP_RETCODE freePythonInterface(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   Py_Finalize();
   if ( PyErr_Occurred() )
   {
      PyErr_Print();
      return SCIP_ERROR;
   }

   SCIPdebugMessage("Freed python interface.\n");

   return SCIP_OKAY;
}
