/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of the program ipboost                               */
/*                                                                           */
/*    an implementation of a boosting learning algorithm.                    */
/*                                                                           */
/*    Copyright (C) 2016-                                                    */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   main.c
 * @brief  main file for ipboost
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <unistd.h>
#include <sys/times.h>

#include <scip/scip.h>
#include <scip/scipdefplugins.h>
#include <scip/scip_randnumgen.h>
#include <scip/pub_misc.h>

#include "readargs.h"
#include "probdata_ipboost.h"
#include "pricer_learning.h"
#include "Python.h"
#include "generateColumn.h"
#include "handle_python.h"

#define DEFAULT_MARGIN             0.01
#define DEFAULT_SEMILOWERBOUND     0.0
#define DEFAULT_L0WEIGHTFACTOR     0.001
#define DEFAULT_ADDSINGLECUTS      TRUE
#define DEFAULT_ADDPAIRCUTS        FALSE
#define DEFAULT_USEINDICATOR       TRUE
#define DEFAULT_USEINFERBRANCH     TRUE
#define DEFAULT_MAXBAGSIZE         30000
#define DEFAULT_BAGSAMPLE          TRUE
#define DEFAULT_TESTDATASHARE      0.2   /* 20 percent of whole data set */
#define DEFAULT_VALIDATION         FALSE
#define DEFAULT_THRESHOLD          0.0
#define DEFAULT_BOUNDLEARNERS      -1
#define DEFAULT_LPBOOST            FALSE
#define DEFAULT_ROBUSTMODEL        FALSE
#define DEFAULT_ROBUSTFRACTION     0.05
#define DEFAULT_MAXDEPTH           1
#define DEFAULT_OPTIMIZEMARGIN     FALSE
#define DEFAULT_OPTMARGINMIN       0.001
#define DEFAULT_MARGINFACTOR       0.9
#define DEFAULT_USESTALLING        TRUE
#define DEFAULT_STALLLIMIT         5000LL
#define DEFAULT_GUESSMARGIN        FALSE
#define DEFAULT_INITWITHBOOST      FALSE
#define DEFAULT_SPARSIFICATION     FALSE
#define DEFAULT_EVALSUBOPTIMAL     FALSE
#define DEFAULT_POSTPROCESS        TRUE
#define DEFAULT_PREDTYPE           1
#define DEFAULT_VOTERTYPE          1
#define DEFAULT_ADABOOSTTYPE       1
#define DEFAULT_ADABOOSTITERS      100

/* ------------- display column ---------------------- */

#define DISP_NAME_PRICEDVARS          "priced"
#define DISP_DESC_PRICEDVARS          "total number of priced variables applied to the LPs"
#define DISP_HEAD_PRICEDVARS          "price"
#define DISP_WIDT_PRICEDVARS          5
#define DISP_PRIO_PRICEDVARS          2100
#define DISP_POSI_PRICEDVARS          3500
#define DISP_STRI_PRICEDVARS          TRUE

/** output method of display column to output file stream 'file' for number of applied pricing vars */
static
SCIP_DECL_DISPOUTPUT(SCIPdispOutputNAppliedPricedvars)
{  /*lint --e{715}*/
   assert( disp != NULL );
   assert( strcmp(SCIPdispGetName(disp), DISP_NAME_PRICEDVARS) == 0 );
   assert( scip != NULL );

   SCIPdispInt(SCIPgetMessagehdlr(scip), file, SCIPgetNPricevarsApplied(scip), DISP_WIDT_PRICEDVARS);

   return SCIP_OKAY;
}


/* ------------- other functions ---------------------- */

/** disable several diving heuristics */
static
SCIP_RETCODE disableDivingHeuristics(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   int i;

#define NEXPENSIVEHEURFREQS 12
   static const char* const expensiveheurfreqs[NEXPENSIVEHEURFREQS] =
      {
         "heuristics/adaptivediving/freq",
         "heuristics/coefdiving/freq",
         "heuristics/distributiondiving/freq",
         "heuristics/fracdiving/freq",
         "heuristics/guideddiving/freq",
         "heuristics/linesearchdiving/freq",
         "heuristics/nlpdiving/freq",
         "heuristics/subnlp/freq",
         "heuristics/objpscostdiving/freq",
         "heuristics/pscostdiving/freq",
         "heuristics/rootsoldiving/freq",
         "heuristics/veclendiving/freq"
      };

   for (i = 0; i < NEXPENSIVEHEURFREQS; ++i)
   {
      SCIP_CALL( SCIPsetIntParam(scip, expensiveheurfreqs[i], -1) );
   }
   return SCIP_OKAY;
}


/** get data from python */
static
SCIP_RETCODE SCIPgetDataFromPython(
   SCIP*                 scip,               /**< SCIP data structure */
   const char*           probname,           /**< problem name */
   const char*           testname,           /**< testing file name (or NULL) */
   int                   randomseed,         /**< random seed */
   ClassificationData**  data,               /**< output: data from python */
   ClassificationData**  testingdata         /**< output: testing data from python or NULL */
   )
{
   int nclass1 = 0;
   int i;

   assert( data != NULL );
   assert( testingdata != NULL );

   *data = generateLearnerData(probname, randomseed);
   if ( PyErr_Occurred() )
   {
      PyErr_Print();
      return SCIP_ERROR;
   }
   assert( *data != NULL );

   /* determine sizes of classes and run check */
   for (i = 0; i < (*data)->npoints; ++i)
   {
      SCIP_Real val;

      val = (*data)->labels[i];
      if ( SCIPisEQ(scip, val, 1.0) )
         ++nclass1;
      else
      {
         if ( ! SCIPisEQ(scip, val, 0.0) )
         {
            SCIPerrorMessage("Labels have values different from {0, 1} (%f).\n", val);
            return SCIP_READERROR;
         }
      }
   }

   SCIPinfoMessage(scip, NULL, "Problem name:\t\t %s\n", probname);
   SCIPinfoMessage(scip, NULL, "Dimension:\t\t %d\n", (*data)->dim);
   SCIPinfoMessage(scip, NULL, "Number of points:\t %d\n", (*data)->npoints);
   SCIPinfoMessage(scip, NULL, "Class 1 size:\t\t %d\n", nclass1);
   SCIPinfoMessage(scip, NULL, "Class 0 size:\t\t %d\n", (*data)->npoints - nclass1);

   if ( nclass1 == (*data)->npoints )
   {
      SCIPerrorMessage("Only labels with 1 values present.\n");
      return SCIP_READERROR;
   }

   if ( nclass1 == 0 )
   {
      SCIPerrorMessage("Only labels with 0 values present.\n");
      return SCIP_READERROR;
   }

   /* possibly read testing data */
   *testingdata = NULL;
   if ( testname != NULL )
   {
      *testingdata = generateTestingData(testname);
      if ( PyErr_Occurred() )
      {
         PyErr_Print();
         return SCIP_ERROR;
      }
      assert( *testingdata != NULL );

      /* determine sizes of classes and run check */
      nclass1 = 0;
      for (i = 0; i < (*testingdata)->npoints; ++i)
      {
         SCIP_Real val;

         val = (*testingdata)->labels[i];
         if ( SCIPisEQ(scip, val, 1.0) )
            ++nclass1;
         else
         {
            if ( ! SCIPisEQ(scip, val, 0.0) )
            {
               SCIPerrorMessage("Labels have values different from {0, 1} (%f).\n", val);
               return SCIP_READERROR;
            }
         }
      }

      SCIPinfoMessage(scip, NULL, "\nTesting file name:\t %s\n", testname);
      SCIPinfoMessage(scip, NULL, "Dimension:\t\t %d\n", (*testingdata)->dim);
      SCIPinfoMessage(scip, NULL, "Number of points:\t %d\n", (*testingdata)->npoints);
      SCIPinfoMessage(scip, NULL, "Class 1 size:\t\t %d\n", nclass1);
      SCIPinfoMessage(scip, NULL, "Class 0 size:\t\t %d\n", (*testingdata)->npoints - nclass1);

      if ( nclass1 == (*testingdata)->npoints )
         SCIPinfoMessage(scip, NULL, "Warning: Only labels with 1 values present.\n");

      if ( nclass1 == 0 )
         SCIPinfoMessage(scip, NULL, "Warning: Only labels with 0 values present.\n");
   }

   return SCIP_OKAY;
}


/** split off training and testing data
 *
 *  Treat special case in which @p testdatashare is 1. In this case, training and testing data is the same.
 */
static
SCIP_RETCODE splitDataset(
   SCIP*                 scip,               /**< SCIP data structure */
   int                   randomseed,         /**< random seed */
   ClassificationData*   origdata,           /**< original data */
   ClassificationData**  trainingdata,       /**< output: training data (should not exist yet) */
   ClassificationData**  testingdata,        /**< output: testing data (might exist, not overwritten in this case) */
   ClassificationData**  validationdata,     /**< output: validation data (should not exist yet) */
   SCIP_Real             testdatashare,      /**< share (in [0,1]) of data for testing */
   SCIP_Real             validationshare     /**< share (in [0,1]) of data for validation */
   )
{
#if ( SCIP_VERSION >= 400 )
   SCIP_RANDNUMGEN* randnumgen;
#endif
   int* randperm;
   int npoints;
   int trainingsize;
   int testingsize;
   int validationsize = 0;
   int dim;
   int i;
   int j;
   int k = 0;

   assert( origdata != NULL );
   assert( trainingdata != NULL );
   assert( testingdata != NULL );
   assert( validationdata != NULL );

   npoints = origdata->npoints;
   dim = origdata->dim;

   /* prepare permutation */
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &randperm, npoints) );

   /* create random permutation to sample */
   for (i = 0; i < npoints; ++i)
      randperm[i] = i;

   /* create random number generator */
#if ( SCIP_VERSION < 400 )
   SCIPpermuteIntArray(randperm, 0, npoints, (unsigned int) &randomseed);
#elif ( SCIP_VERSION >= 600 )
   SCIP_CALL( SCIPcreateRandom(scip, &randnumgen, (unsigned int) randomseed, FALSE) );
   SCIPrandomPermuteIntArray(randnumgen, randperm, 0, npoints);
   SCIPfreeRandom(scip, &randnumgen);
#else
   SCIP_CALL( SCIPrandomCreate(&randnumgen, SCIPblkmem(scip), (unsigned int) randomseed) );
   SCIPrandomPermuteIntArray(randnumgen, randperm, 0, npoints);
   SCIPrandomFree(&randnumgen);
#endif

   assert( testdatashare + validationshare <= 1.0 );
   assert( validationshare < 1.0 );

   if ( *testingdata != NULL )
   {
      testingsize = 0;
      validationsize = (int) (npoints * validationshare);
      trainingsize = npoints - validationsize;
      assert( trainingsize > 0 );

      SCIPinfoMessage(scip, NULL, "\n");
      SCIPinfoMessage(scip, NULL, "Size of training data:\t\t %d\n", trainingsize);
      SCIPinfoMessage(scip, NULL, "Size of validation data:\t %d\n", validationsize);
   }
   else if ( SCIPisEQ(scip, testdatashare, 1.0) )
   {
      assert( validationshare == 0.0 );

      testingsize = npoints;
      trainingsize = npoints;

      SCIPinfoMessage(scip, NULL, "\n");
      SCIPinfoMessage(scip, NULL, "Training and testing data are the same (size: %d).\n", npoints);
   }
   else
   {
      testingsize = (int) (npoints * testdatashare);
      validationsize = (int) (npoints * validationshare);
      trainingsize = npoints - testingsize - validationsize;
      assert( trainingsize > 0 );

      SCIPinfoMessage(scip, NULL, "\n");
      SCIPinfoMessage(scip, NULL, "Size of training data:\t\t %d\n", trainingsize);
      SCIPinfoMessage(scip, NULL, "Size of testing data:\t\t %d\n", testingsize);
      SCIPinfoMessage(scip, NULL, "Size of validation data:\t %d\n", validationsize);
   }

   /* allocate space for test data */
   if ( *testingdata == NULL )
   {
      SCIP_CALL( SCIPallocBlockMemory(scip, testingdata) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*testingdata)->features, testingsize * dim) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*testingdata)->labels, testingsize) );
      (*testingdata)->dim = origdata->dim;
      (*testingdata)->npoints = testingsize;

      /* first copy testing data */
      for (i = 0; i < testingsize; ++i)
      {
         assert( 0 <= randperm[i] && randperm[i] < npoints );
         for (j = 0; j < dim; ++j)
            (*testingdata)->features[k++] = origdata->features[randperm[i] * dim + j];
         (*testingdata)->labels[i] = origdata->labels[randperm[i]];
      }
      assert( k == testingsize * dim );
   }

   /* allocate space for validation data */
   if ( validationsize > 0 )
   {
      SCIP_CALL( SCIPallocBlockMemory(scip, validationdata) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*validationdata)->features, validationsize * dim) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*validationdata)->labels, validationsize) );
      (*validationdata)->dim = origdata->dim;
      (*validationdata)->npoints = validationsize;

      /* copy validation data */
      k = 0;
      for (i = testingsize; i < testingsize + validationsize; ++i)
      {
         assert( 0 <= randperm[i] && randperm[i] < npoints );
         for (j = 0; j < dim; ++j)
            (*validationdata)->features[k++] = origdata->features[randperm[i] * dim + j];
         (*validationdata)->labels[i - testingsize] = origdata->labels[randperm[i]];
      }
      assert( k == validationsize * dim );
   }
   else
   {
      assert( validationsize == 0 );
      *validationdata = NULL;
   }

   /* allocate space for training data */
   SCIP_CALL( SCIPallocBlockMemory(scip, trainingdata) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*trainingdata)->features, trainingsize * dim) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*trainingdata)->labels, trainingsize) );
   (*trainingdata)->dim = origdata->dim;
   (*trainingdata)->npoints = trainingsize;

   /* if training and testing data should be the same */
   if ( SCIPisEQ(scip, testdatashare, 1.0) && (*testingdata != NULL) )
   {
      testingsize = 0;
      validationsize = 0;
   }

   /* then copy training data */
   k = 0;
   for (i = testingsize + validationsize; i < npoints; ++i)
   {
      assert( 0 <= randperm[i] && randperm[i] < npoints );
      for (j = 0; j < dim; ++j)
         (*trainingdata)->features[k++] = origdata->features[randperm[i] * dim + j];
      (*trainingdata)->labels[i - testingsize - validationsize] = origdata->labels[randperm[i]];
   }
   assert( k == trainingsize * dim );

   SCIPfreeBlockMemoryArray(scip, &randperm, npoints);

   return SCIP_OKAY;
}


/** initialize problem with variables and solution from boosting */
static
SCIP_RETCODE initFromBoosting(
   SCIP*                 scip,               /**< SCIP data structure */
   int                   dim,                /**< dimension */
   int                   npoints,            /**< number of points */
   SCIP_Real             margin,             /**< margin */
   int                   predtype,           /**< type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   SCIP_Bool             lpboost,            /**< if we are running LP-boost */
   BoosterResult*        boosterresult       /**< data returned from running boosting */
   )
{
   SCIP_PROBDATA* probdata;
   SCIP_SOL* sol;
   SCIP_VAR** origvars;
   SCIP_Bool stored;
   SCIP_Real val;
   SCIP_Real sum = 0.0;
   SCIP_VAR** binvars;
   int norigbinvars;
   int norigcontvars;
   int nmisclassified = 0;
   int cnt = 0;
   int l;
   int j;
   int i;

   assert( boosterresult != NULL );
   assert( boosterresult->dim == dim );
   assert( boosterresult->npoints = npoints );

   SCIPinfoMessage(scip, NULL, "Adding %d variables from boosting.\n", boosterresult->nlearner);
   for (l = 0; l < boosterresult->nlearner; ++l)
   {
      SCIP_CALL( SCIPaddLearnerVar(scip, dim, npoints, &boosterresult->learners[l * npoints], &boosterresult->predict[l * npoints], TRUE) );
      sum += boosterresult->weights[l];
   }

   /* creating solution for LP-boost not yet implemented */
   if ( lpboost )
      return SCIP_OKAY;

   /* normalize weights */
   for (l = 0; l < boosterresult->nlearner; ++l)
   {
      boosterresult->weights[l] /= sum;
   }

#if 0
   SCIP_CALL( SCIPprintOrigProblem(scip, NULL, NULL, FALSE) );
#endif

   /* add solution from boosing */
   SCIP_CALL( SCIPcreateOrigSol(scip, &sol, NULL) );

   /* get original variables */
   SCIP_CALL( SCIPgetOrigVarsData(scip, &origvars, NULL, &norigbinvars, NULL, NULL, &norigcontvars) );
   for (j = norigbinvars; j < norigbinvars + norigcontvars; ++j)
   {
      SCIP_VAR* var;
      var = origvars[j];
      if ( strcmp(SCIPvarGetName(var), "margin") == 0 )
      {
         SCIP_CALL( SCIPsetSolVal(scip, sol, var, margin) );
      }
      else
      {
         val = boosterresult->weights[cnt++];
         SCIP_CALL( SCIPsetSolVal(scip, sol, var, val) );
      }
   }

   /* compute violations */
   probdata = SCIPgetProbData(scip);
   assert( probdata != NULL );
   binvars = IpboostProbGetBinaryvars(probdata);
   assert( binvars != NULL );
   for (i = 0; i < npoints; ++i)
   {
      sum = 0.0;
      for (l = 0; l < boosterresult->nlearner; ++l)
      {
         assert( SCIPisFeasGE(scip, boosterresult->predict[l * npoints + i], -1) || SCIPisFeasLT(scip, boosterresult->predict[l * npoints + i], 1.0) );
         assert( predtype != 0 || (SCIPisFeasEQ(scip, boosterresult->predict[l * npoints + i], -1) || SCIPisFeasEQ(scip, boosterresult->predict[l * npoints + i], 1.0)) );
         val = boosterresult->weights[l];
         sum += boosterresult->predict[l * npoints + i] * val;
      }

      if ( SCIPisLT(scip, sum, margin) )
      {
         SCIP_CALL( SCIPsetSolVal(scip, sol, binvars[i], 1.0) );
         ++nmisclassified;
      }
   }
   SCIPinfoMessage(scip, NULL, "Numer of misclassified points in solution: %d.\n", nmisclassified);

   SCIP_CALL( SCIPcheckSolOrig(scip, sol, &stored, TRUE, TRUE) );
   if ( stored )
   {
      SCIP_CALL( SCIPaddSolFree(scip, &sol, &stored) );
      assert( stored );
   }
   else
   {
      SCIP_CALL( SCIPfreeSol(scip, &sol) );
   }

   return SCIP_OKAY;
}


/** run bagging approach */
static
SCIP_RETCODE runBagging(
   SCIP*                 scip,               /**< SCIP data structure */
   const char*           probname,           /**< problem name */
   SCIP_Real             timelimit,          /**< total time limit for computations */
   SCIP_Real             memlimit,           /**< memory limit */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   int                   maxdepth,           /**< maximal depth of decision trees used for column generation */
   int                   randomseed,         /**< random seed */
   ClassificationData*   trainingdata,       /**< training data */
   ClassificationData*   testingdata,        /**< testing data */
   ClassificationData*   validationdata,     /**< validation data */
   int                   maxbagsize,         /**< maximal size of bags */
   SCIP_Real             marginval,          /**< margin that should be used (unless guessmargin or optimizemargin is true) */
   SCIP_Bool             lpboost,            /**< whether to use LP boosting */
   SCIP_Bool             robustmodel,        /**< whether we should use the robust model */
   SCIP_Real             robustfraction,     /**< fraction of points that are allowed to be ignored in robust model */
   SCIP_Bool             optimizemargin,     /**< wether we want to optimize the margin */
   SCIP_Real             optmarginmin,       /**< minimal margin when optimizemargin is true */
   SCIP_Real             marginfactor,       /**< factor to compute margin from estimate */
   SCIP_Real             semilowerbound,     /**< Lower bound for sparsification problem */
   SCIP_Bool             usestalling,        /**< Use stall limit in column generation? Turns approach into a heuristic. */
   SCIP_Longint          stalllimit,         /**< stall limit for column generation */
   SCIP_Bool             guessmargin,        /**< Guess margin based on other boosting algorithm? */
   SCIP_Bool             initwithboost,      /**< Run boosting algorithm to initialize problem */
   int                   boundlearners,      /**< maximal number of learners for a solution to be accepted if validation is true */
   SCIP_Bool             sparsification,     /**< wether we want to run a sparsification routine after IPBoost */
   SCIP_Bool             evalsuboptimal,     /**< whether suboptimal solutions should be evaluated after IPBoost */
   SCIP_Bool             postprocess,        /**< whether postprocessing of the solution should be performed */
   SCIP_Bool             bagsample,          /**< whether to use only one bag, i.e., sample maxbagsize points and labels to form the training set */
   SCIP_Bool             useinferbranch,     /**< */
   int                   predtype,           /**< type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   int                   votertype           /**< type of prediction used for voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   )
{
#if ( SCIP_VERSION >= 400 )
   SCIP_RANDNUMGEN* randnumgen;
#endif
   SCIP* subscip = NULL;
   SCIP_PROBDATA* probdata;
   BoosterResult* boosterresult = NULL;
   ScoreResult score;
   SCIP_Real* features;
   SCIP_Real* labels;
   SCIP_Real* bagfeatures = NULL;
   SCIP_Real* baglabels = NULL;
   SCIP_Real* bagdualvals = NULL;
   SCIP_Real* bagweights = NULL;
   SCIP_Real tl;
   int* nbaglearners = NULL;
   int* nbagpoints = NULL;
   int* randperm;
   int nmaxbagdualvals = 0;
   int nbagdualvals = 0;
   int nmaxbagweights = 0;
   int nbagweights = 0;
   int nmaxbagfeatures = 0;
   int nmaxbaglabels = 0;
   int nbagfeatures = 0;
   int nbaglabels = 0;
   int nbags;
   int dim;
   int i;
   int j;
   int b;

   assert( trainingdata != NULL );
   assert( testingdata != NULL );
   /* note that validationdata may be NULL */

   /* adjust maxbagsize for small instances */
   if ( trainingdata->npoints < maxbagsize )
      maxbagsize = trainingdata->npoints;

   /* determine number of bags (runs) */
   nbags = (int) ceil((SCIP_Real) trainingdata->npoints / (SCIP_Real) maxbagsize);
   dim = trainingdata->dim;

   if ( nbags > 1 )
   {
      if ( bagsample )
      {
         nbags = 1;
         SCIPinfoMessage(scip, NULL, "\nUse one bag, i.e., sample %d points to form the training set.\n", maxbagsize);
      }
      else
         SCIPinfoMessage(scip, NULL, "\nUsing bagging approach with %d bags (maxbagsize: %d).\n", nbags, maxbagsize);
   }

   /* alloc space */
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &randperm, trainingdata->npoints) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &features, maxbagsize * dim) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &labels, maxbagsize) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &nbaglearners, nbags) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &nbagpoints, nbags) );

   /* compute individual time limits */
   tl = timelimit;
   if ( ! SCIPisInfinity(scip, timelimit) )
      tl = timelimit / (SCIP_Real) nbags;

   /* possibly set stall limit */
   if ( usestalling )
   {
      SCIPinfoMessage(scip, NULL, "\nUsing stall limit %" SCIP_LONGINT_FORMAT ".\n\n", stalllimit);
      SCIP_CALL( SCIPsetLongintParam(scip, "limits/stallnodes", stalllimit) );
   }

   /* create random number generator */
#if ( SCIP_VERSION >= 600 )
   SCIP_CALL( SCIPcreateRandom(scip, &randnumgen, (unsigned int) randomseed, FALSE) );
#elif ( SCIP_VERSION >= 400 )
   SCIP_CALL( SCIPrandomCreate(&randnumgen, SCIPblkmem(scip), (unsigned int) randomseed) );
#endif

   /* init subscip if we use sparsification */
   if ( sparsification )
   {
      SCIP_CALL( SCIPcreate(&subscip) );
      SCIP_CALL( SCIPincludeDefaultPlugins(subscip) );
   }

   /* start loop */
   for (b = 0; b < nbags; ++b)
   {
      SCIP_Bool runsparsification;
      SCIP_Real errorrate = 0.0;
      SCIP_Real margin;
      int noldbagweights;
      int npoints;
      int k = 0;

      npoints = maxbagsize;

      SCIPinfoMessage(scip, NULL, "\n-----------------------------------------------------------\n");
      if ( nbags > 1 )
         SCIPinfoMessage(scip, NULL, "Considering bag %d with size %d.\n", b, npoints);

      /* create random permutation to sample - take points with replacement! */
      for (i = 0; i < trainingdata->npoints; ++i)
         randperm[i] = i;

#if ( SCIP_VERSION < 400 )
      SCIPpermuteIntArray(randperm, 0, trainingdata->npoints, (unsigned int*) &randomseed);
#else
      SCIPrandomPermuteIntArray(randnumgen, randperm, 0, trainingdata->npoints);
#endif

      /* copy data */
      for (i = 0; i < npoints; ++i)
      {
         assert( 0 <= randperm[i] && randperm[i] < trainingdata->npoints );
         for (j = 0; j < dim; ++j)
            features[k++] = trainingdata->features[randperm[i] * dim + j];
         assert( k <= maxbagsize * dim );
         labels[i] = trainingdata->labels[randperm[i]];
         assert( SCIPisEQ(scip, labels[i], 0.0) || SCIPisEQ(scip, labels[i], 1.0) );
      }

      margin = marginval;
      if ( optimizemargin )
      {
         /* Run AdaBoost to get good estimation of error rate. This is used to set the objective for the margin optimization model. */
         if ( predtype )
            score = AdaBoost(dim, npoints, features, labels, 20, maxdepth, randomseed, 1, npoints, features, labels, NULL, NULL);
         else
            score = AdaBoost(dim, npoints, features, labels, 20, maxdepth, randomseed, 0, npoints, features, labels, NULL, NULL);
         errorrate = ((SCIP_Real) npoints - (SCIP_Real)npoints * score.trainingscore)/((SCIP_Real) npoints);
         SCIPinfoMessage(scip, NULL, "\nEstimating error rate by AdaBoost to be %f.\n", errorrate);
      }
      else if ( guessmargin )
      {
         /* Run AdaBoost to get good estimation on the margin */
         if ( predtype )
            score = AdaBoost(dim, npoints, features, labels, 100, maxdepth, randomseed, 1, npoints, features, labels, &margin, NULL);
         else
            score = AdaBoost(dim, npoints, features, labels, 100, maxdepth, randomseed, 0, npoints, features, labels, &margin, NULL);
#if 0
         margin *= 1.2;  /* add 20% */
#endif
         SCIPinfoMessage(scip, NULL, "\nGuessing margin to be %f.\n", margin);
      }

      /* possibly init with variables from boosting */
      if ( initwithboost )
      {
#if 0
         boosterresult = SimpleBooster(dim, npoints, features, labels, 10, maxdepth, randomseed, predtype);
         SCIPinfoMessage(scip, NULL, "\nRunning Simple Boost for initialization (margin: %f, training score: %f).\n", boosterresult->margin, boosterresult->trainingscore);
#else
         /* use AdaBoost to compute initial solution, use same parameters as below */
         boosterresult = AdaBooster(dim, npoints, features, labels, 100, maxdepth, randomseed, predtype);
         SCIPinfoMessage(scip, NULL, "\nRunning AdaBoost for initialization (margin: %f, training score: %f).\n", boosterresult->margin, boosterresult->trainingscore);
#endif

         margin = boosterresult->margin;
      }

      /* create problem: use objective to be the estimated error rate + 10% (see Demiriz et al.) */
      SCIP_CALL( SCIPcreateProbIpboost(scip, probname, maxdepth, randomseed, dim, npoints, features, labels, margin, lpboost, robustmodel, robustfraction, optimizemargin,
            optmarginmin, MIN(errorrate * marginfactor * npoints, 1000.0), semilowerbound, predtype, votertype) );

      /* We set the dualbound to 0, since this cannot be automatically derived due to pricing. This speeds up computations
       * if a solution with 0 misclassifications is found. Be careful if the objective is changed somehow. */
      SCIP_CALL( SCIPupdateLocalDualbound(scip, 0.0) );

      /* possibly init with variables from boosting */
      if ( initwithboost )
      {
         SCIP_CALL( initFromBoosting(scip, dim, npoints, margin, predtype, lpboost, boosterresult) );

         free(boosterresult->predict);
         free(boosterresult->learners);

         free(boosterresult);
         boosterresult = NULL;
      }
      assert( boosterresult == NULL );

      /* activate pricer */
      SCIP_CALL( SCIPactivatePricer(scip, SCIPfindPricer(scip, "learning")) );

      /* set individual time limit */
      if ( ! SCIPisInfinity(scip, tl) )
      {
         SCIPinfoMessage(scip, NULL, "\nSetting time limit to %.2f.\n", tl);
         SCIP_CALL( SCIPsetRealParam(scip, "limits/time", tl) );
      }

      /* solve problem */
      SCIPinfoMessage(scip, NULL, "\nsolving problem ...\n\n");

#if 0
      SCIP_CALL( SCIPsetIntParam(scip, "limits/bestsol", 1) );
#endif

      SCIP_CALL( SCIPsolve(scip) );

      /* SCIP_CALL( SCIPwriteTransProblem(scip, "debug.lp", "lp", FALSE) ); */

#if 0
      SCIP_CALL( SCIPsetIntParam(scip, "limits/bestsol", -1) );
#endif

      /* always output statistics for later analysis */
      SCIP_CALL( SCIPprintStatistics(scip, NULL) );

      /* evaluate result */
      probdata = SCIPgetProbData(scip);
      assert( probdata != NULL );

      noldbagweights = nbagweights;

      /* evaluate suboptimal solutions */
      if ( evalsuboptimal )
      {
         SCIP_CALL( IpboostEvalStrongLearnerSuboptimal(scip, probdata, threshold, testingdata) );
      }

      /* check time limt for sparsification */
      runsparsification = sparsification;
      if ( ! SCIPisInfinity(scip, tl) )
      {
         if ( tl <= SCIPgetSolvingTime(scip) )
            runsparsification = FALSE;
      }

      if ( runsparsification )
      {
         assert( ! postprocess );  /* cannot currently post process with sparsification */

         /* initialize sub-SCIP */
         SCIP_CALL( SCIPcreateProb(subscip, "subscip", NULL, NULL, NULL, NULL, NULL, NULL, NULL) );

         if ( ! SCIPisInfinity(subscip, memlimit) )
         {
            SCIP_CALL( SCIPsetRealParam(subscip, "limits/memory", memlimit) );
         }

         if ( ! SCIPisInfinity(scip, tl) )
         {
            assert( tl > SCIPgetSolvingTime(scip) );
            SCIP_CALL( SCIPsetRealParam(subscip, "limits/time", tl - SCIPgetSolvingTime(scip)) );
            SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nUsing timelimt %.2f for sparsification subproblem ...\n", tl - SCIPgetSolvingTime(scip));
         }

         SCIP_CALL( IpboostSparsifyStrongLearner(scip, probdata, subscip) );

         SCIP_CALL( IpboostEvalStrongLearner(subscip, NULL, npoints, threshold,
               &nbagdualvals, &nmaxbagdualvals, &nbagweights, &nmaxbagweights, &bagdualvals, &bagweights) );
      }
      else
      {
         /* if we should postprocess the solution */
         if ( postprocess && ! lpboost )
         {
            if ( optimizemargin )
            {
               SCIPerrorMessage("Cannot postprocess solution when margin is optimized!\n");
               return SCIP_ERROR;
            }
            assert( ! robustmodel );

            SCIPinfoMessage(scip, NULL, "\nRunning postprocessing of solution ...\n");

            SCIP_CALL( runPostprocessing(scip, probdata, probname, dim, npoints, features, labels, threshold, maxdepth, randomseed, margin, predtype, votertype,
                  validationdata, &nbagdualvals, &nmaxbagdualvals, &nbagweights, &nmaxbagweights, &bagdualvals, &bagweights) );
         }
         else
         {
            if ( validationdata != NULL )
            {
               SCIP_CALL( IpboostEvalStrongLearnerValidation(scip, probdata, threshold, validationdata, boundlearners,
                     &nbagdualvals, &nmaxbagdualvals, &nbagweights, &nmaxbagweights, &bagdualvals, &bagweights) );
            }
            else
            {
               SCIP_CALL( IpboostEvalStrongLearner(scip, probdata, npoints, threshold,
                     &nbagdualvals, &nmaxbagdualvals, &nbagweights, &nmaxbagweights, &bagdualvals, &bagweights) );
            }
         }
      }

      /* add features and labels */
      nbaglearners[b] = nbagweights - noldbagweights;
      nbagpoints[b] = npoints;

      /* ensure space for features and labels */
      if ( bagfeatures == NULL )
      {
         assert( baglabels == NULL );
         assert( nbagfeatures == 0 );
         assert( nbaglabels == 0 );

         SCIP_CALL( SCIPallocBlockMemoryArray(scip, &bagfeatures, npoints * dim) );
         SCIP_CALL( SCIPallocBlockMemoryArray(scip, &baglabels, npoints) );
         nmaxbagfeatures = npoints * dim;
         nmaxbaglabels = npoints;
      }
      else
      {
         assert( baglabels != NULL );

         if ( nbagfeatures + npoints * dim > nmaxbagfeatures )
         {
            int nnewfeatures;
            nnewfeatures = nbagfeatures + 2 * npoints * dim;
            SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &bagfeatures, nmaxbagfeatures, nnewfeatures) );
            nmaxbagfeatures = nnewfeatures;
         }

         if ( nbaglabels + npoints > nmaxbaglabels )
         {
            int nnewfeatures;
            nnewfeatures = nbaglabels + 2 * npoints;
            SCIP_CALL( SCIPreallocBlockMemoryArray(scip, &baglabels, nmaxbaglabels, nnewfeatures) );
            nmaxbaglabels = nnewfeatures;
         }
      }
      assert( baglabels != NULL );
      assert( bagfeatures != NULL );

      /* collect data */
      for (i = 0; i < npoints; ++i)
      {
         for (j = 0; j < dim; ++j)
            bagfeatures[nbagfeatures++] = features[i * dim + j];
         baglabels[nbaglabels++] = labels[i];
      }
      assert( nbagfeatures <= nmaxbagfeatures );
      assert( nbaglabels <= nmaxbaglabels );

      /* free problem */
      SCIP_CALL( SCIPfreeProb(scip) );

      if ( runsparsification )
      {
         SCIP_CALL( SCIPfreeProb(subscip) );
      }
   }

   /* reset time limit */
   if ( ! SCIPisInfinity(scip, timelimit) )
   {
      SCIP_CALL( SCIPsetRealParam(scip, "limits/time", timelimit) );
   }

   /* reset stall limit */
   SCIP_CALL( SCIPsetLongintParam(scip, "limits/stallnodes", -1LL) );

   SCIPfreeBlockMemoryArray(scip, &labels, maxbagsize);
   SCIPfreeBlockMemoryArray(scip, &features, maxbagsize * dim);
   SCIPfreeBlockMemoryArray(scip, &randperm, trainingdata->npoints);

   SCIPinfoMessage(scip, NULL, "\n-----------------------------------------------------------\n");

   /* call evaluation function from python */
   if ( lpboost )
      SCIPinfoMessage(scip, NULL, "Evaluating the performance of LP boosting (maxdepth = %d).\n", maxdepth);
   else
      SCIPinfoMessage(scip, NULL, "Evaluating the performance of IP boosting (maxdepth = %d).\n", maxdepth);

   /* if there would be only one bag, we could use the following */
   /*
     score = evalStrongLearner(dim, trainingdata->npoints, bagfeatures, baglabels, nbagweights, bagdualvals, bagweights,
     testingdata->npoints, testingdata->features, testingdata->labels, maxdepth, randomseed, votertype);
   */
   score = evalBagStrongLearners(nbags, dim, nbagpoints, nbaglearners, bagfeatures, baglabels, bagdualvals, bagweights,
        testingdata->npoints, testingdata->features, testingdata->labels, maxdepth, randomseed, votertype);

   if ( lpboost )
   {
      SCIPinfoMessage(scip, NULL, "Score of LP boosting on training set: %f  (#misclassified: %d)\n",
         score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
      SCIPinfoMessage(scip, NULL, "Score of LP boosting on testing set:  %f  (#misclassified: %d)\n",
         score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   }
   else
   {
      SCIPinfoMessage(scip, NULL, "Score of IP boosting on training set: %f  (#misclassified: %d)\n",
         score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
      SCIPinfoMessage(scip, NULL, "Score of IP boosting on testing set: %f  (#misclassified: %d)\n",
         score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   }
   SCIPinfoMessage(scip, NULL, "Number of IP boosting weak learners: %d\n", nbagweights);
   SCIPfreeBlockMemoryArray(scip, &nbagpoints, nbags);
   SCIPfreeBlockMemoryArray(scip, &nbaglearners, nbags);

   /* free dual values and weights */
   if ( sparsification )
   {
      SCIPfreeBlockMemoryArrayNull(subscip, &bagdualvals, nmaxbagdualvals);
      SCIPfreeBlockMemoryArrayNull(subscip, &bagweights, nmaxbagweights);
   }
   else
   {
      SCIPfreeBlockMemoryArray(scip, &bagdualvals, nmaxbagdualvals);
      SCIPfreeBlockMemoryArray(scip, &bagweights, nmaxbagweights);
   }
   SCIPfreeBlockMemoryArray(scip, &bagfeatures, nmaxbagfeatures);
   SCIPfreeBlockMemoryArray(scip, &baglabels, nmaxbaglabels);

#if ( SCIP_VERSION >= 600 )
   SCIPfreeRandom(scip, &randnumgen);
#elif ( SCIP_VERSION >= 400 )
   SCIPrandomFree(&randnumgen);
#endif

   if ( sparsification )
   {
      SCIP_CALL( SCIPfree(&subscip) );
   }

   return SCIP_OKAY;
}


/** run LP boosting */
static
SCIP_RETCODE runLPBoost(
   SCIP*                 scip,               /**< SCIP data structure */
   const char*           probname,           /**< problem name */
   SCIP_Real             timelimit,          /**< total time limit for computations */
   SCIP_Real             marginval,          /**< margin that should be used (unless guessmargin or optimizemargin is true) */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   int                   maxdepth,           /**< maximal depth of decision trees used for column generation */
   int                   randomseed,         /**< random seed */
   ClassificationData*   trainingdata,       /**< training data */
   ClassificationData*   testingdata,        /**< testing data */
   ClassificationData*   validationdata,     /**< validation data */
   int                   boundlearners,      /**< maximal number of learners for a solution to be accepted if validation is true */
   SCIP_Bool             optimizemargin,     /**< wether we want to optimize the margin */
   SCIP_Real             optmarginmin,       /**< minimal margin when optimizemargin is true */
   SCIP_Real             marginfactor,       /**< factor to compute margin from estimate */
   SCIP_Bool             guessmargin,        /**< Guess margin based on other boosting algorithm? */
   int                   predtype,           /**< type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   int                   votertype,          /**< type of prediction used for voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   ScoreResult*          score               /**< pointer to store the resulting score */
   )
{
   SCIP_PROBDATA* probdata;
   SCIP_Real* dualvals = NULL;
   SCIP_Real* weights = NULL;
   SCIP_Real errorrate = 0.0;
   SCIP_Real margin;
   int verblevel;
   int maxdualvals = 0;
   int maxweights = 0;
   int ndualvals = 0;
   int nweights = 0;
   int npoints;
   int dim;

   assert( trainingdata != NULL );
   assert( testingdata != NULL );
   assert( score != NULL );

   score->trainingscore = -1.0;
   score->testingscore = -1.0;

   npoints = trainingdata->npoints;
   dim = trainingdata->dim;

   /* turn off output */
   SCIP_CALL( SCIPgetIntParam(scip, "display/verblevel", &verblevel) );
   SCIP_CALL( SCIPsetIntParam(scip, "display/verblevel", 0) );

   /* Turn off heuristics, but leave simplerounding to produce feasible primal solutions, even if we run into the time limit. */
   SCIP_CALL( SCIPsetHeuristics(scip, SCIP_PARAMSETTING_OFF, TRUE) );
   SCIP_CALL( SCIPsetIntParam(scip, "heuristics/simplerounding/freq", 0) );

   margin = marginval;
   if ( optimizemargin )
   {
      ScoreResult s;

      /* Run AdaBoost to get good estimation of error rate. This is used to set the objective for the margin optimization model. */
      if ( predtype )
         s = AdaBoost(dim, npoints, trainingdata->features, trainingdata->labels, 20, maxdepth, randomseed, 1, npoints, trainingdata->features, trainingdata->labels, NULL, NULL);
      else
         s = AdaBoost(dim, npoints, trainingdata->features, trainingdata->labels, 20, maxdepth, randomseed, 0, npoints, trainingdata->features, trainingdata->labels, NULL, NULL);
      errorrate = ((SCIP_Real) npoints - (SCIP_Real)npoints * s.trainingscore)/((SCIP_Real) npoints);
   }
   else if ( guessmargin )
   {
      /* Run AdaBoost to get good estimation on the margin */
      if ( predtype )
         (void) AdaBoost(dim, npoints, trainingdata->features, trainingdata->labels, 100, maxdepth, randomseed, 1,
            npoints, trainingdata->features, trainingdata->labels, &margin, NULL);
      else
         (void) AdaBoost(dim, npoints, trainingdata->features, trainingdata->labels, 100, maxdepth, randomseed, 0,
            npoints, trainingdata->features, trainingdata->labels, &margin, NULL);
      SCIPinfoMessage(scip, NULL, "\nGuessing margin to be %f.\n", margin);
   }

   /* create problem: use objective to be the estimated error rate + 10% (see Demiriz et al.) */
   SCIP_CALL( SCIPcreateProbIpboost(scip, probname, maxdepth, randomseed, dim, npoints, trainingdata->features, trainingdata->labels, margin, TRUE,
         FALSE, 0.0, optimizemargin, optmarginmin, MAX(errorrate * marginfactor * npoints, 0.0001), 0.0, predtype, votertype) );

   /* activate pricer */
   SCIP_CALL( SCIPactivatePricer(scip, SCIPfindPricer(scip, "learning")) );

   /* We set the dualbound to 0, since this cannot be automatically derived due to pricing. This speeds up computations
    * if a solution with 0 misclassifications is found. Be careful if the objective is changed somehow. */
   SCIP_CALL( SCIPupdateLocalDualbound(scip, 0.0) );

   /* set individual time limit */
   if ( ! SCIPisInfinity(scip, timelimit) )
   {
      SCIP_CALL( SCIPsetRealParam(scip, "limits/time", timelimit) );
   }

   /* solve problem */
   SCIP_CALL( SCIPsolve(scip) );

   /* SCIP_CALL( SCIPprintStatistics(scip, NULL) ); */

   /* evaluate result */
   probdata = SCIPgetProbData(scip);
   assert( probdata != NULL );

   if ( validationdata != NULL )
   {
      SCIP_CALL( IpboostEvalStrongLearnerValidation(scip, probdata, threshold, validationdata, boundlearners,
            &ndualvals, &maxdualvals, &nweights, &maxweights, &dualvals, &weights) );
   }
   else
   {
      SCIP_CALL( IpboostEvalStrongLearner(scip, probdata, npoints, threshold,
            &ndualvals, &maxdualvals, &nweights, &maxweights, &dualvals, &weights) );
   }
   /* SCIP_CALL( IpboostEvalStrongLearner(scip, probdata, threshold, &ndualvals, &maxdualvals, &nweights, &maxweights, &dualvals, &weights) ); */
#if 1
   *score = evalStrongLearner(dim, npoints, trainingdata->features, trainingdata->labels, nweights, dualvals, weights,
      testingdata->npoints, testingdata->features, testingdata->labels, maxdepth, randomseed, votertype);
#else
   *score = evalBagStrongLearners(1, dim, &npoints, &nweights, trainingdata->features, trainingdata->labels, dualvals, weights,
      testingdata->npoints, testingdata->features, testingdata->labels, maxdepth, randomseed, votertype);
#endif

   SCIPinfoMessage(scip, NULL, "Number of LP boosting weak learners: %d\n", nweights);

   /* free dual values and weights */
   SCIPfreeBlockMemoryArray(scip, &dualvals, maxdualvals);
   SCIPfreeBlockMemoryArray(scip, &weights, maxweights);

   /* reset output */
   SCIP_CALL( SCIPsetIntParam(scip, "display/verblevel", verblevel) );

   /* reset heuristics */
   SCIP_CALL( SCIPsetHeuristics(scip, SCIP_PARAMSETTING_DEFAULT, TRUE) );

   return SCIP_OKAY;
}


/** run scip with commandline arguments */
static
SCIP_RETCODE runSCIP(
   int                   argc,
   const char**          argv
   )
{
   ClassificationData* origdata;
   ClassificationData* trainingdata;
   ClassificationData* testingdata;
   ClassificationData* validationdata;
   ScoreResult score;
   SCIP* scip = NULL;
   const char* probname;
   const char* testname;
   const char* settingsname;
   SCIP_Longint stalllimit;
   SCIP_Bool usestalling;
   SCIP_Real testdatashare;
   SCIP_Real validationshare = 0.0;
   SCIP_Bool validation;
   SCIP_Real timelimit;
   SCIP_Real memlimit;
   SCIP_Longint nodelimit;
   SCIP_Real margin;
   SCIP_Real semilowerbound;
   SCIP_Real l0weightfactor;
   SCIP_Bool addsinglecuts;
   SCIP_Bool addpaircuts;
   SCIP_Bool useindicator;
   SCIP_Bool useinferbranch;
   SCIP_Real threshold;
   SCIP_Bool lpboost;
   SCIP_Bool robustmodel;
   SCIP_Real robustfraction;
   SCIP_Bool optimizemargin;
   SCIP_Real optmarginmin;
   SCIP_Real marginfactor;
   SCIP_Real trainingmargin;
   SCIP_Real testingmargin;
   SCIP_Bool guessmargin;
   SCIP_Bool initwithboost;
   SCIP_Bool sparsification;
   SCIP_Bool evalsuboptimal;
   SCIP_Bool postprocess;
   SCIP_Bool bagsample;
   int boundlearners;
   int maxbagsize;
   int maxdepth;
   int dispfreq;
   int randomseed;
   int predtype;
   int votertype;
   int adaboosttype;
   int adaboostiters;

   /* variables for timing */
   struct tms timer_beg;
   struct tms timer_end;

   /* parse command line arguments */
   SCIP_CALL( readArguments(argc, argv, &probname, &testname, &settingsname, &timelimit, &memlimit, &nodelimit, &dispfreq, &randomseed) );
   if ( randomseed < 0 )
      randomseed = 0;
   assert( probname != NULL );

   /* initialize SCIP */
   SCIP_CALL( SCIPcreate(&scip) );
   SCIPprintVersion(scip, NULL);
   SCIPinfoMessage(scip, NULL, "[ipboost GitHash: %s]\n", IPBOOSTGITHASH);

   /* include default SCIP plugins */
   SCIP_CALL( SCIPincludeDefaultPlugins(scip) );
   SCIP_CALL( SCIPincludePricerLearning(scip) );

   /* add display column */
   assert( SCIPfindDisp(scip, DISP_NAME_PRICEDVARS) == NULL );
   SCIP_CALL( SCIPincludeDisp(scip, DISP_NAME_PRICEDVARS, DISP_DESC_PRICEDVARS, DISP_HEAD_PRICEDVARS,
         SCIP_DISPSTATUS_AUTO, NULL, NULL, NULL, NULL, NULL, NULL, SCIPdispOutputNAppliedPricedvars, NULL,
         DISP_WIDT_PRICEDVARS, DISP_PRIO_PRICEDVARS, DISP_POSI_PRICEDVARS, DISP_STRI_PRICEDVARS) );

   /* add parameters */
   SCIP_CALL( SCIPaddRealParam(scip,
         "ipboost/margin",
         "Bound for margin",
         &margin, TRUE, DEFAULT_MARGIN, 0.0, 1.0, NULL, NULL) );

   SCIP_CALL( SCIPaddRealParam(scip,
         "ipboost/semilowerbound",
         "Lower bound for sparsification problem",
         &semilowerbound, TRUE, DEFAULT_SEMILOWERBOUND, 0.0, 1.0, NULL, NULL) );

   SCIP_CALL( SCIPaddRealParam(scip,
         "ipboost/l0weightfactor",
         "Weight factor for l0 part in sparsification problem",
         &l0weightfactor, TRUE, DEFAULT_L0WEIGHTFACTOR, 0.0, 1.0, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/addsinglecuts",
         "Add single row cuts to sparsification problem?",
         &addsinglecuts, TRUE, DEFAULT_ADDSINGLECUTS, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/addpaircuts",
         "Add pairwise cuts to sparsification problem?",
         &addpaircuts, TRUE, DEFAULT_ADDPAIRCUTS, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/useindicator",
         "Use indicator constraints for sparsification problem?",
         &useindicator, TRUE, DEFAULT_USEINDICATOR, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/useinferbranch",
         "Use inference branching for IPboost?",
         &useinferbranch, TRUE, DEFAULT_USEINFERBRANCH, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "ipboost/maxbagsize",
         "Maximal size of one instance in bagging approach",
         &maxbagsize, TRUE, DEFAULT_MAXBAGSIZE, 0, INT_MAX, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/bagsample",
         "whether to use only one bag, i.e., sample maxbagsize points and labels to form the training set",
         &bagsample, TRUE, DEFAULT_BAGSAMPLE, NULL, NULL) );

   SCIP_CALL( SCIPaddRealParam(scip,
         "ipboost/threshold",
         "threshold for the minimum weights of weak learners",
         &threshold, TRUE, DEFAULT_THRESHOLD, 0.0, 1.0, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "ipboost/boundlearners",
         "maximal number of learners for a solution to be accepted if validation is true",
         &boundlearners, TRUE, DEFAULT_BOUNDLEARNERS, -1, INT_MAX, NULL, NULL) );

   SCIP_CALL( SCIPaddRealParam(scip,
         "ipboost/testdatashare",
         "Share of points in test data (split off from original data)",
         &testdatashare, TRUE, DEFAULT_TESTDATASHARE, 0.0, 1.0, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/validation",
         "Use validation to choose solution?",
         &validation, TRUE, DEFAULT_VALIDATION, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/lpboost",
         "Use LP boost?",
         &lpboost, TRUE, DEFAULT_LPBOOST, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/robustmodel",
         "Use robust model?",
         &robustmodel, TRUE, DEFAULT_ROBUSTMODEL, NULL, NULL) );

   SCIP_CALL( SCIPaddRealParam(scip,
         "ipboost/robustfraction",
         "Fraction of points that are allowed to be robustified",
         &robustfraction, TRUE, DEFAULT_ROBUSTFRACTION, 0.0, 1.0, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "ipboost/maxdepth",
         "Maximal depth of decision trees used for boosting",
         &maxdepth, TRUE, DEFAULT_MAXDEPTH, 1, INT_MAX, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/optimizemargin",
         "Optimize the margin in the model?",
         &optimizemargin, TRUE, DEFAULT_OPTIMIZEMARGIN, NULL, NULL) );

   SCIP_CALL( SCIPaddRealParam(scip,
         "ipboost/optmarginmin",
         "Minimal margin when optimizing the margin in the model?",
         &optmarginmin, TRUE, DEFAULT_OPTMARGINMIN, 0.0, 1.0, NULL, NULL) );

   SCIP_CALL( SCIPaddRealParam(scip,
         "ipboost/marginfactor",
         "Factor for computing the margin value from error estimation",
         &marginfactor, TRUE, DEFAULT_MARGINFACTOR, 0.0, SCIP_REAL_MAX, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/usestalling",
         "Use stall limit in column generation; turns approach into a heuristic",
         &usestalling, TRUE, DEFAULT_USESTALLING, NULL, NULL) );

   SCIP_CALL( SCIPaddLongintParam(scip,
         "ipboost/stalllimit",
         "Stall limit for column generation",
         &stalllimit, TRUE, DEFAULT_STALLLIMIT, -1, SCIP_LONGINT_MAX, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/guessmargin",
         "Guess margin based on other boosting algorithm",
         &guessmargin, TRUE, DEFAULT_GUESSMARGIN, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/initwithboost",
         "Run simple boosting and initialize problem with result, adjusting the margin",
         &initwithboost, TRUE, DEFAULT_INITWITHBOOST, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/sparsification",
         "Use sparsification on result of ipboost",
         &sparsification, TRUE, DEFAULT_SPARSIFICATION, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/evalsuboptimal",
         "wether suboptimal solutions should be evaluated after IPBoost",
         &evalsuboptimal, TRUE, DEFAULT_EVALSUBOPTIMAL, NULL, NULL) );

   SCIP_CALL( SCIPaddBoolParam(scip,
         "ipboost/postprocess",
         "perform postprocessing with fixed margin",
         &postprocess, TRUE, DEFAULT_POSTPROCESS, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "ipboost/predtype",
         "type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)",
         &predtype, TRUE, DEFAULT_PREDTYPE, 0, 2, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "ipboost/votertype",
         "type of prediction used for voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities)",
         &votertype, TRUE, DEFAULT_VOTERTYPE, 0, 2, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "ipboost/adaboosttype",
         "type of prediction used for adaboost (0: SAMME, 1: SAMME.R)",
         &adaboosttype, TRUE, DEFAULT_ADABOOSTTYPE, 0, 1, NULL, NULL) );

   SCIP_CALL( SCIPaddIntParam(scip,
         "ipboost/adaboostiters",
         "number of iterations of simple boosting and AdaBoost",
         &adaboostiters, TRUE, DEFAULT_ADABOOSTITERS, 1, 100000, NULL, NULL) );

   /* set limit */
   if ( ! SCIPisInfinity(scip, timelimit) )
      SCIP_CALL( SCIPsetRealParam(scip, "limits/time", timelimit) );
   if ( ! SCIPisInfinity(scip, memlimit) )
      SCIP_CALL( SCIPsetRealParam(scip, "limits/memory", memlimit) );
   if ( nodelimit < SCIP_LONGINT_MAX )
      SCIP_CALL( SCIPsetLongintParam(scip, "limits/nodes", nodelimit) );
   if ( dispfreq >= 0 )
      SCIP_CALL( SCIPsetIntParam(scip, "display/freq", dispfreq) );

#if 0
   /* disable subscips in order to avoid expensive heuristics - this produces mixed results */
   SCIP_CALL( SCIPsetSubscipsOff(scip, TRUE) );
#endif

   /* use CPU time */
   SCIP_CALL( SCIPsetIntParam(scip, "timing/clocktype", SCIP_CLOCKTYPE_CPU) );

   /* make sure that we do not restart */
   SCIP_CALL( SCIPsetCharParam(scip, "estimation/restarts/restartpolicy", 'n') );

   /* Disable the row representation switch, since this currently generates a lot of switches between primal and dual simplex. */
   SCIP_CALL( SCIPsetRealParam(scip, "lp/rowrepswitch", -1.0) );

   /* The following needs to be evaluated - not clear whether it helps ... */
   /* enable inference branching if we are looking for a good solution */
   if ( useinferbranch )
   {
      SCIP_CALL( SCIPsetIntParam(scip, "branching/inference/priority", INT_MAX / 4) );
   }

   /* disable #cuts and activate cols display */
   SCIP_CALL( SCIPsetIntParam(scip, "display/separounds/active", 0) );
   SCIP_CALL( SCIPsetIntParam(scip, "display/curconss/active", 0) );
   SCIP_CALL( SCIPsetIntParam(scip, "display/curcols/active", 2) );

   /* disable diving heuristics - tend to be unsuccessful */
   SCIP_CALL( disableDivingHeuristics(scip) );

   /* check settings */
   if ( settingsname != NULL )
   {
      if ( SCIPfileExists(settingsname) )
      {
	 SCIPinfoMessage(scip, NULL, "reading user parameter file <%s> ...\n", settingsname);
	 SCIP_CALL( SCIPreadParams(scip, settingsname) );
      }
      else
      {
	 SCIPinfoMessage(scip, NULL, "parameter file <%s> not found - using default parameters.\n", settingsname);
      }
   }

   /* output changed parameters */
   SCIP_CALL( SCIPwriteParams(scip, NULL, FALSE, TRUE) );
   SCIPinfoMessage(scip, NULL, "\n");

   SCIPinfoMessage(scip, NULL, "Using random seed:     %d\n", randomseed);
   SCIPinfoMessage(scip, NULL, "Using prediction type: %d\n", predtype);
   SCIPinfoMessage(scip, NULL, "Using voter type:      %d\n\n", votertype);

   /* init python */
   SCIP_CALL( initPythonInterface(scip) );

   /* get data from python */
   SCIP_CALL( SCIPgetDataFromPython(scip, probname, testname, randomseed, &origdata, &testingdata) );

   /* adjust validationshare */
   if ( validation )
   {
      validationshare = 0.16;  /* 20% of the 80% of the training data */
   }

   /* split of training, testing, and validation data */
   SCIP_CALL( splitDataset(scip, randomseed, origdata, &trainingdata, &testingdata, &validationdata, testdatashare, validationshare) );

   /* solve problem */
   (void) times(&timer_beg);
   SCIP_CALL( runBagging(scip, probname, timelimit, memlimit, threshold, maxdepth, randomseed, trainingdata, testingdata, validationdata, maxbagsize, margin, lpboost,
         robustmodel, robustfraction, optimizemargin, optmarginmin, marginfactor, semilowerbound, usestalling, stalllimit, guessmargin, initwithboost, boundlearners,
         sparsification, evalsuboptimal, postprocess, bagsample, useinferbranch, predtype, votertype) );
   (void) times(&timer_end);
   if ( lpboost )
      SCIPinfoMessage(scip, NULL, "Time for LP boosting strong learner: %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));
   else
      SCIPinfoMessage(scip, NULL, "Time for IP boosting strong learner: %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));

   /* run LP boost */
   if ( ! lpboost )
   {
      SCIPinfoMessage(scip, NULL, "\nEvaluating performance of LP boosting on testset (maxdepth = %d).\n", maxdepth);
      (void) times(&timer_beg);
      SCIP_CALL( runLPBoost(scip, probname, timelimit, margin, threshold, maxdepth, randomseed, trainingdata, testingdata, validationdata, boundlearners,
            optimizemargin, optmarginmin, marginfactor, guessmargin, predtype, votertype, &score) );
      (void) times(&timer_end);
      SCIPinfoMessage(scip, NULL, "Score of LP boosting on training set: %f  (#misclassified: %d)\n",
         score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
      SCIPinfoMessage(scip, NULL, "Score of LP boosting on testing set:  %f  (#misclassified: %d)\n",
         score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
      SCIPinfoMessage(scip, NULL, "Time for LP boosting strong learner: %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));
   }

   /* run simple boosting */
   SCIPinfoMessage(scip, NULL, "\nEvaluating performance of simple boosting on testset (maxdepth = %d, niters = %d).\n", maxdepth, adaboostiters);
   (void) times(&timer_beg);
   score = SimpleBoost(trainingdata->dim, trainingdata->npoints, trainingdata->features, trainingdata->labels, adaboostiters, maxdepth, randomseed,
      testingdata->npoints, testingdata->features, testingdata->labels, predtype, votertype);
   (void) times(&timer_end);
   SCIPinfoMessage(scip, NULL, "Score of simple boosting on training set: %f  (#misclassified: %d)\n",
      score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Score of simple boosting on testing set:  %f  (#misclassified: %d)\n",
      score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Number of simple boosting weak learners: %d\n", score.nnonz);
   SCIPinfoMessage(scip, NULL, "Time for simple boosting: %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));

   /* run Ada boosting */
   SCIPinfoMessage(scip, NULL, "\nEvaluating performance of AdaBoost on testset (maxdepth = %d, niters = %d, type = %d).\n", maxdepth, adaboostiters, adaboosttype);
   (void) times(&timer_beg);
   score = AdaBoost(trainingdata->dim, trainingdata->npoints, trainingdata->features, trainingdata->labels, adaboostiters, maxdepth, randomseed, adaboosttype,
      testingdata->npoints, testingdata->features, testingdata->labels, &trainingmargin, &testingmargin);
   (void) times(&timer_end);
   SCIPinfoMessage(scip, NULL, "Score of AdaBoost on training set: %f  (#misclassified: %d)\n",
      score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Score of AdaBoost on testing set: %f  (#misclassified: %d)\n",
      score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Minimal AdaBoost margin for training set: %f\n", trainingmargin);
   SCIPinfoMessage(scip, NULL, "Minimal AdaBoost margin for testing set:  %f\n", testingmargin);
   SCIPinfoMessage(scip, NULL, "Number of AdaBoost weak learners: %d\n", score.nnonz);
   SCIPinfoMessage(scip, NULL, "Time for AdaBoost: %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));

   /* run gradient boosting */
   SCIPinfoMessage(scip, NULL, "\nEvaluating performance of gradient boosting on testset (maxdepth = %d).\n", maxdepth);
   (void) times(&timer_beg);
   score = GradientBoosting(trainingdata->dim, trainingdata->npoints, trainingdata->features, trainingdata->labels, maxdepth, randomseed,
      testingdata->npoints, testingdata->features, testingdata->labels);
   (void) times(&timer_end);
   SCIPinfoMessage(scip, NULL, "Score of gradient boosting on training set: %f  (#misclassified: %d)\n",
      score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Score of gradient boosting on testing set:  %f  (#misclassified: %d)\n",
      score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Number of gradient boosting weak learners: %d\n", score.nnonz);
   SCIPinfoMessage(scip, NULL, "Time for gradient boosting: %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));

   /* run random forest with depth 1 */
   SCIPinfoMessage(scip, NULL, "\nEvaluating performance of random forest (maxdepth = 1).\n");
   (void) times(&timer_beg);
   score = RandomForest(trainingdata->dim, trainingdata->npoints, trainingdata->features, trainingdata->labels, 1, randomseed,
      testingdata->npoints, testingdata->features, testingdata->labels);
   (void) times(&timer_end);
   SCIPinfoMessage(scip, NULL, "Score of random forest (maxdepth 1) on training set: %f  (#misclassified: %d)\n",
      score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Score of random forest (maxdepth 1) on testing set:  %f  (#misclassified: %d)\n",
      score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Total number of random forest (maxdepth 1) nodes in trees: %d\n", score.nnonz);
   SCIPinfoMessage(scip, NULL, "Time for random forest (maxdepth 1): %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));

   /* run random forest */
   SCIPinfoMessage(scip, NULL, "\nEvaluating performance of random forest.\n");
   (void) times(&timer_beg);
   score = RandomForest(trainingdata->dim, trainingdata->npoints, trainingdata->features, trainingdata->labels, -1, randomseed,
      testingdata->npoints, testingdata->features, testingdata->labels);
   (void) times(&timer_end);
   SCIPinfoMessage(scip, NULL, "Score of random forest (maxdepth none) on training set: %f  (#misclassified: %d)\n",
      score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Score of random forest (maxdepth none) on testing set:  %f  (#misclassified: %d)\n",
      score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Total number of random forest (maxdepth none) nodes in trees: %d\n", score.nnonz);
   SCIPinfoMessage(scip, NULL, "Time for random forest (maxdepth none): %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));

   /* run decision tree with depth 1 */
   SCIPinfoMessage(scip, NULL, "\nEvaluating performance of decision tree on testset (maxdepth = 1).\n");
   (void) times(&timer_beg);
   score = DecTree(trainingdata->dim, trainingdata->npoints, trainingdata->features, trainingdata->labels, 1, randomseed,
      testingdata->npoints, testingdata->features, testingdata->labels);
   (void) times(&timer_end);
   SCIPinfoMessage(scip, NULL, "Score of decision tree (maxdepth 1) on training set: %f  (#misclassified: %d)\n",
      score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Score of decision tree (maxdepth 1) on testing set:  %f  (#misclassified: %d)\n",
      score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Number of decision tree (maxdepth 1) nodes in tree: %d\n", score.nnonz);
   SCIPinfoMessage(scip, NULL, "Time for decision tree (maxdepth 1): %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));

   /* run decision tree */
   SCIPinfoMessage(scip, NULL, "\nEvaluating performance of decision tree on testset.\n", maxdepth);
   (void) times(&timer_beg);
   score = DecTree(trainingdata->dim, trainingdata->npoints, trainingdata->features, trainingdata->labels, -1, randomseed,
      testingdata->npoints, testingdata->features, testingdata->labels);
   (void) times(&timer_end);
   SCIPinfoMessage(scip, NULL, "Score of decision tree (maxdepth none) on training set: %f  (#misclassified: %d)\n", score,
      score.trainingscore, trainingdata->npoints - (int) round(score.trainingscore * trainingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Score of decision tree (maxdepth none) on testing set: %f  (#misclassified: %d)\n", score,
      score.testingscore, testingdata->npoints - (int) round(score.testingscore * testingdata->npoints));
   SCIPinfoMessage(scip, NULL, "Number of decision tree (maxdepth none) nodes in tree: %d\n", score.nnonz);
   SCIPinfoMessage(scip, NULL, "Time for decision tree (maxdepth none): %.2f\n", (timer_end.tms_utime - timer_beg.tms_utime) / (double)sysconf(_SC_CLK_TCK));

   /* free python */
   SCIP_CALL( freePythonInterface(scip) );

   /* free memory */
   if ( testname != NULL )
   {
      free(testingdata->labels);
      free(testingdata->features);
      free(testingdata);
   }
   else
   {
      SCIPfreeBlockMemoryArray(scip, &testingdata->labels, testingdata->npoints);
      SCIPfreeBlockMemoryArray(scip, &testingdata->features, testingdata->npoints * testingdata->dim);
      SCIPfreeBlockMemory(scip, &testingdata);
   }

   assert( trainingdata != NULL );
   SCIPfreeBlockMemoryArray(scip, &trainingdata->labels, trainingdata->npoints);
   SCIPfreeBlockMemoryArray(scip, &trainingdata->features, trainingdata->npoints * trainingdata->dim);
   SCIPfreeBlockMemory(scip, &trainingdata);

   if ( validationdata != NULL )
   {
      SCIPfreeBlockMemoryArray(scip, &validationdata->labels, validationdata->npoints);
      SCIPfreeBlockMemoryArray(scip, &validationdata->features, validationdata->npoints * validationdata->dim);
      SCIPfreeBlockMemory(scip, &validationdata);
   }

   /* free */
   SCIP_CALL( SCIPfree(&scip) );

   free(origdata->labels);
   free(origdata->features);
   free(origdata);

   BMScheckEmptyMemory();

   printf("\nFinished optimization run.\n");

   return SCIP_OKAY;
}



/** main function */
int
main(
   int                   argc,
   const char**          argv
   )
{
   SCIP_RETCODE retcode;

   retcode = runSCIP(argc, argv);
   if ( retcode != SCIP_OKAY )
   {
      SCIPprintError(retcode);
      return -1;
   }

   return 0;
}
