/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of the program ipboost                               */
/*                                                                           */
/*    an implementation of a boosting learning algorithm.                    */
/*                                                                           */
/*    Copyright (C) 2016-                                                    */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   pricer_learning.c
 * @brief  variable pricer for generating learning data with boosting
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "Python.h"

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "pricer_learning.h"
#include "vardata.h"
#include <scip/cons_linear.h>

#include "generateColumn.h"


#define PRICER_NAME            "learning"
#define PRICER_DESC            "pricer for learning"
#define PRICER_PRIORITY        5000000
#define PRICER_DELAY           TRUE     /* only call pricer if all problem variables have non-negative reduced costs */


/*
 * Data structures
 */

/** variable pricer data */
struct SCIP_PricerData
{
   SCIP*                 scip;               /**< SCIP data structure */
   int                   ngen;               /**< number of generated variables */
};


/*
 * Local methods
 */

/** delete variables */
static
SCIP_DECL_VARDELTRANS(vardeltrans)
{
   IPBOOST_VARDATA** data;

   assert( scip != NULL );
   assert( vardata != NULL );

   data = (IPBOOST_VARDATA**) vardata;

   SCIPfreeBlockMemoryArray(scip, &(*data)->dualvals, (*data)->npoints);
   SCIPfreeBlockMemory(scip, data);

   return SCIP_OKAY;
}

/** adds variable based on passed information */
SCIP_RETCODE SCIPaddLearnerVar(
   SCIP*                 scip,               /**< SCIP data structure */
   int                   dim,                /**< dimension */
   int                   npoints,            /**< number of points */
   double*               dist,               /**< weights for variable */
   double*               pred,               /**< predictions for variable */
   SCIP_Bool             initial             /**< whether the variable is initial */
   )
{
   char name[SCIP_MAXSTRLEN];
   SCIP_PRICERDATA* pricerdata;
   IPBOOST_VARDATA* vardata;
   SCIP_PROBDATA* probdata;
   SCIP_CONS** linconss;
   SCIP_CONS* simplexcons;
   SCIP_PRICER* pricer;
   SCIP_VAR* var;
   SCIP_Real* dualvals;
#ifndef NDEBUG
   int predtype;
#endif
   int i;

   assert( scip != NULL );
   assert( dist != NULL );
   assert( pred != NULL );

   /* get pricer */
   pricer = SCIPfindPricer(scip, "learning");
   if ( pricer == NULL )
   {
      SCIPerrorMessage("Could not find learning pricer.\n");
      return SCIP_PLUGINNOTFOUND;
   }

   /* get pricerdata */
   pricerdata = SCIPpricerGetData(pricer);
   assert( pricerdata != NULL );

   probdata = SCIPgetProbData(scip);
   assert( probdata != NULL );
   assert( dim == IpboostProbGetDim(probdata) );
   assert( npoints == IpboostProbGetNPoints(probdata) );

#ifndef NDEBUG
   predtype = IpboostProbGetPredType(probdata);
#endif
   linconss = IpboostProbGetLinconss(probdata);
   simplexcons = IpboostProbGetSimplexcons(probdata);
   assert( linconss != NULL );
   assert( simplexcons != NULL );

   SCIP_CALL( SCIPallocBlockMemory(scip, &vardata) );
   SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &dualvals, dist, npoints) );

   vardata->number = pricerdata->ngen++;
   vardata->npoints = npoints;
   vardata->dualvals = dualvals;  /* freed with the variable */

   /* add corresponding column */
   if ( initial )
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "init%d", vardata->number);
   else
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "var%d", vardata->number);
   SCIP_CALL( SCIPcreateVar(scip, &var, name, 0.0, SCIPinfinity(scip), 0.0, SCIP_VARTYPE_CONTINUOUS, FALSE, TRUE,
         NULL, NULL, vardeltrans, NULL, (SCIP_VARDATA*) vardata) );

   /* add variable to constraints */
   for (i = 0; i < npoints; ++i)
   {
      assert( SCIPisFeasGE(scip, pred[i], -1) || SCIPisFeasLT(scip, pred[i], 1.0) );
      assert( predtype != 0 || (SCIPisFeasEQ(scip, pred[i], -1) || SCIPisFeasEQ(scip, pred[i], 1.0)) );
      SCIP_CALL( SCIPaddCoefLinear(scip, linconss[i], var, pred[i]) );
   }
   SCIP_CALL( SCIPaddCoefLinear(scip, simplexcons, var, 1.0) );

   SCIPvarMarkDeletable(var);
   SCIP_CALL( SCIPaddVar(scip, var) );
   SCIP_CALL( SCIPreleaseVar(scip, &var) );

#ifndef NDEBUG
   /* check whether column can be reproduced */
   {
      SCIP_Real* column;
      double* features;
      double* labels;

      features = IpboostProbGetFeatures(probdata);
      labels = IpboostProbGetLabels(probdata);
      column = generateCol(dim, npoints, dist, features, labels, IpboostProbGetMaxDepth(probdata), IpboostProbGetRandomSeed(probdata), IpboostProbGetPredType(probdata));

      for (i = 0; i < npoints; ++i)
         assert(SCIPisEQ(scip, pred[i], column[i]));

      free(column);
   }
#endif

   return SCIP_OKAY;
}



/** perform pricing */
static
SCIP_RETCODE pricing(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PRICERDATA*      pricerdata,         /**< data of pricer */
   SCIP_Bool             farkas              /**< whether we are in Farkas pricing */
   )
{
   char name[SCIP_MAXSTRLEN];
   IPBOOST_VARDATA* vardata;
   SCIP_PROBDATA* probdata;
   SCIP_CONS** linconss;
   SCIP_CONS* simplexcons;
   SCIP_Real* dualvals;
   SCIP_Real* features;
   SCIP_Real* labels;
   SCIP_Real* column;
   SCIP_Real constval;
   SCIP_Real sum = 0.0;
   SCIP_VAR* var;
   int maxdepth;
   int npoints;
   int randomseed;
   int predtype;
   int dim;
   int i;

   assert( scip != NULL );

   probdata = SCIPgetProbData(scip);
   assert( probdata != NULL );

   dim = IpboostProbGetDim(probdata);
   npoints = IpboostProbGetNPoints(probdata);
   randomseed = IpboostProbGetRandomSeed(probdata);
   features = IpboostProbGetFeatures(probdata);
   labels = IpboostProbGetLabels(probdata);
   linconss = IpboostProbGetLinconss(probdata);
   simplexcons = IpboostProbGetSimplexcons(probdata);
   maxdepth = IpboostProbGetMaxDepth(probdata);
   predtype = IpboostProbGetPredType(probdata);
   assert( linconss != NULL );
   assert( simplexcons != NULL );

   /* get dual values */
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &dualvals, npoints) ); /* use block memory, since this is stored in the variable */
   for (i = 0; i < npoints; ++i)
   {
      assert( linconss[i] != NULL );
      if ( farkas )
         dualvals[i] = SCIPgetDualfarkasLinear(scip, linconss[i]);
      else
         dualvals[i] = SCIPgetDualsolLinear(scip, linconss[i]);
   }
   if ( farkas )
      constval = SCIPgetDualfarkasLinear(scip, simplexcons);
   else
      constval = SCIPgetDualsolLinear(scip, simplexcons);

   /* obtain column */
   column = generateCol(dim, npoints, dualvals, features, labels, maxdepth, randomseed, predtype);

   /* test whether column improves ... */
   for (i = 0; i < npoints; ++i)
      sum += dualvals[i] * column[i];

   SCIPdebugMessage("Computed variable with reduced cost %f.\n", sum + constval);
   if ( SCIPisFeasPositive(scip, sum + constval) )
   {
      SCIP_CALL( SCIPallocBlockMemory(scip, &vardata) );

      vardata->number = pricerdata->ngen++;
      vardata->npoints = npoints;
      vardata->dualvals = dualvals;

      /* add corresponding column */
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "var%d", vardata->number);
      SCIP_CALL( SCIPcreateVar(scip, &var, name, 0.0, SCIPinfinity(scip), 0.0, SCIP_VARTYPE_CONTINUOUS, FALSE, TRUE,
            NULL, NULL, vardeltrans, NULL, (SCIP_VARDATA*) vardata) );

      for (i = 0; i < npoints; ++i)
      {
         /* predictions must be between -1 and 1 */
         assert( predtype == 2 || (SCIPisFeasGE(scip, column[i], -1) || SCIPisFeasLT(scip, column[i], 1.0)) );
         assert( predtype != 0 || (SCIPisFeasEQ(scip, column[i], -1) || SCIPisFeasEQ(scip, column[i], 1.0)) );
         SCIP_CALL( SCIPaddCoefLinear(scip, linconss[i], var, column[i]) );
      }
      SCIP_CALL( SCIPaddCoefLinear(scip, simplexcons, var, 1.0) );

      SCIPvarMarkDeletable(var);
      SCIP_CALL( SCIPaddPricedVar(scip, var, 1.0) );
      SCIP_CALL( SCIPreleaseVar(scip, &var) );
   }
   else
   {
      SCIPfreeBlockMemoryArray(scip, &dualvals, npoints);
      SCIPdebugMessage("LP optimal.\n");
   }

   /* free column obtained from python */
   free(column);

   return SCIP_OKAY;
}


/*
 * Callback methods of variable pricer
 */

/** copy method for pricer plugins (called when SCIP copies plugins) */
static
SCIP_DECL_PRICERCOPY(pricerCopyLearning)
{  /*lint --e{715}*/
   assert( scip != NULL );
   assert( pricer != NULL );
   assert( strcmp(SCIPpricerGetName(pricer), PRICER_NAME) == 0 );

   /* Do not copy the pricer. Note that pricers are not copied for large scale neighborhood heuristics.
    * Note that valid is FALSE by default.
    * @todo Check whether we want to copy the pricer, too. */

   return SCIP_OKAY;
}


/** destructor of variable pricer to free user data (called when SCIP is exiting) */
static
SCIP_DECL_PRICERFREE(pricerFreeLearning)
{
   SCIP_PRICERDATA* pricerdata;

   assert( scip != NULL );

   SCIPdebugMessage("freeing learing pricer.\n");

   /* get pricerdata */
   pricerdata = SCIPpricerGetData(pricer);

   /* free memory for pricerdata*/
   SCIPfreeBlockMemoryNull(scip, &pricerdata);
   SCIPpricerSetData(pricer, NULL);

   return SCIP_OKAY;
}


/** solving process initialization method of variable pricer (called when branch and bound process is about to begin) */
static
SCIP_DECL_PRICERINITSOL(pricerInitsolLearning)
{
   assert( scip != NULL );
   assert( pricer != NULL );

   SCIPdebugMessage("Initsol\n");

   /* currently do nothing */

   return SCIP_OKAY;
}


/** solving process deinitialization method of variable pricer (called before branch and bound process data is freed) */
static
SCIP_DECL_PRICEREXITSOL(pricerExitsolLearning)
{
   assert( scip != NULL );
   assert( pricer != NULL );

   /* currently do nothing */

   return SCIP_OKAY;
}


/** reduced cost pricing method of variable pricer for feasible LPs */
static
SCIP_DECL_PRICERREDCOST(pricerRedcostLearning)
{
   SCIP_PRICERDATA* pricerdata;

   assert( scip != NULL );
   assert( pricer != NULL );
   assert( result != NULL );

   SCIPdebugMessage("Reduced cost pricing ...\n");

   /* get pricerdata */
   pricerdata = SCIPpricerGetData(pricer);

   SCIP_CALL( pricing(scip, pricerdata, FALSE) );
   *result = SCIP_SUCCESS;

   return SCIP_OKAY;
}


/** farkas pricing method of variable pricer for infeasible LPs */
static
SCIP_DECL_PRICERFARKAS(pricerFarkasLearning)
{
   SCIP_PRICERDATA* pricerdata;

   assert( scip != NULL );
   assert( pricer != NULL );
   assert( result != NULL );

   SCIPdebugMessage("Farkas pricing ...\n");

   /* get pricerdata */
   pricerdata = SCIPpricerGetData(pricer);

   SCIP_CALL( pricing(scip, pricerdata, TRUE) );
   *result = SCIP_SUCCESS;

   return SCIP_OKAY;
}


/*
 * variable pricer specific interface methods
 */

/** creates the learning variable pricer and includes it in SCIP */
SCIP_RETCODE SCIPincludePricerLearning(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_PRICERDATA* pricerdata;
   SCIP_PRICER* pricer = NULL;

   SCIP_CALL( SCIPallocBlockMemory(scip, &pricerdata) );
   pricerdata->scip = scip;
   pricerdata->ngen = 0;

   /* include variable pricer */
   SCIP_CALL( SCIPincludePricerBasic(scip, &pricer, PRICER_NAME, PRICER_DESC, PRICER_PRIORITY, PRICER_DELAY,
         pricerRedcostLearning, pricerFarkasLearning, pricerdata) );
   assert( pricer != NULL );

   /* include non-fundamental callbacks via setter functions */
   SCIP_CALL( SCIPsetPricerCopy(scip, pricer, pricerCopyLearning) );
   SCIP_CALL( SCIPsetPricerFree(scip, pricer, pricerFreeLearning) );
   SCIP_CALL( SCIPsetPricerInitsol(scip, pricer, pricerInitsolLearning) );
   SCIP_CALL( SCIPsetPricerExitsol(scip, pricer, pricerExitsolLearning) );

   return SCIP_OKAY;
}
