/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of the program ipboost                               */
/*                                                                           */
/*    an implementation of a boosting learning algorithm.                    */
/*                                                                           */
/*    Copyright (C) 2016-                                                    */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   probdata_ipboost.c
 * @brief  problem data for ipboost
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include "Python.h"

#include "probdata_ipboost.h"

#include <scip/cons_linear.h>
#include <scip/scipdefplugins.h>

#include "vardata.h"
#include "pricer_learning.h"


/** probdata */
struct SCIP_ProbData
{
   int                   dim;                /**< dimension of points */
   int                   npoints;            /**< number of points */
   SCIP_Real*            features;           /**< features - vector of size dim x npoints */
   SCIP_Real*            labels;             /**< labels - array of size npoints */
   SCIP_CONS**           linconss;           /**< linear constraints (size: npoints) */
   SCIP_CONS*            simplexcons;        /**< linear constraints for simplex equation */
   SCIP_CONS*            robustcons;         /**< linear constraint for bounding robust variables */
   SCIP_VAR**            binvars;            /**< binary variables if lpboost is false (size: npoints) */
   SCIP_VAR**            robustvars;         /**< binary variables for robust model if lpboost is false and robustmodel is true (size: npoints) */
   SCIP_VAR**            slackvars;          /**< continuous slack variables if lpboost is true (size: npoints) */
   SCIP_VAR*             marginvar;          /**< variable for maximizing the margin */
   int                   randomseed;         /**< random seed */
   int                   maxdepth;           /**< maximal depth of decision trees used for column generation */
   SCIP_Bool             lpboost;            /**< whether we are using LP boost */
   SCIP_Bool             robustmodel;        /**< whether we are using the robust model */
   SCIP_Bool             optimizemargin;     /**< wether we want to optimize the margin */
   SCIP_Real             optmarginmin;       /**< minimal margin when optimizemargin is true */
   SCIP_Real             semilowerbound;     /**< lower bound for semicontinuous problem */
   SCIP_Real             margin;             /**< margin bound */
   int                   predtype;           /**< type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   int                   votertype;          /**< type of prediction used for voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
};


/*
 * Local methods
 */


/*
 * Callback methods of probdata
 */

/** transforms the problem */
static
SCIP_DECL_PROBTRANS(probtransIpboost)
{
   assert( scip != NULL );
   assert( sourcedata != NULL );
   assert( targetdata != NULL );

   /* allocate memory */
   SCIP_CALL( SCIPallocBlockMemory(scip, targetdata) );

   (*targetdata)->dim = sourcedata->dim;
   (*targetdata)->npoints = sourcedata->npoints;
   (*targetdata)->features = sourcedata->features;
   (*targetdata)->labels = sourcedata->labels;
   (*targetdata)->lpboost = sourcedata->lpboost;
   (*targetdata)->maxdepth = sourcedata->maxdepth;
   (*targetdata)->robustmodel = sourcedata->robustmodel;
   (*targetdata)->optimizemargin = sourcedata->optimizemargin;
   (*targetdata)->optmarginmin = sourcedata->optmarginmin;
   (*targetdata)->randomseed = sourcedata->randomseed;
   (*targetdata)->margin = sourcedata->margin;
   (*targetdata)->predtype = sourcedata->predtype;
   (*targetdata)->votertype = sourcedata->votertype;

   if ( sourcedata->lpboost )
   {
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*targetdata)->slackvars, sourcedata->npoints) );
      SCIP_CALL( SCIPtransformVars(scip, sourcedata->npoints, sourcedata->slackvars, (*targetdata)->slackvars) );
      (*targetdata)->binvars = NULL;
      (*targetdata)->robustvars = NULL;
   }
   else
   {
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*targetdata)->binvars, sourcedata->npoints) );
      SCIP_CALL( SCIPtransformVars(scip, sourcedata->npoints, sourcedata->binvars, (*targetdata)->binvars) );
      (*targetdata)->slackvars = NULL;

      if ( sourcedata->robustmodel )
      {
         assert( sourcedata->robustvars != NULL );
         SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*targetdata)->robustvars, sourcedata->npoints) );
         SCIP_CALL( SCIPtransformVars(scip, sourcedata->npoints, sourcedata->robustvars, (*targetdata)->robustvars) );
      }
      else
         (*targetdata)->robustvars = NULL;
   }

   if ( sourcedata->optimizemargin )
   {
      SCIP_CALL( SCIPtransformVar(scip, sourcedata->marginvar, &(*targetdata)->marginvar) );
   }

   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*targetdata)->linconss, sourcedata->npoints) );
   SCIP_CALL( SCIPtransformConss(scip, sourcedata->npoints, sourcedata->linconss, (*targetdata)->linconss) );

   SCIP_CALL( SCIPtransformCons(scip, sourcedata->simplexcons, &(*targetdata)->simplexcons) );
   if ( sourcedata->robustmodel )
   {
      SCIP_CALL( SCIPtransformCons(scip, sourcedata->robustcons, &(*targetdata)->robustcons) );
      assert( (*targetdata)->robustcons != NULL );
   }
   else
      (*targetdata)->robustcons = NULL;

   return SCIP_OKAY;
}

/** deletes the transformed problem */
static
SCIP_DECL_PROBDELTRANS(probdeltransIpboost)
{
   int i;

   assert( scip != NULL );
   assert( probdata != NULL );

   /* release simplex condition */
   SCIP_CALL( SCIPreleaseCons(scip, &(*probdata)->simplexcons) );
   if ( (*probdata)->robustmodel )
   {
      SCIP_CALL( SCIPreleaseCons(scip, &(*probdata)->robustcons) );
   }

   /* release binary variables and linear inequalities */
   for (i = (*probdata)->npoints-1; i >= 0; --i)
   {
      if ( (*probdata)->lpboost )
      {
         SCIP_CALL( SCIPreleaseVar(scip, &(*probdata)->slackvars[i]) );
      }
      else
      {
         SCIP_CALL( SCIPreleaseVar(scip, &(*probdata)->binvars[i]) );
         if ( (*probdata)->robustmodel )
         {
            SCIP_CALL( SCIPreleaseVar(scip, &(*probdata)->robustvars[i]) );
         }
      }
      SCIP_CALL( SCIPreleaseCons(scip, &(*probdata)->linconss[i]) );
   }
   SCIPfreeBlockMemoryArray(scip, &(*probdata)->linconss, (*probdata)->npoints);

   if ( (*probdata)->lpboost )
   {
      SCIPfreeBlockMemoryArray(scip, &(*probdata)->slackvars, (*probdata)->npoints);
   }
   else
   {
      SCIPfreeBlockMemoryArray(scip, &(*probdata)->binvars, (*probdata)->npoints);
      if ( (*probdata)->robustmodel )
      {
         SCIPfreeBlockMemoryArray(scip, &(*probdata)->robustvars, (*probdata)->npoints);
      }
   }

   if ( (*probdata)->optimizemargin )
   {
      SCIP_CALL( SCIPreleaseVar(scip, &(*probdata)->marginvar) );
   }

   /* free probdata */
   SCIPfreeBlockMemory(scip, probdata);

   return SCIP_OKAY;
}

/** delete the original problem */
static
SCIP_DECL_PROBDELORIG(probdelorigIpboost)
{
   int i;

   assert( probdata != NULL );
   assert( *probdata != NULL );

   assert( (*probdata)->dim > 0 );

   /* release simplex condition */
   SCIP_CALL( SCIPreleaseCons(scip, &(*probdata)->simplexcons) );
   if ( (*probdata)->robustmodel )
   {
      SCIP_CALL( SCIPreleaseCons(scip, &(*probdata)->robustcons) );
   }

   /* release binary variables and linear inequalities */
   for (i = (*probdata)->npoints-1; i >= 0; --i)
   {
      if ( (*probdata)->lpboost )
      {
         SCIP_CALL( SCIPreleaseVar(scip, &(*probdata)->slackvars[i]) );
      }
      else
      {
         SCIP_CALL( SCIPreleaseVar(scip, &(*probdata)->binvars[i]) );
         if ( (*probdata)->robustmodel )
         {
            SCIP_CALL( SCIPreleaseVar(scip, &(*probdata)->robustvars[i]) );
         }
      }
      SCIP_CALL( SCIPreleaseCons(scip, &(*probdata)->linconss[i]) );
   }
   SCIPfreeBlockMemoryArray(scip, &(*probdata)->linconss, (*probdata)->npoints);

   if ( (*probdata)->lpboost )
   {
      SCIPfreeBlockMemoryArray(scip, &(*probdata)->slackvars, (*probdata)->npoints);
   }
   else
   {
      SCIPfreeBlockMemoryArray(scip, &(*probdata)->binvars, (*probdata)->npoints);
      if ( (*probdata)->robustmodel )
      {
         SCIPfreeBlockMemoryArray(scip, &(*probdata)->robustvars, (*probdata)->npoints);
      }
   }

   if ( (*probdata)->optimizemargin )
   {
      SCIP_CALL( SCIPreleaseVar(scip, &(*probdata)->marginvar) );
   }

   /* free features and labels */
   SCIPfreeBlockMemoryArray(scip, &(*probdata)->features, (*probdata)->dim * (*probdata)->npoints);
   SCIPfreeBlockMemoryArray(scip, &(*probdata)->labels, (*probdata)->npoints);

   /* free probdata */
   SCIPfreeBlockMemory(scip, probdata);

   return SCIP_OKAY;
}


/*
 * probdata specific interface methods
 */

/** sets up the problem data */
SCIP_RETCODE SCIPcreateProbIpboost(
   SCIP*                 scip,               /**< SCIP data structure */
   const char*           probname,           /**< problem name */
   int                   maxdepth,           /**< maximal depth of decision trees used for column generation */
   int                   randomseed,         /**< random seed */
   int                   dim,                /**< dimension of problem (number of features) */
   int                   npoints,            /**< number of points */
   SCIP_Real*            features,           /**< matrix containing features */
   SCIP_Real*            labels,             /**< array of labels */
   SCIP_Real             margin,             /**< margin */
   SCIP_Bool             lpboost,            /**< whether to use LP boosting */
   SCIP_Bool             robustmodel,        /**< whether we should use the robust model */
   SCIP_Real             robustfraction,     /**< fraction of points that are allowed to be ignored in robust model */
   SCIP_Bool             optimizemargin,     /**< wether we want to optimize the margin */
   SCIP_Real             optmarginmin,       /**< minimal margin when optimizemargin is true */
   SCIP_Real             objmargin,          /**< objective coefficient for margin variable */
   SCIP_Real             semilowerbound,     /**< Lower bound for sparsification problem */
   int                   predtype,           /**< type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   int                   votertype           /**< type of prediction used for voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   )
{
   char name[SCIP_MAXSTRLEN];
   SCIP_PROBDATA* probdata = NULL;
   SCIP_Real val;
   SCIP_Real rhs;
   int i;

   assert( scip != NULL );
   assert( randomseed >= 0 );
   assert( features != NULL );
   assert( labels != NULL );
   assert( ! (lpboost && robustmodel) ); /* both is currently no possible */
   assert( optmarginmin >= 0.0 );

   /* allocate memory */
   SCIP_CALL( SCIPallocBlockMemory(scip, &probdata) );

   /* copy data */
   probdata->dim = dim;
   probdata->npoints = npoints;
   probdata->randomseed = randomseed;
   probdata->lpboost = lpboost;
   probdata->robustmodel = robustmodel;
   probdata->maxdepth = maxdepth;
   probdata->optimizemargin = optimizemargin;
   probdata->optmarginmin = optmarginmin;
   probdata->margin = margin;
   probdata->predtype = predtype;
   probdata->votertype = votertype;
   probdata->semilowerbound = semilowerbound;
   SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &probdata->features, features, npoints * dim) );
   SCIP_CALL( SCIPduplicateBlockMemoryArray(scip, &probdata->labels, labels, npoints) );

   /* prepare to set up problem */

   /* create problem in SCIP */
   SCIP_CALL( SCIPcreateProb(scip, probname, probdelorigIpboost, probtransIpboost, probdeltransIpboost, NULL, NULL, NULL, probdata) );

   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nInitializing model ...\n");

   /* create variables */
   if ( lpboost )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nUsing LP boosting ...\n\n");
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &probdata->slackvars, probdata->npoints) );
      for (i = 0; i < npoints; ++i)
      {
         (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "s%d", i);
         SCIP_CALL( SCIPcreateVar(scip, &probdata->slackvars[i], name, 0.0, SCIPinfinity(scip), 1.0, SCIP_VARTYPE_CONTINUOUS,
               TRUE, FALSE, NULL, NULL, NULL, NULL, NULL) );
         SCIP_CALL( SCIPaddVar(scip, probdata->slackvars[i]) );
      }
      probdata->binvars = NULL;
      probdata->robustvars = NULL;
   }
   else
   {
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &probdata->binvars, probdata->npoints) );
      if ( robustmodel )
      {
         SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Using robust model (allow to relax up to %d constraints) ...\n", (int) SCIPfloor(scip, robustfraction * npoints));
         SCIP_CALL( SCIPallocBlockMemoryArray(scip, &probdata->robustvars, probdata->npoints) );
      }

      for (i = 0; i < npoints; ++i)
      {
         (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "z%d", i);
         SCIP_CALL( SCIPcreateVar(scip, &probdata->binvars[i], name, 0.0, 1.0, 1.0, SCIP_VARTYPE_BINARY, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL) );
         SCIP_CALL( SCIPaddVar(scip, probdata->binvars[i]) );

         if ( robustmodel )
         {
            (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "a%d", i);
            SCIP_CALL( SCIPcreateVar(scip, &probdata->robustvars[i], name, 0.0, 1.0, 0.0, SCIP_VARTYPE_BINARY, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL) );
            SCIP_CALL( SCIPaddVar(scip, probdata->robustvars[i]) );
         }
      }
      probdata->slackvars = NULL;
   }

   /* if we want to optimize the margin */
   if ( optimizemargin )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Using margin optimization (objective: %g, min margin: %g) ...\n", fabs(objmargin), optmarginmin);

      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "margin");
      SCIP_CALL( SCIPcreateVar(scip, &probdata->marginvar, name, optmarginmin, 1.0, -fabs(objmargin), SCIP_VARTYPE_CONTINUOUS, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL) );
      SCIP_CALL( SCIPaddVar(scip, probdata->marginvar) );
   }
   else
      probdata->marginvar = NULL;

   /* create linear inequalities */
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, &probdata->linconss, npoints) );

   /* determine coefficient in front of relaxing/slack variable */
   if ( optimizemargin )
   {
      rhs = 0.0;
      if ( lpboost )
         val = 1.0;
      else
         val = 2.0; /* 1.0 to compensate lhs values + 1.0 for maximal margin value */
   }
   else
   {
      rhs = probdata->margin;
      if ( lpboost )
         val = 1.0;
      else
      {
         if ( predtype == 2 )
            val = 0.5 * fabs(log(1000.0)) + rhs;
         else
            val = 1.0 + probdata->margin;
      }
   }

   for (i = 0; i < npoints; ++i)
   {
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "cover%d", i);
      if ( lpboost )
      {
         SCIP_CALL( SCIPcreateConsLinear(scip, &probdata->linconss[i], name, 1, &probdata->slackvars[i], &val, rhs, SCIPinfinity(scip),
               TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, TRUE, FALSE, FALSE, FALSE) );
         SCIP_CALL( SCIPaddCons(scip, probdata->linconss[i]) );
      }
      else
      {
         SCIP_CALL( SCIPcreateConsLinear(scip, &probdata->linconss[i], name, 1, &probdata->binvars[i], &val, rhs, SCIPinfinity(scip),
               TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, TRUE, FALSE, FALSE, FALSE) );
         SCIP_CALL( SCIPaddCons(scip, probdata->linconss[i]) );
         if ( robustmodel )
         {
            SCIP_CALL( SCIPaddCoefLinear(scip, probdata->linconss[i], probdata->robustvars[i], val) );
         }
      }

      if ( optimizemargin )
      {
         SCIP_CALL( SCIPaddCoefLinear(scip, probdata->linconss[i], probdata->marginvar, -1.0) );
      }
   }

   /* create simplex equation */
   (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "simplex");
   SCIP_CALL( SCIPcreateConsLinear(scip, &probdata->simplexcons, name, 0, NULL, NULL, 1.0, 1.0,
         TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, TRUE, FALSE, FALSE, FALSE) );
   SCIP_CALL( SCIPaddCons(scip, probdata->simplexcons) );

   /* possibly add bound for robust variables */
   if ( robustmodel )
   {
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "robust");
      SCIP_CALL( SCIPcreateConsLinear(scip, &probdata->robustcons, name, 0, NULL, NULL, -SCIPinfinity(scip), SCIPfloor(scip, robustfraction * npoints),
            TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, TRUE, FALSE, FALSE, FALSE) );
      SCIP_CALL( SCIPaddCons(scip, probdata->robustcons) );
      for (i = 0; i < npoints; ++i)
      {
         SCIP_CALL( SCIPaddCoefLinear(scip, probdata->robustcons, probdata->robustvars[i], 1.0) );
      }
      assert( probdata->robustcons != NULL );
   }
   else
      probdata->robustcons = NULL;

   if ( robustmodel )
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nInitialized model with %d constraints and %d variables.\n", npoints + 2, 2 * npoints);
   else
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nInitialized model with %d constraints and %d variables.\n", npoints + 1, npoints);

   if ( ! optimizemargin )
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Using margin: %g.\n", probdata->margin);

#if 0
   SCIP_CALL( SCIPprintOrigProblem(scip, NULL, NULL, FALSE) );
#endif

   return SCIP_OKAY;
}


/* ----------------------------------- external methods -------------------------- */

/** returns the dimension */
int IpboostProbGetDim(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->dim;
}

/** returns the number of points */
int IpboostProbGetNPoints(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->npoints;
}

/** returns the random seed */
int IpboostProbGetRandomSeed(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->randomseed;
}

/** returns the feature matrix */
SCIP_Real* IpboostProbGetFeatures(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->features;
}

/** returns the labels */
SCIP_Real* IpboostProbGetLabels(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->labels;
}

/** returns the linear constraints */
SCIP_CONS** IpboostProbGetLinconss(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->linconss;
}

/** returns the simplex constraint */
SCIP_CONS* IpboostProbGetSimplexcons(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->simplexcons;
}

/** returns the binary variables */
SCIP_VAR** IpboostProbGetBinaryvars(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->binvars;
}

/** returns the maximal depth of decision trees used for column generation */
int IpboostProbGetMaxDepth(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->maxdepth;
}

/** returns the prediction type (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
int IpboostProbGetPredType(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->predtype;
}

/** returns the prediction type of the voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
int IpboostProbGetVoterType(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->votertype;
}

/** returns whether lp boost should be performed */
SCIP_Bool IpboostProbGetLpBoost(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   )
{
   assert( probdata != NULL );

   return probdata->lpboost;
}


/** evaluate resulting strong learner
 *
 *  We append the dual values and weights.
 */
SCIP_RETCODE IpboostEvalStrongLearner(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PROBDATA*        probdata,           /**< probdata structure (or NULL) */
   int                   npoints,            /**< number of points */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   int*                  ndualvals,          /**< output: size of dualvalues */
   int*                  nmaxdualvals,       /**< output: maximal number of dualvalues (size of dualvals) */
   int*                  nweights,           /**< output: size of weights */
   int*                  nmaxweights,        /**< output: maximal number of weights (size of weights) */
   SCIP_Real**           dualvals,           /**< output: corresponding dual values for learners */
   SCIP_Real**           weights             /**< output: weights of learners */
   )
{
   SCIP_SOL* sol;
   SCIP_VAR** vars;
   SCIP_Real marginval = -1;
   int ncontvars;
   int nbinvars;
   int nintvars;
   int nimplvars;
   int nmisclassified = 0;
   int nlearner = 0;
   int j;
   int i;

   assert( scip != NULL );
   assert( ndualvals != NULL );
   assert( nmaxdualvals != NULL );
   assert( nweights != NULL );
   assert( nmaxweights != NULL );
   assert( dualvals != NULL );
   assert( weights != NULL );

   sol = SCIPgetBestSol(scip);
   if ( sol == NULL )
      return SCIP_OKAY;

   /* loop through variables */
   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nEvaluating performance of strong learner ...\n");

   /* get variable data */
   SCIP_CALL( SCIPgetVarsData(scip, &vars, NULL, &nbinvars, &nintvars, &nimplvars, &ncontvars) );
   assert( nintvars == 0 );
   assert( nimplvars == 0 );

   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weights of weak learners ...\n");
   for (j = nbinvars; j < nbinvars + ncontvars; ++j)
   {
      SCIP_Real val;
      SCIP_VAR* var;

      var = vars[j];
      assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );

      if ( strcmp(SCIPvarGetName(var), "margin") == 0 || strcmp(SCIPvarGetName(var), "t_margin") == 0 )
         marginval = SCIPgetSolVal(scip, sol, var);

      /* only use conex combination variables: these are the only one using vardata */
      if ( SCIPvarGetData(var) == NULL )
         continue;

      val = SCIPgetSolVal(scip, sol, var);

      /* only treat columns with nonzero value */
      if ( SCIPisFeasGT(scip, val, threshold) )
      {
         ++nlearner;
         SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weight <%s>: %f.\n", SCIPvarGetName(var), val);
      }
      else
      {
         if ( SCIPisFeasPositive(scip, val) )
            SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weight <%s>: %f  (truncated).\n", SCIPvarGetName(var), val);
      }
   }

   /* determine number of miscalssified points in the solution itself */
   if ( probdata != NULL )
   {
      if ( ! probdata->optimizemargin )
         marginval = probdata->margin;

      for (i = 0; i < npoints; ++i)
      {
         SCIP_Real val;

         if ( probdata->lpboost )
         {
            val = SCIPgetSolVal(scip, sol, probdata->slackvars[i]);
            if ( SCIPisFeasGT(scip, val, marginval) )
               ++nmisclassified;
         }
         else
         {
            val = SCIPgetSolVal(scip, sol, probdata->binvars[i]);
            assert( SCIPisFeasIntegral(scip, val) );
            if ( val > 0.5 )
               ++nmisclassified;
         }
      }
   }


   if ( probdata != NULL && probdata->optimizemargin )
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nValue of margin: %g (min. value: %g).\n", marginval, probdata->optmarginmin);

   if ( SCIPisZero(scip, threshold) )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nStrong learner uses %d weak learners.\n", nlearner);
      if ( probdata != NULL )
      {
         if ( probdata->lpboost )
            SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Number of misclassified points with respect to LP: %d (score: %f)\n", nmisclassified,
               1.0 - (SCIP_Real) nmisclassified/((SCIP_Real) npoints));
         else
            SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Number of misclassified points with respect to MIP: %d (score: %f)\n", nmisclassified,
            1.0 - (SCIP_Real) nmisclassified/((SCIP_Real) npoints));
      }
   }
   else
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nStrong learner uses %d truncated weak learners.\n", nlearner);
   }
   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Solution found at time: %f.\n", SCIPsolGetTime(sol));

   /* ensure space for weights and dual values */
   if ( *weights == NULL )
   {
      assert( *dualvals == NULL );
      assert( *ndualvals == 0 );
      assert( *nweights == 0 );
      assert( *nmaxdualvals == 0 );
      assert( *nmaxweights == 0 );

      SCIP_CALL( SCIPallocBlockMemoryArray(scip, dualvals, nlearner * npoints) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, weights, nlearner) );
      *nmaxdualvals = nlearner * npoints;
      *nmaxweights = nlearner;
   }
   else
   {
      if ( *ndualvals + nlearner * npoints > *nmaxdualvals )
      {
         int nnewvals;
         nnewvals = *ndualvals + 2 * nlearner * npoints;
         SCIP_CALL( SCIPreallocBlockMemoryArray(scip, dualvals, *nmaxdualvals, nnewvals) );
         *nmaxdualvals = nnewvals;
      }

      if ( *nweights + nlearner > *nmaxweights )
      {
         int nnewweights;
         nnewweights = *nweights + 2 * nlearner;
         SCIP_CALL( SCIPreallocBlockMemoryArray(scip, weights, *nmaxweights, nnewweights) );
         *nmaxweights = nnewweights;
      }
   }

   /* copy dual values and weights */
   for (j = nbinvars; j < nbinvars + ncontvars; ++j)
   {
      IPBOOST_VARDATA* vardata;
      SCIP_Real val;
      SCIP_VAR* var;

      var = vars[j];
      assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );

      /* only use conex combination variables: these are the only one using vardata */
      if ( SCIPvarGetData(var) == NULL )
         continue;

      val = SCIPgetSolVal(scip, sol, var);

      /* only treat columns with nonzero value */
      if ( SCIPisFeasGT(scip, val, threshold) )
      {
         vardata = (IPBOOST_VARDATA*) SCIPvarGetData(var);
         assert( vardata != NULL );
         assert( vardata->npoints == npoints );

         /* set up matrix of dual variables */
         for (i = 0; i < npoints; ++i)
            (*dualvals)[(*ndualvals)++] = vardata->dualvals[i];
         (*weights)[(*nweights)++] = val;

#ifndef NDEBUG
         /* check whether column can be reproduced */
         if ( probdata != NULL )
         {
            SCIP_Real* column;

            column = generateCol(probdata->dim, npoints, vardata->dualvals, probdata->features, probdata->labels,
               probdata->maxdepth, probdata->randomseed, probdata->predtype);

            for (i = 0; i < npoints; ++i)
            {
               SCIP_VAR** linvars;
               SCIP_Real* linvals;
               int nlinvars;
               int l;

               linvars = SCIPgetVarsLinear(scip, probdata->linconss[i]);
               linvals = SCIPgetValsLinear(scip, probdata->linconss[i]);
               nlinvars = SCIPgetNVarsLinear(scip, probdata->linconss[i]);

               /* find variable */
               for (l = 0; l < nlinvars; ++l)
               {
                  if ( linvars[l] == var )
                  {
                     assert( SCIPisEQ(scip, linvals[l], column[i]) );
                     break;
                  }
               }
               assert( l < nlinvars );
            }
            free(column);
         }
#endif
      }
   }
   assert( *ndualvals <= *nmaxdualvals );
   assert( *nweights <= *nmaxweights );

   return SCIP_OKAY;
}


/** evaluate best strong learner with respect to validation data
 *
 *  We append the dual values and weights.
 */
SCIP_RETCODE IpboostEvalStrongLearnerValidation(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PROBDATA*        probdata,           /**< probdata structure */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   ClassificationData*   validationdata,     /**< validation data (or NULL) */
   int                   boundlearners,      /**< maximal number of learners for a solution to be accepted if validation is true */
   int*                  ndualvals,          /**< output: size of dualvalues */
   int*                  nmaxdualvals,       /**< output: maximal number of dualvalues (size of dualvals) */
   int*                  nweights,           /**< output: size of weights */
   int*                  nmaxweights,        /**< output: maximal number of weights (size of weights) */
   SCIP_Real**           dualvals,           /**< output: corresponding dual values for learners */
   SCIP_Real**           weights             /**< output: weights of learners */
   )
{
   SCIP_SOL** sols;
   SCIP_SOL* sol;
   SCIP_VAR** vars;
   SCIP_Real marginval = -1.0;
   SCIP_Real bestscore = -1.0;
   SCIP_Real val;
   SCIP_VAR* var;
   int nmisclassified = 0;
   int bestsol = -1;
   int bestnlearner = -1;
   int nlearner;
   int ncontvars;
   int nbinvars;
   int nintvars;
   int nimplvars;
   int npoints;
   int nsols;
   int s;
   int j;
   int i;

   assert( scip != NULL );
   assert( probdata != NULL );
   assert( ndualvals != NULL );
   assert( nmaxdualvals != NULL );
   assert( nweights != NULL );
   assert( nmaxweights != NULL );
   assert( dualvals != NULL );
   assert( weights != NULL );
   assert( validationdata != NULL );

   nsols = SCIPgetNSols(scip);
   if ( nsols <= 0 )
      return SCIP_OKAY;
   sols = SCIPgetSols(scip);

   npoints = probdata->npoints;

   /* loop through variables */
   if ( boundlearners >= 0 )
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nEvaluating performance of strong learner (%d solutions, bound on # of learners: %d) ...\n",
         nsols, boundlearners);
   else
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nEvaluating performance of strong learner (%d solutions) ...\n", nsols);

   /* get variable data */
   SCIP_CALL( SCIPgetVarsData(scip, &vars, NULL, &nbinvars, &nintvars, &nimplvars, &ncontvars) );
   assert( nintvars == 0 );
   assert( nimplvars == 0 );

   for (s = 0; s < nsols; ++s)
   {
      ScoreResult score;
      SCIP_Real* localdualvals;
      SCIP_Real* localweights;
      SCIP_Real validationscore;
      int cnt = 0;

      sol = sols[s];
      assert( sol != NULL );

      /* count number of learners */
      nlearner = 0;
      for (j = nbinvars; j < nbinvars + ncontvars; ++j)
      {
         var = vars[j];
         assert( var != NULL );
         assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );

         /* only use conex combination variables: these are the only one using vardata */
         if ( SCIPvarGetData(var) == NULL )
            continue;

         val = SCIPgetSolVal(scip, sol, var);

         /* only treat columns with nonzero value */
         if ( SCIPisFeasGT(scip, val, threshold) )
            ++nlearner;
      }

      /* avoid solution that uses too many learners - but make sure that we keep at least one solution */
      if ( s > 0 && boundlearners != -1 && nlearner > boundlearners )
         continue;

      /* allocate space */
      SCIP_CALL( SCIPallocBufferArray(scip, &localdualvals, nlearner * npoints) );
      SCIP_CALL( SCIPallocBufferArray(scip, &localweights, nlearner) );

      /* copy data */
      nlearner = 0;
      for (j = nbinvars; j < nbinvars + ncontvars; ++j)
      {
         IPBOOST_VARDATA* vardata;

         var = vars[j];
         assert( var != NULL );
         assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );

         /* only use conex combination variables: these are the only one using vardata */
         if ( SCIPvarGetData(var) == NULL )
            continue;

         val = SCIPgetSolVal(scip, sol, var);

         /* only treat columns with nonzero value */
         if ( SCIPisFeasGT(scip, val, threshold) )
         {
            vardata = (IPBOOST_VARDATA*) SCIPvarGetData(var);
            assert( vardata != NULL );
            assert( vardata->npoints == npoints );

            /* set up matrix of dual variables */
            for (i = 0; i < npoints; ++i)
               localdualvals[cnt++] = vardata->dualvals[i];

            localweights[nlearner++] = val;

#ifdef SCIP_CHECKCOLUMN
            /* check whether column can be reproduced */
            {
               SCIP_Real* column;

               column = generateCol(probdata->dim, npoints, vardata->dualvals, probdata->features, probdata->labels,
                  probdata->maxdepth, probdata->randomseed, probdata->predtype);

               for (i = 0; i < npoints; ++i)
               {
                  SCIP_VAR** linvars;
                  SCIP_Real* linvals;
                  int nlinvars;
                  int l;

                  linvars = SCIPgetVarsLinear(scip, probdata->linconss[i]);
                  linvals = SCIPgetValsLinear(scip, probdata->linconss[i]);
                  nlinvars = SCIPgetNVarsLinear(scip, probdata->linconss[i]);

                  /* find variable */
                  for (l = 0; l < nlinvars; ++l)
                  {
                     if ( linvars[l] == var )
                     {
                        if ( ! SCIPisEQ(scip, linvals[l], column[i]) )
                        {
                           SCIPerrorMessage("columns not equal: %f != %f\n", linvals[l], column[i]);
                        }
                        break;
                     }
                  }
                  if ( l >= nlinvars )
                     abort();
               }
               free(column);
            }
#endif
         }
      }

      /* evaluate performance on validation */
      score = evalStrongLearner(probdata->dim, npoints, probdata->features, probdata->labels, nlearner, localdualvals, localweights,
         validationdata->npoints, validationdata->features, validationdata->labels, probdata->maxdepth, probdata->randomseed, probdata->votertype);
      validationscore = score.testingscore;
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Solution with validation score %f (objective value: %f, nlearner: %d).\n",
         validationscore, SCIPgetSolOrigObj(scip, sol), nlearner);

#if 0
      if ( validationscore > bestscore || (fabs(validationscore - bestscore) < 0.01 && nlearner < bestnlearner) )
      {
         bestscore = validationscore;
         bestsol = s;
         bestnlearner = nlearner;
      }
#else
      /* only take solution if validation score is signficantly better than the currently best */
      if ( validationscore > bestscore + 0.01 )
      {
         bestscore = validationscore;
         bestsol = s;
         bestnlearner = nlearner;
      }
#endif

      SCIPfreeBufferArray(scip, &localweights);
      SCIPfreeBufferArray(scip, &localdualvals);
   }

   assert( bestscore >= 0.0 );
   assert( bestsol >= 0 );

   /* return best solution */
   sol = sols[bestsol];
   nlearner = bestnlearner;

   /* determine number of miscalssified points */
   for (i = 0; i < npoints; ++i)
   {
      if ( probdata->lpboost )
      {
         val = SCIPgetSolVal(scip, sol, probdata->slackvars[i]);
         if ( SCIPisFeasPositive(scip, val) )
            ++nmisclassified;
      }
      else
      {
         val = SCIPgetSolVal(scip, sol, probdata->binvars[i]);
         assert( SCIPisFeasIntegral(scip, val) );
         if ( val > 0.5 )
            ++nmisclassified;
      }
   }

   /* ensure space for weights and dual values */
   if ( *weights == NULL )
   {
      assert( *dualvals == NULL );
      assert( *ndualvals == 0 );
      assert( *nweights == 0 );
      assert( *nmaxdualvals == 0 );
      assert( *nmaxweights == 0 );

      SCIP_CALL( SCIPallocBlockMemoryArray(scip, dualvals, nlearner * npoints) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, weights, nlearner) );
      *nmaxdualvals = nlearner * npoints;
      *nmaxweights = nlearner;
   }
   else
   {
      if ( *ndualvals + nlearner * npoints > *nmaxdualvals )
      {
         int nnewvals;
         nnewvals = *ndualvals + 2 * nlearner * npoints;
         SCIP_CALL( SCIPreallocBlockMemoryArray(scip, dualvals, *nmaxdualvals, nnewvals) );
         *nmaxdualvals = nnewvals;
      }

      if ( *nweights + nlearner > *nmaxweights )
      {
         int nnewweights;
         nnewweights = *nweights + 2 * nlearner;
         SCIP_CALL( SCIPreallocBlockMemoryArray(scip, weights, *nmaxweights, nnewweights) );
         *nmaxweights = nnewweights;
      }
   }

   /* copy dual values and weights */
   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nWeights of weak learners of best validation solution (score: %f, objective: %f)...\n", bestscore,
      SCIPgetSolOrigObj(scip, sols[bestsol]));
   for (j = nbinvars; j < nbinvars + ncontvars; ++j)
   {
      IPBOOST_VARDATA* vardata;

      var = vars[j];
      assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );

      if ( strcmp(SCIPvarGetName(var), "margin") == 0 || strcmp(SCIPvarGetName(var), "t_margin") == 0 )
         marginval = SCIPgetSolVal(scip, sol, var);

      /* only use conex combination variables: these are the only one using vardata */
      if ( SCIPvarGetData(var) == NULL )
         continue;

      val = SCIPgetSolVal(scip, sol, var);

      /* only treat columns with nonzero value */
      if ( SCIPisFeasGT(scip, val, threshold) )
      {
         vardata = (IPBOOST_VARDATA*) SCIPvarGetData(var);
         assert( vardata != NULL );
         assert( vardata->npoints == npoints );

         SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weight <%s>: %f.\n", SCIPvarGetName(var), val);

         /* set up matrix of dual variables */
         for (i = 0; i < npoints; ++i)
            (*dualvals)[(*ndualvals)++] = vardata->dualvals[i];

         (*weights)[(*nweights)++] = val;
      }
      else
      {
         if ( SCIPisFeasPositive(scip, val) )
            SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weight <%s>: %f  (truncated).\n", SCIPvarGetName(var), val);
      }
   }
   assert( *ndualvals <= *nmaxdualvals );
   assert( *nweights <= *nmaxweights );

   if ( SCIPisZero(scip, threshold) )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nStrong learner uses %d weak learners.\n", nlearner);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Number of misclassified points: %d (score: %f)\n", nmisclassified,
         1.0 - (SCIP_Real) nmisclassified/((SCIP_Real) npoints));
   }
   else
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nStrong learner uses %d truncated weak learners.\n", nlearner);
   }

   if ( probdata->optimizemargin )
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nValue of margin: %g (min. margin: %g).\n", marginval, probdata->optmarginmin);
   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Solution found at time: %f.\n", SCIPsolGetTime(sol));

   return SCIP_OKAY;
}


/** run postprocessing
 *
 *  Fix integer variables and maximize margin.
 */
SCIP_RETCODE runPostprocessing(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PROBDATA*        probdata,           /**< problem data */
   const char*           probname,           /**< problem name */
   int                   dim,                /**< dimension of problem (number of features) */
   int                   npoints,            /**< number of points */
   SCIP_Real*            features,           /**< matrix containing features */
   SCIP_Real*            labels,             /**< array of labels */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   int                   maxdepth,           /**< maximal depth of decision trees used for column generation */
   int                   randomseed,         /**< random seed */
   SCIP_Real             margin,             /**< margin value */
   int                   predtype,           /**< type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   int                   votertype,          /**< type of prediction used for voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   ClassificationData*   validationdata,     /**< validation data */
   int*                  ndualvals,          /**< output: size of dualvalues */
   int*                  nmaxdualvals,       /**< output: maximal number of dualvalues (size of dualvals) */
   int*                  nweights,           /**< output: size of weights */
   int*                  nmaxweights,        /**< output: maximal number of weights (size of weights) */
   SCIP_Real**           dualvals,           /**< output: corresponding dual values for learners */
   SCIP_Real**           weights             /**< output: weights of learners */
   )
{
   SCIP_SOL* sol;
   SCIP* postscip;
   SCIP_VAR** postvars;
   SCIP_VAR** vars;
   SCIP_PROBDATA* postprobdata;
   SCIP_Real marginval = -1.0;
   int npostbinvars;
   int npostcontvars;
   int nbinvars;
   int nintvars;
   int nimplvars;
   int ncontvars;
   int nlearner = 0;
   int npostlearner = 0;
   int nmisclassified = 0;
   int i;
   int j;

   /* get best solution */
   sol = SCIPgetBestSol(scip);
   if ( sol == NULL )
   {
      SCIPinfoMessage(scip, NULL, "No solution found.\n");
      return SCIP_OKAY;
   }

   /* create new problem */
   SCIP_CALL( SCIPcreate(&postscip) );
   SCIP_CALL( SCIPincludeDefaultPlugins(postscip) );
   SCIP_CALL( SCIPincludePricerLearning(postscip) );

   /* turn off output */
   SCIP_CALL( SCIPsetIntParam(postscip, "display/verblevel", 0) );

   /* use margin optimization */
   SCIP_CALL( SCIPcreateProbIpboost(postscip, probname, maxdepth, randomseed, dim, npoints, features, labels, margin, FALSE, FALSE, 0.0, TRUE,
         0.0, 1.0, 0.0, predtype, votertype) );

   postprobdata = SCIPgetProbData(postscip);
   assert( postprobdata != NULL );

   /* loop through the binary variables and fix them in postscip */
   SCIP_CALL( SCIPgetOrigVarsData(postscip, &postvars, NULL, &npostbinvars, NULL, NULL, NULL) );
   assert( npostbinvars == npoints );
   for (i = 0; i < npoints; ++i)
   {
      SCIP_Real val;
      SCIP_Bool infeasible;
      SCIP_Bool fixed;

      val = SCIPgetSolVal(scip, sol, probdata->binvars[i]);
      assert( SCIPisFeasIntegral(scip, val) );
      assert( postvars[i] != NULL );
      if ( val > 0.5 )
      {
         SCIP_CALL( SCIPfixVar(postscip, postvars[i], 1.0, &infeasible, &fixed) );
         ++nmisclassified;
      }
      else
      {
         SCIP_CALL( SCIPfixVar(postscip, postvars[i], 0.0, &infeasible, &fixed) );
      }
      assert( ! infeasible );
      assert( fixed );
   }

   /* get variable data of original SCIP */
   SCIP_CALL( SCIPgetVarsData(scip, &vars, NULL, &nbinvars, &nintvars, &nimplvars, &ncontvars) );
   assert( nintvars == 0 );
   assert( nimplvars == 0 );

   /* loop through the continuous variables and create corresponding variable in postscip */
   for (j = nbinvars; j < nbinvars + ncontvars; ++j)
   {
      SCIP_Real val;
      SCIP_VAR* var;

      var = vars[j];
      assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );

      /* only use conex combination variables: these are the only one using vardata */
      if ( SCIPvarGetData(var) == NULL )
         continue;

      val = SCIPgetSolVal(scip, sol, var);

      /* only treat columns with nonzero value - NOTE: we could as well take all generated columns! */
      if ( SCIPisFeasPositive(scip, val) )
      {
         IPBOOST_VARDATA* vardata;
         SCIP_Real* pred;

         vardata = (IPBOOST_VARDATA*) SCIPvarGetData(var);
         assert( vardata != NULL );
         assert( vardata->npoints == npoints );

         pred = generateCol(dim, npoints, vardata->dualvals, features, labels, maxdepth, randomseed, predtype);

         SCIP_CALL( SCIPaddLearnerVar(postscip, dim, npoints, vardata->dualvals, pred, TRUE) );
         ++nlearner;

         free(pred);
      }
   }
   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Computed strong learner uses %d weak learners.\n", nlearner);


   /* make sure that the constraints are not modifiable */
   for (i = 0; i < npoints; ++i)
   {
      SCIP_CALL( SCIPsetConsModifiable(postscip, postprobdata->linconss[i], FALSE) );
   }

#ifdef SCIP_MORE_DEBUG
   SCIP_CALL( SCIPprintOrigProblem(postscip, NULL, NULL, FALSE) );
#endif

   /* solve postprocessing LP */
   SCIP_CALL( SCIPsolve(postscip) );

   /* get best solution of postprocessing problem */
   sol = SCIPgetBestSol(postscip);
   if ( sol == NULL )
   {
      SCIPinfoMessage(scip, NULL, "No solution of postprocessing problem found.\n");
      return SCIP_OKAY;
   }

   SCIP_CALL( SCIPgetOrigVarsData(postscip, &postvars, NULL, &npostbinvars, NULL, NULL, &npostcontvars) );
   assert( npostbinvars == npoints );

   /* count number of learners */
   SCIPinfoMessage(scip, NULL, "Weights of postprocessed solution:\n");
   for (j = npostbinvars; j < npostbinvars + npostcontvars; ++j)
   {
      SCIP_VAR* postvar;
      SCIP_Real postval;

      postvar = postvars[j];
      assert( postvar != NULL );
      assert( SCIPvarGetType(postvar) == SCIP_VARTYPE_CONTINUOUS );

      if ( strcmp(SCIPvarGetName(postvar), "margin") == 0 || strcmp(SCIPvarGetName(postvar), "t_margin") == 0 )
         marginval = SCIPgetSolVal(postscip, sol, postvar);

      /* only use conex combination variables: these are the only one using vardata */
      if ( SCIPvarGetData(postvar) == NULL )
         continue;

      postval = SCIPgetSolVal(postscip, sol, postvar);

      /* only treat columns with nonzero value */
      if ( SCIPisFeasGT(scip, postval, threshold) )
      {
         ++npostlearner;
         SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weight <%s>: %f.\n", SCIPvarGetName(postvar), postval);
      }
      else
      {
         if ( SCIPisFeasPositive(postscip, postval) )
            SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weight <%s>: %f  (truncated).\n", SCIPvarGetName(postvar), postval);
      }
   }

   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nValue of margin: %g.\n", marginval);

   if ( SCIPisZero(postscip, threshold) )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nStrong learner uses %d weak learners.\n", npostlearner);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Number of misclassified points with respect to MIP: %d (score: %f)\n", nmisclassified,
         1.0 - (SCIP_Real) nmisclassified/((SCIP_Real) npoints));
   }
   else
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nStrong learner uses %d truncated weak learners.\n", npostlearner);
   }

   /* ensure space for weights and dual values */
   if ( *weights == NULL )
   {
      assert( *dualvals == NULL );
      assert( *ndualvals == 0 );
      assert( *nweights == 0 );
      assert( *nmaxdualvals == 0 );
      assert( *nmaxweights == 0 );

      SCIP_CALL( SCIPallocBlockMemoryArray(scip, dualvals, nlearner * npoints) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, weights, nlearner) );
      *nmaxdualvals = nlearner * npoints;
      *nmaxweights = nlearner;
   }
   else
   {
      if ( *ndualvals + nlearner * npoints > *nmaxdualvals )
      {
         int nnewvals;
         nnewvals = *ndualvals + 2 * nlearner * npoints;
         SCIP_CALL( SCIPreallocBlockMemoryArray(scip, dualvals, *nmaxdualvals, nnewvals) );
         *nmaxdualvals = nnewvals;
      }

      if ( *nweights + nlearner > *nmaxweights )
      {
         int nnewweights;
         nnewweights = *nweights + 2 * nlearner;
         SCIP_CALL( SCIPreallocBlockMemoryArray(scip, weights, *nmaxweights, nnewweights) );
         *nmaxweights = nnewweights;
      }
   }

   /* copy dual values and weights */
   for (j = npostbinvars; j < npostbinvars + npostcontvars; ++j)
   {
      IPBOOST_VARDATA* postvardata;
      SCIP_Real postval;
      SCIP_VAR* postvar;

      postvar = postvars[j];
      assert( SCIPvarGetType(postvar) == SCIP_VARTYPE_CONTINUOUS );

      /* only use conex combination variables: these are the only one using vardata */
      if ( SCIPvarGetData(postvar) == NULL )
         continue;

      postval = SCIPgetSolVal(postscip, sol, postvar);

      /* only treat columns with nonzero value */
      if ( SCIPisFeasGT(postscip, postval, threshold) )
      {
         postvardata = (IPBOOST_VARDATA*) SCIPvarGetData(postvar);
         assert( postvardata != NULL );
         assert( postvardata->npoints == npoints );

         /* set up matrix of dual variables */
         for (i = 0; i < npoints; ++i)
            (*dualvals)[(*ndualvals)++] = postvardata->dualvals[i];
         (*weights)[(*nweights)++] = postval;

#ifndef NDEBUG
         /* check whether column can be reproduced */
         if ( postprobdata != NULL )
         {
            SCIP_Real* column;

            column = generateCol(postprobdata->dim, npoints, postvardata->dualvals, postprobdata->features, postprobdata->labels,
               postprobdata->maxdepth, postprobdata->randomseed, postprobdata->predtype);

            for (i = 0; i < npoints; ++i)
            {
               SCIP_VAR** linvars;
               SCIP_Real* linvals;
               int nlinvars;
               int l;

               linvars = SCIPgetVarsLinear(postscip, postprobdata->linconss[i]);
               linvals = SCIPgetValsLinear(postscip, postprobdata->linconss[i]);
               nlinvars = SCIPgetNVarsLinear(postscip, postprobdata->linconss[i]);

               /* find variable */
               for (l = 0; l < nlinvars; ++l)
               {
                  if ( linvars[l] == postvar )
                  {
                     assert( SCIPisEQ(postscip, linvals[l], column[i]) );
                     break;
                  }
               }
               assert( l < nlinvars );
            }
            free(column);
         }
#endif
      }
   }
   assert( *ndualvals <= *nmaxdualvals );
   assert( *nweights <= *nmaxweights );

   SCIP_CALL( SCIPfree(&postscip) );

   SCIPinfoMessage(scip, NULL, "Finished postprocessing.\n");

   return SCIP_OKAY;
}


/** evaluate all solutions */
SCIP_RETCODE IpboostEvalStrongLearnerSuboptimal(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PROBDATA*        probdata,           /**< probdata structure */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   ClassificationData*   testingdata         /**< testing data */
   )
{
   SCIP_SOL** sols;
   SCIP_SOL* sol;
   SCIP_VAR** vars;
   SCIP_Real val;
   SCIP_VAR* var;
   int ncontvars;
   int nbinvars;
   int nintvars;
   int nimplvars;
   int npoints;
   int nsols;
   int s;
   int j;
   int i;

   assert( scip != NULL );
   assert( probdata != NULL );
   assert( testingdata != NULL );

   nsols = SCIPgetNSols(scip);
   if ( nsols <= 0 )
      return SCIP_OKAY;
   sols = SCIPgetSols(scip);

   npoints = probdata->npoints;

   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\n\nEvaluating performance of all solutions found (%d solutions) ...\n", nsols);

   /* get variable data */
   SCIP_CALL( SCIPgetVarsData(scip, &vars, NULL, &nbinvars, &nintvars, &nimplvars, &ncontvars) );
   assert( nintvars == 0 );
   assert( nimplvars == 0 );

   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "#   training  testing  objective  nlearner\n");
   for (s = 0; s < nsols; ++s)
   {
      ScoreResult score;
      SCIP_Real* localdualvals;
      SCIP_Real* localweights;
      int nlearner = 0;
      int cnt = 0;

      sol = sols[s];
      assert( sol != NULL );

      /* count number of learners */
      for (j = nbinvars; j < nbinvars + ncontvars; ++j)
      {
         var = vars[j];
         assert( var != NULL );
         assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );

         /* only use conex combination variables: these are the only one using vardata */
         if ( SCIPvarGetData(var) == NULL )
            continue;

         val = SCIPgetSolVal(scip, sol, var);

         /* only treat columns with nonzero value */
         if ( SCIPisFeasGT(scip, val, threshold) )
            ++nlearner;
      }

      /* allocate space */
      SCIP_CALL( SCIPallocBufferArray(scip, &localdualvals, nlearner * npoints) );
      SCIP_CALL( SCIPallocBufferArray(scip, &localweights, nlearner) );

      /* copy data */
      nlearner = 0;
      for (j = nbinvars; j < nbinvars + ncontvars; ++j)
      {
         IPBOOST_VARDATA* vardata;

         var = vars[j];
         assert( var != NULL );
         assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );

         /* only use conex combination variables: these are the only one using vardata */
         if ( SCIPvarGetData(var) == NULL )
            continue;

         val = SCIPgetSolVal(scip, sol, var);

         /* only treat columns with nonzero value */
         if ( SCIPisFeasGT(scip, val, threshold) )
         {
            vardata = (IPBOOST_VARDATA*) SCIPvarGetData(var);
            assert( vardata != NULL );
            assert( vardata->npoints == npoints );

            /* set up matrix of dual variables */
            for (i = 0; i < npoints; ++i)
               localdualvals[cnt++] = vardata->dualvals[i];

            localweights[nlearner++] = val;
         }
      }

      /* evaluate performance on training and testing data */
      score = evalStrongLearner(probdata->dim, npoints, probdata->features, probdata->labels, nlearner, localdualvals, localweights,
         testingdata->npoints, testingdata->features, testingdata->labels, probdata->maxdepth, probdata->randomseed, probdata->votertype);
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "%-2d  %-7.2f   %-7.2f  %-7.1f    %-3d\n",
         s, 100.0 * score.trainingscore, 100.0 * score.testingscore, SCIPgetSolOrigObj(scip, sol), nlearner);

      SCIPfreeBufferArray(scip, &localweights);
      SCIPfreeBufferArray(scip, &localdualvals);
   }

   return SCIP_OKAY;
}


/** add single row cuts to sparsification problem */
static
SCIP_RETCODE IpboostAddSingleCuts(
   SCIP*                 scip,               /**< SCIP */
   int                   npoints,            /**< number of points */
   int                   nlearner,           /**< number of columns */
   int*                  rownnonz,           /**< number of nonzeros in each row */
   int**                 rowidx,             /**< array for the indices of the columns for each row */
   SCIP_Real**           rowval,             /**< array for the values of the columns for each row */
   SCIP_VAR**            classvars,          /**< classification variables */
   SCIP_VAR**            sparsevars          /**< sparsification variables (one for each continuous var.) */
   )
{
   char name[SCIP_MAXSTRLEN];
   SCIP_VAR** consvars;
   SCIP_CONS* cons;
   int i;

   assert( scip != NULL );
   assert( rowidx != NULL );
   assert( rowval != NULL );

   SCIP_CALL( SCIPallocBufferArray(scip, &consvars, nlearner + 1) );

   for (i = 0; i < npoints; ++i)
   {
      int cnt = 0;
      int j;

      assert( 0 < rownnonz[i] && rownnonz[i] <= nlearner );
      for (j = 0; j < rownnonz[i]; ++j)
      {
         if ( SCIPisPositive(scip, rowval[i][j]) )
         {
            assert( 0 <= rowidx[i][j] && rowidx[i][j] < nlearner );
            consvars[cnt++] = sparsevars[rowidx[i][j]];
         }
      }
      consvars[cnt++] = classvars[i];

      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "cut%d", i);
      /* lazy cuts: not inital, separate, non enforcing, no check, propagation */
      SCIP_CALL( SCIPcreateConsSetcover(scip, &cons, name, cnt, consvars, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE, FALSE, TRUE, TRUE, FALSE) );
      SCIP_CALL( SCIPaddCons(scip, cons) );
      SCIP_CALL( SCIPreleaseCons(scip, &cons) );
   }
   SCIPfreeBufferArray(scip, &consvars);

   return SCIP_OKAY;
}


/** add cuts from pairs of points to sparsification problem
 *
 *  If for two points there are two points \f$i\f$ and \f$k\f$ which are correctly classified by many classifiers, i.e.,
 *  their coefficient is 1 in the matrix, it may make sense to add the following constraints.
 *  \f$z_i + z_k + \sum \{ y_j: j \text{ classifies } i \text{ or } k \text{ correctly} \}\f$.
 *
 *  This cut is not dominated by the sum of the two single row cuts if there are points that classify both points
 *  correctly since then the coefficient of the corresponding classifier variables is 1 and not 2.
 */
static
SCIP_RETCODE IpboostAddPairCuts(
   SCIP*                 scip,               /**< SCIP */
   int                   npoints,            /**< number of points */
   int                   nlearner,           /**< number of columns */
   int*                  rownnonz,           /**< number of nonzeros in each row */
   int**                 rowidx,             /**< array for the indices of the columns for each row */
   SCIP_Real**           rowval,             /**< array for the values of the columns for each row */
   SCIP_VAR**            classvars,          /**< classification variables */
   SCIP_VAR**            sparsevars          /**< sparsification variables (one for each continuous var.) */
   )
{
   char name[SCIP_MAXSTRLEN];
   SCIP_VAR** consvars;
   SCIP_CONS* cons;
   int i;
   int k;

   assert( scip != NULL );
   assert( rowidx != NULL );
   assert( rowval != NULL );

   SCIP_CALL( SCIPallocBufferArray(scip, &consvars, nlearner + 2) );

   for (i = 0; i < npoints; ++i)
   {
      for (k = i + 1; k < npoints; ++k)
      {
         SCIP_Real val;
         int cnt = 0;
         int nboth = 0;
         int j = 0;
         int l = 0;

         while ( j < rownnonz[i] && l < rownnonz[k] )
         {
            while ( j < rownnonz[i] && rowidx[i][j] < rowidx[k][l] )
               ++j;

            if ( j >= rownnonz[i] )
               break;

            while ( l < rownnonz[k] && rowidx[k][l] < rowidx[i][j] )
               ++l;

            if ( l >= rownnonz[k] )
               break;

            assert( rowidx[i][j] == rowidx[k][l] );
            assert( rowidx[i][j] < nlearner );
            assert( rowidx[k][l] < nlearner );
            assert( SCIPisEQ(scip, rowval[i][j], 1.0) || SCIPisEQ(scip, rowval[i][j], -1.0) );
            assert( SCIPisEQ(scip, rowval[k][l], 1.0) || SCIPisEQ(scip, rowval[k][l], -1.0) );

            val = rowval[i][j] + rowval[k][l];

            if ( val > 1.0 )
               ++nboth;

            if ( val > -0.5 )
               consvars[cnt++] = sparsevars[rowidx[i][j]];

            ++j;
            ++l;
         }
         assert( cnt <= nlearner );

         /* It would suffice to check that nboth > 0, but we produce stronger cuts if it is larger.
          * Similarly, cnt < nlearner would suffice, but we produce stronger cuts. */
         if ( nboth > 2 && cnt < nlearner - 1 )
         {
            consvars[cnt++] = classvars[i];
            consvars[cnt++] = classvars[k];
            (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "pair%d#%d", i, k);
            /* lazy cuts: not inital, separate, non enforcing, no check, propagation */
            SCIP_CALL( SCIPcreateConsLogicor(scip, &cons, name, cnt, consvars, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE, FALSE, TRUE, TRUE, FALSE) );
            SCIP_CALL( SCIPaddCons(scip, cons) );
            SCIP_CALL( SCIPreleaseCons(scip, &cons) );
         }
      }
   }

   SCIPfreeBufferArray(scip, &consvars);

   return SCIP_OKAY;
}


/** extract matrix from column generation problem
 *
 *  We extract the part of the matrix of the column generation problem that corresponds to the conditions of the points
 *  and the generated (continuous) columns with nonzero value in the given solution.
 *
 *  The arrays are created using SCIPallocBlockMemoryArray().
 */
static
SCIP_RETCODE IpboostExtractContinuousMatrix(
   SCIP*                 scip,               /**< SCIP */
   SCIP_SOL*             sol,                /**< use only nonzero variables w.r.t. this solution */
   SCIP_CONS**           linconss,           /**< linear constraints */
   int                   npoints,            /**< number of points */
   int*                  ncols,              /**< return: number of columns */
   int**                 rownnonz,           /**< return: number of nonzeros in each row */
   int***                rowidx,             /**< return: array for the indices of the columns for each row */
   SCIP_Real***          rowval              /**< return: array for the values of the columns for each row */
   )
{
   SCIP_VAR** origvars;
   int* varidx;
   int norigbinvars;
   int norigintvars;
   int norigimplvars;
   int norigcontvars;
   int k = 0;
   int i;
   int j;

   assert( scip != NULL );
   assert( sol != NULL );
   assert( ncols != NULL );
   assert( rownnonz != NULL );
   assert( rowidx != NULL );
   assert( rowval != NULL );

   /* get data for original problem */
   SCIP_CALL( SCIPgetVarsData(scip, &origvars, NULL, &norigbinvars, &norigintvars, &norigimplvars, &norigcontvars) );
   assert( norigintvars == 0 );
   assert( norigimplvars == 0 );

   /* build data structure for variables */
   SCIP_CALL( SCIPallocBufferArray(scip, &varidx, norigcontvars) );
   for (j = 0; j < norigcontvars; ++j)
   {
      SCIP_VAR* var;
      SCIP_Real val;

      var = origvars[norigbinvars + j];
      assert( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS );
      assert( SCIPvarGetProbindex(var) == norigbinvars + j );

      val = SCIPgetSolVal(scip, sol, var);

      /* only treat columns with nonzero value */
      if ( SCIPisFeasPositive(scip, val) )
         varidx[j] = k++;
      else
         varidx[j] = -1;
   }
   *ncols = k;

   /* allocate memory */
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, rownnonz, npoints) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, rowidx, npoints) );
   SCIP_CALL( SCIPallocBlockMemoryArray(scip, rowval, npoints) );

   /* loop through constraints */
   for (i = 0; i < npoints; ++i)
   {
      SCIP_Real* linvals;
      SCIP_CONS* lincons;
      SCIP_VAR** linvars;
      int nlinvars;
      int cnt = 0;
      int l;

      lincons = linconss[i];
      assert( lincons != NULL );

      nlinvars = SCIPgetNVarsLinear(scip, lincons);
      linvars = SCIPgetVarsLinear(scip, lincons);
      linvals = SCIPgetValsLinear(scip, lincons);
      assert( linvars != 0 );
      assert( linvals != 0 );

      /* pass one: count nonzero variables in constraint */
      for (l = 0; l < nlinvars; ++l)
      {
         SCIP_VAR* var;
         int idx;

         var = linvars[l];
         assert( var != NULL );

         if ( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS )
         {
            idx = SCIPvarGetProbindex(var) - norigbinvars;
            assert( 0 <= idx && idx < norigcontvars );
            if ( varidx[idx] >= 0 )
               ++cnt;
         }
      }
      assert( 0 < cnt && cnt <= nlinvars && cnt <= *ncols );

      /* alloc memory */
      (*rownnonz)[i] = cnt;
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*rowidx)[i], cnt) );
      SCIP_CALL( SCIPallocBlockMemoryArray(scip, &(*rowval)[i], cnt) );

      /* pass two: collect values */
      cnt = 0;
      for (l = 0; l < nlinvars; ++l)
      {
         SCIP_VAR* var;
         int idx;

         var = linvars[l];
         assert( var != NULL );

         if ( SCIPvarGetType(var) == SCIP_VARTYPE_CONTINUOUS )
         {
            idx = SCIPvarGetProbindex(var) - norigbinvars;
            assert( 0 <= idx && idx < norigcontvars );
            if ( varidx[idx] >= 0 )
            {
               (*rowidx)[i][cnt] = varidx[idx];
               (*rowval)[i][cnt] = linvals[l];
               ++cnt;
            }
         }
      }
      assert( cnt == (*rownnonz)[i] );
   }
   SCIPfreeBufferArray(scip, &varidx);

   return SCIP_OKAY;
}


/** sparsify strong learner
 *
 *  Read solution from scip and set up a model in which the variables are semicontinuous, i.e., they have to be 0 or
 *  larger than a given lower bound. In this way, the solution is sparsified.
 */
SCIP_RETCODE IpboostSparsifyStrongLearner(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PROBDATA*        probdata,           /**< probdata structure */
   SCIP*                 subscip             /**< subscip problem */
   )
{
   char name[SCIP_MAXSTRLEN];
   SCIP_SOL* sol;
   SCIP_VAR** subcontvars;
   SCIP_VAR** subclassvars;
   SCIP_VAR** subsparsevars;
   SCIP_VAR** vars;
   SCIP_CONS** subconss;
   SCIP_CONS* subsimplexcons;
   SCIP_Real* vals;
   SCIP_Real semilowerbound;
   SCIP_Real margin;
   SCIP_Real l0weightfactor;
   SCIP_Real l0weight;
   SCIP_Bool addsinglecuts;
   SCIP_Bool addpaircuts;
   SCIP_Bool useindicator;
   SCIP_Real** rowval;
   SCIP_CONS* cons;
   SCIP_VAR** origvars;
   int** rowidx;
   int* rownnonz;
   int nmisclassified = 0;
   int nsparselearner = 0;
   int norigcontvars;
   int norigbinvars;
   int npoints;
   int nlearner;
   int i;
   int j;

   assert( scip != NULL );
   assert( probdata != NULL );

   /* get best (or optimal) solution */
   sol = SCIPgetBestSol(scip);
   if ( sol == NULL )
      return SCIP_OKAY;

   /* extract matrix */
   npoints = probdata->npoints;
   SCIP_CALL( IpboostExtractContinuousMatrix(scip, sol, probdata->linconss, npoints, &nlearner, &rownnonz, &rowidx, &rowval) );

   /* get parameters */
   SCIP_CALL( SCIPgetRealParam(scip, "ipboost/semilowerbound", &semilowerbound) );
   SCIP_CALL( SCIPgetRealParam(scip, "ipboost/margin", &margin) );
   SCIP_CALL( SCIPgetRealParam(scip, "ipboost/l0weightfactor", &l0weightfactor) );
   SCIP_CALL( SCIPgetBoolParam(scip, "ipboost/addsinglecuts", &addsinglecuts) );
   SCIP_CALL( SCIPgetBoolParam(scip, "ipboost/addpaircuts", &addpaircuts) );
   SCIP_CALL( SCIPgetBoolParam(scip, "ipboost/useindicator", &useindicator) );

   /* set objective weight */
   l0weight = probdata->npoints * l0weightfactor;

   /* alloc space for variables */
   SCIP_CALL( SCIPallocBlockMemoryArray(subscip, &subclassvars, npoints) );
   SCIP_CALL( SCIPallocBlockMemoryArray(subscip, &subcontvars, nlearner) );
   SCIP_CALL( SCIPallocBlockMemoryArray(subscip, &subsparsevars, nlearner) );

   /* create binary indicator variables for the rows (points) */
   for (i = 0; i < npoints; ++i)
   {
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "z%d", i);
      SCIP_CALL( SCIPcreateVar(subscip, &subclassvars[i], name, 0.0, 1.0, 1.0, SCIP_VARTYPE_BINARY, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL) );
      SCIP_CALL( SCIPaddVar(subscip, subclassvars[i]) );
   }

   /* get variables */
   SCIP_CALL( SCIPgetVarsData(scip, &origvars, NULL, &norigbinvars, NULL, NULL, &norigcontvars) );

   /* create continuous and binary variables for the columns (learners) */
   for (j = 0; j < nlearner; ++j)
   {
      /* Just copy vardata for original variable, do not specify delition function, so not memory problem should arise. */
      /* create continuous variable */
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "l%d", j);
      SCIP_CALL( SCIPcreateVar(subscip, &subcontvars[j], name, 0.0, 1.0, 0.0, SCIP_VARTYPE_CONTINUOUS, TRUE, FALSE, NULL, NULL, NULL, NULL,
            SCIPvarGetData(origvars[norigbinvars + j])) );
      SCIP_CALL( SCIPaddVar(subscip, subcontvars[j]) );

      /* create binary variable */
      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "s%d", j);
      SCIP_CALL( SCIPcreateVar(subscip, &subsparsevars[j], name, 0.0, 1.0, l0weight, SCIP_VARTYPE_BINARY, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL) );
      SCIP_CALL( SCIPaddVar(subscip, subsparsevars[j]) );
   }

   /* create linear inequalities */
   SCIP_CALL( SCIPallocBlockMemoryArray(subscip, &subconss, npoints) );
   SCIP_CALL( SCIPallocBufferArray(scip, &vars, nlearner + 1) );
   SCIP_CALL( SCIPallocBufferArray(scip, &vals, nlearner + 1) );
   for (i = 0; i < npoints; ++i)
   {
      /* copy variables and values */
      for (j = 0; j < rownnonz[i]; ++j)
      {
         assert( 0 <= rowidx[i][j] && rowidx[i][j] < nlearner );
         vars[j] = subcontvars[rowidx[i][j]];
         if ( useindicator )
            vals[j] = -rowval[i][j];
         else
            vals[j] = rowval[i][j];
      }

      (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "cover%d", i);
      if ( useindicator )
      {
         SCIP_VAR* negbinvar;

         SCIP_CALL( SCIPgetNegatedVar(subscip, subclassvars[i], &negbinvar) );
         SCIP_CALL( SCIPcreateConsIndicator(subscip, &subconss[i], name, negbinvar, rownnonz[i], vars, vals, -margin,
               TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE) );

#if 0 // ( SCIP_VERSION >= 400 && SCIP_SUBVERSION >= 1 )
         SCIP_CALL( SCIPsetSlackVarUb(subscip, subconss[i], 1.0 + margin) );
#else
         SCIP_CALL( SCIPchgVarUb(subscip, SCIPgetSlackVarIndicator(subconss[i]), 1.0 + margin) );
#endif
      }
      else
      {
         /* add linear constraints */
         vars[rownnonz[i]] = subclassvars[i];
         vals[rownnonz[i]] = 1.0 + margin;
         SCIP_CALL( SCIPcreateConsLinear(subscip, &subconss[i], name, rownnonz[i] + 1, vars, vals, margin, SCIPinfinity(scip),
               TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );
      }
      SCIP_CALL( SCIPaddCons(subscip, subconss[i]) );
   }

   /* create simplex equation */
   for (j = 0; j < nlearner; ++j)
   {
      vars[j] = subcontvars[j];
      vals[j] = 1.0;
   }
   (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "simplex");
   SCIP_CALL( SCIPcreateConsLinear(subscip, &subsimplexcons, name, nlearner, vars, vals, 1.0, 1.0,
         TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );
   SCIP_CALL( SCIPaddCons(subscip, subsimplexcons) );

   SCIPfreeBufferArray(scip, &vals);
   SCIPfreeBufferArray(scip, &vars);

   /* add coupling constraints for columns/learners */
   for (j = 0; j < nlearner; ++j)
   {
      SCIP_Real consval = 1.0;

      if ( useindicator )
      {
         SCIP_VAR* negbinvar;

         /* model currently only works for 0 lower bounds */
         assert( SCIPisZero(subscip, semilowerbound) );

         (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "s_ub%d", j);
         SCIP_CALL( SCIPgetNegatedVar(subscip, subsparsevars[j], &negbinvar) );
         assert( negbinvar != NULL );
         SCIP_CALL( SCIPcreateConsBasicIndicator(subscip, &cons, name, negbinvar, 1, &subcontvars[j], &consval, 0.0) );
         SCIP_CALL( SCIPaddCons(subscip, cons) );
         SCIP_CALL( SCIPreleaseCons(subscip, &cons) );
      }
      else
      {
         if ( SCIPisPositive(subscip, semilowerbound) )
         {
            (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "s_lb%d", j);
            SCIP_CALL( SCIPcreateConsVarbound(subscip, &cons, name, subcontvars[j], subsparsevars[j], -semilowerbound, 0.0, SCIPinfinity(subscip),
                  TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );
            SCIP_CALL( SCIPaddCons(subscip, cons) );
            SCIP_CALL( SCIPreleaseCons(subscip, &cons) );
         }

         (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "s_ub%d", j);
         SCIP_CALL( SCIPcreateConsVarbound(subscip, &cons, name, subcontvars[j], subsparsevars[j], -1.0, -SCIPinfinity(subscip), 0.0,
               TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE) );
         SCIP_CALL( SCIPaddCons(subscip, cons) );
         SCIP_CALL( SCIPreleaseCons(subscip, &cons) );
      }
   }

   /* possibly add cuts */
   if (SCIPisPositive(subscip, margin) )
   {
      if ( addsinglecuts )
      {
         SCIP_CALL( IpboostAddSingleCuts(subscip, npoints, nlearner, rownnonz, rowidx, rowval, subclassvars, subsparsevars) );
      }
      if ( addpaircuts )
      {
         SCIP_CALL( IpboostAddPairCuts(subscip, npoints, nlearner, rownnonz, rowidx, rowval, subclassvars, subsparsevars) );
      }
   }

#if 0
   SCIP_CALL( SCIPprintOrigProblem(subscip, NULL, NULL, FALSE) );
#endif
#ifdef SCIP_WRITEFILE
   (void) SCIPsnprintf(name, SCIP_MAXSTRLEN, "%s_sparsify.lp", SCIPgetProbName(scip));
   SCIP_CALL( SCIPwriteOrigProblem(subscip, name, "lp", FALSE) );
   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nWrote sparsification problem to file <%s>.\n", name);
#endif

   if ( useindicator )
   {
      /* enable the separation of IIS cuts: */
      SCIP_CALL( SCIPsetBoolParam(subscip, "constraints/indicator/sepaalternativelp", TRUE) );
      /* SCIP_CALL( SCIPsetBoolParam(subscip, "constraints/indicator/useobjectivecut", TRUE) ); */

      /* turn of usage of other constraints, i.e., the simplex constraint */
      SCIP_CALL( SCIPsetBoolParam(subscip, "constraints/indicator/useotherconss", TRUE) );

      /* separate coupling constraints, do not use variable bounds */
      SCIP_CALL( SCIPsetBoolParam(subscip, "constraints/indicator/addcouplingcons", FALSE) );
      SCIP_CALL( SCIPsetBoolParam(subscip, "constraints/indicator/sepacouplingcuts", TRUE) );

      /* turn on restart */
      /* SCIP_CALL( SCIPsetBoolParam(subscip, "constraints/indicator/forcerestart", TRUE) ); */

      /* turn on heuristic */
      SCIP_CALL( SCIPsetBoolParam(subscip, "constraints/indicator/trysolfromcover", TRUE) );
      SCIP_CALL( SCIPsetBoolParam(subscip, "heuristics/indicator/oneopt", FALSE) );

      /* turn on decomposition */
      /* SCIP_CALL( SCIPsetBoolParam(subscip, "constraints/indicator/nolinconscont", TRUE) ); */
   }

   /* solve problem */
   SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nSolving sparsification problem with %d starting columns ...\n\n", nlearner);
   SCIP_CALL( SCIPsolve(subscip) );
#ifdef SCIP_DEBUG
   SCIP_CALL( SCIPprintStatistics(subscip, NULL) );
   SCIP_CALL( SCIPprintBestSol(subscip, NULL, FALSE) );
#endif

   /* evaluate solution (reuse sol) */
   sol = SCIPgetBestSol(subscip);
   if ( sol == NULL )
   {
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Sparsification problem found no solution.\n");
   }
   else
   {
      SCIP_Real val;

      for (i = 0; i < npoints; ++i)
      {
         val = SCIPgetSolVal(subscip, sol, subclassvars[i]);
         assert( SCIPisFeasIntegral(subscip, val) );
         if ( val > 0.5 )
            ++nmisclassified;
      }
      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "\nFinal number of misclassified points: %d\n", nmisclassified);

      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weights of learners ...\n");
      for (j = 0; j < nlearner; ++j)
      {
         val = SCIPgetSolVal(subscip, sol, subcontvars[j]);
         if ( SCIPisFeasPositive(subscip, val) )
         {
            ++nsparselearner;
            SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Weight <%s>: %f.\n", SCIPvarGetName(subcontvars[j]), val);
         }
      }

      SCIPverbMessage(scip, SCIP_VERBLEVEL_NORMAL, NULL, "Final number of used learners:        %d\n", nsparselearner);
   }

   /* free data */
   for (i = npoints-1; i >= 0; --i)
   {
      SCIPfreeBlockMemoryArray(scip, &rowval[i], rownnonz[i]);
      SCIPfreeBlockMemoryArray(scip, &rowidx[i], rownnonz[i]);
      SCIP_CALL( SCIPreleaseCons(subscip, &subconss[i]) );
      SCIP_CALL( SCIPreleaseVar(subscip, &subclassvars[i]) );
   }
   SCIPfreeBlockMemoryArray(scip, &rowval, npoints);
   SCIPfreeBlockMemoryArray(scip, &rowidx, npoints);
   SCIPfreeBlockMemoryArray(scip, &rownnonz, npoints);
   SCIPfreeBlockMemoryArray(subscip, &subconss, npoints);

   SCIP_CALL( SCIPreleaseCons(subscip, &subsimplexcons) );
   SCIPfreeBlockMemoryArray(subscip, &subclassvars, npoints);

   for (j = 0; j < nlearner; ++j)
   {
      SCIP_CALL( SCIPreleaseVar(subscip, &subcontvars[j]) );
      SCIP_CALL( SCIPreleaseVar(subscip, &subsparsevars[j]) );
   }
   SCIPfreeBlockMemoryArray(subscip, &subsparsevars, nlearner);
   SCIPfreeBlockMemoryArray(subscip, &subcontvars, nlearner);

   return SCIP_OKAY;
}
