/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of the program ipboost                               */
/*                                                                           */
/*    an implementation of a boosting learning algorithm.                    */
/*                                                                           */
/*    Copyright (C) 2016-                                                    */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   probdata_coloring.h
 * @brief  problem data for ipboost
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#ifndef __SCIP_PROBDATA_IPBOOST__
#define __SCIP_PROBDATA_IPBOOST__

#include <assert.h>

#include "scip/scip.h"
#include "Python.h"
#include "generateColumn.h"

#ifdef __cplusplus
extern "C" {
#endif

/** returns the dimension */
extern
int IpboostProbGetDim(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the number of points */
extern
int IpboostProbGetNPoints(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the random seed */
extern
int IpboostProbGetRandomSeed(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the feature matrix */
extern
SCIP_Real* IpboostProbGetFeatures(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the labels */
extern
SCIP_Real* IpboostProbGetLabels(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the linear constraints */
extern
SCIP_CONS** IpboostProbGetLinconss(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the simplex constraint */
extern
SCIP_CONS* IpboostProbGetSimplexcons(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the binary variables */
extern
SCIP_VAR** IpboostProbGetBinaryvars(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the maximal depth of decision trees used for column generation */
extern
int IpboostProbGetMaxDepth(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the prediction type (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
extern
int IpboostProbGetPredType(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns the prediction type of the voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
extern
int IpboostProbGetVoterType(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** returns whether lp boost should be performed */
extern
SCIP_Bool IpboostProbGetLpBoost(
   SCIP_PROBDATA*        probdata            /**< probdata structure */
   );

/** evaluate resulting strong learner
 *
 *  We append the dual values and weights.
 */
SCIP_RETCODE IpboostEvalStrongLearner(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PROBDATA*        probdata,           /**< probdata structure (or NULL) */
   int                   npoints,            /**< number of points */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   int*                  ndualvals,          /**< output: size of dualvalues */
   int*                  nmaxdualvals,       /**< output: maximal number of dualvalues (size of dualvals) */
   int*                  nweights,           /**< output: size of weights */
   int*                  nmaxweights,        /**< output: maximal number of weights (size of weights) */
   SCIP_Real**           dualvals,           /**< output: corresponding dual values for learners */
   SCIP_Real**           weights             /**< output: weights of learners */
   );

/** evaluate best strong learner with respect to validation data
 *
 *  We append the dual values and weights.
 */
extern
SCIP_RETCODE IpboostEvalStrongLearnerValidation(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PROBDATA*        probdata,           /**< probdata structure */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   ClassificationData*   validationdata,     /**< validation data (or NULL) */
   int                   boundlearners,      /**< maximal number of learners for a solution to be accepted if validation is true */
   int*                  ndualvals,          /**< output: size of dualvalues */
   int*                  nmaxdualvals,       /**< output: maximal number of dualvalues (size of dualvals) */
   int*                  nweights,           /**< output: size of weights */
   int*                  nmaxweights,        /**< output: maximal number of weights (size of weights) */
   SCIP_Real**           dualvals,           /**< output: corresponding dual values for learners */
   SCIP_Real**           weights             /**< output: weights of learners */
   );

/** run postprocessing
 *
 *  Fix integer variables and maximize margin.
 */
extern
SCIP_RETCODE runPostprocessing(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_PROBDATA*        probdata,           /**< problem data */
   const char*           probname,           /**< problem name */
   int                   dim,                /**< dimension of problem (number of features) */
   int                   npoints,            /**< number of points */
   SCIP_Real*            features,           /**< matrix containing features */
   SCIP_Real*            labels,             /**< array of labels */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   int                   maxdepth,           /**< maximal depth of decision trees used for column generation */
   int                   randomseed,         /**< random seed */
   SCIP_Real             margin,             /**< margin value */
   int                   predtype,           /**< type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   int                   votertype,          /**< type of prediction used for voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   ClassificationData*   validationdata,     /**< validation data */
   int*                  ndualvals,          /**< output: size of dualvalues */
   int*                  maxdualvals,        /**< output: maximal number of dualvalues (size of dualvals) */
   int*                  nweights,           /**< output: size of weights */
   int*                  maxweights,         /**< output: maximal number of weights (size of weights) */
   SCIP_Real**           dualvals,           /**< output: corresponding dual values for learners */
   SCIP_Real**           weights             /**< output: weights of learners */
   );

/** evaluate all solutions */
extern
SCIP_RETCODE IpboostEvalStrongLearnerSuboptimal(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PROBDATA*        probdata,           /**< probdata structure */
   SCIP_Real             threshold,          /**< threshold for values of weak learners */
   ClassificationData*   testingdata         /**< testing data */
   );

/** sparsify strong learner */
extern
SCIP_RETCODE IpboostSparsifyStrongLearner(
   SCIP*                 scip,               /**< SCIP */
   SCIP_PROBDATA*        probdata,           /**< probdata structure */
   SCIP*                 subscip             /**< subscip problem */
   );

/* create probdata */

/** sets up the problem data */
extern
SCIP_RETCODE SCIPcreateProbIpboost(
   SCIP*                 scip,               /**< SCIP data structure */
   const char*           probname,           /**< problem name */
   int                   maxdepth,           /**< maximal depth of decision trees used for column generation */
   int                   seed,               /**< random seed */
   int                   dim,                /**< dimension of problem (number of features) */
   int                   npoints,            /**< number of points */
   SCIP_Real*            features,           /**< matrix containing features */
   SCIP_Real*            labels,             /**< array of labels */
   SCIP_Real             margin,             /**< margin */
   SCIP_Bool             lpboost,            /**< whether to use LP boosting */
   SCIP_Bool             robustmodel,        /**< whether we should use the robust model */
   SCIP_Real             robustfraction,     /**< fraction of points that are allowed to be ignored in robust model */
   SCIP_Bool             optimizemargin,     /**< wether we want to optimize the margin */
   SCIP_Real             optmarginmin,       /**< minimal margin when optimizemargin is true */
   SCIP_Real             objmargin,          /**< objective coefficient for margin variable */
   SCIP_Real             semilowerbound,     /**< Lower bound for sparsification problem */
   int                   predtype,           /**< type of prediction used (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   int                   votertype           /**< type of prediction used for voter (0: usual 0/1 prediction, 1: probabilities, 2: log probabilities) */
   );

#ifdef __cplusplus
}
#endif

#endif
