/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*    This file is part of the program ipboost                               */
/*                                                                           */
/*    an implementation of a boosting learning algorithm.                    */
/*                                                                           */
/*    Copyright (C) 2016-                                                    */
/*                                                                           */
/*                                                                           */
/*    Based on SCIP  --- Solving Constraint Integer Programs                 */
/*       SCIP is distributed under the terms of the SCIP Academic Licence,   */
/*       see file COPYING in the SCIP distribution.                          */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**@file   readargs.c
 * @brief  read comand line arguments
 */

#include "readargs.h"
#include <string.h>


/** get problem name */
SCIP_RETCODE getProblemName(
   const char*          filename,            /**< filename */
   char*                probname,            /**< returned problem name */
   int                  maxsize              /**< maximal size of probname */
   )
{
   int i = 0;
   int l;
   int j;

   /* first find end of string */
   while ( filename[i] != 0) /* find end of string */
      ++i;
   l = i;                /* end of string */

   /* if we found ".gz" */
   if ( l > 3 && filename[l-3] == '.' && filename[l-2] == 'g' && filename[l-1] == 'z' )
   {
      l = l - 4;
      i = l;
   }

   /* go back until '.' or '/' appears */
   while ((i > 0) && (filename[i] != '.') && (filename[i] != '/'))
      --i;
   assert( i > 0 );

   /* if we found '.', search for '/' */
   if (filename[i] == '.')
   {
      l = i;
      while ((i > 0) && (filename[i] != '/'))
	 --i;
   }

   /* correct counter */
   if (filename[i] == '/')
      ++i;

   /* copy name */
   j = 0;
   while ( (i < l) && (filename[i] != 0) )
   {
      probname[j++] = filename[i++];
      if ( j >= maxsize-1 )
         return SCIP_ERROR;
   }
   probname[j] = 0;

   return SCIP_OKAY;
}


/** read comand line arguments */
SCIP_RETCODE readArguments(
   int                   argc,               /**< number of shell parameters */
   const char**          argv,               /**< array with shell parameters */
   const char**          probname,           /**< problem name from arguments */
   const char**          testname,           /**< testing file name (optional) */
   const char**          settingsname,       /**< name of settings file */
   SCIP_Real*            timelimit,          /**< time limit read from arguments */
   SCIP_Real*            memlimit,           /**< memory limit read from arguments */
   SCIP_Longint*         nodelimit,          /**< node limit read from arguments */
   int*                  dispfreq,           /**< display frequency */
   int*                  seed                /**< random seed */
   )
{  /*lint --e{818}*/
   char usage[256];
   int i;

   assert( argc > 0 );
   assert( argv != NULL );
   assert( probname != NULL );
   assert( testname != NULL );
   assert( settingsname != NULL );
   assert( timelimit != NULL );
   assert( memlimit != NULL );
   assert( nodelimit != NULL );
   assert( dispfreq != NULL );
   assert( seed != NULL );

   /* init usage text */
#ifndef NDEBUG
   {
      int status;
      status = snprintf(usage, 255, "usage: %s <problem name> [<testing file name>] [-s <setting file>] [-t <time limit>] [-m <mem limit>] [-n <node limit>] [-d <display frequency>] [-p <seed>]", argv[0]);
      assert( 0 <= status && status < 256 );
   }
#else
   (void) snprintf(usage, 255, "usage: %s <problem name> [<testing file name>] [-s <setting file>] [-t <time limit>] [-m <mem limit>] [-n <node limit>] [-d <display frequency>] [-p <seed>]", argv[0]);
#endif

   /* init arguments */
   *timelimit = 1e20;
   *memlimit = 1e20;
   *nodelimit = SCIP_LONGINT_MAX;
   *probname = NULL;
   *testname = NULL;
   *settingsname = NULL;
   *dispfreq = -1;
   *seed = 1;

   /* check all arguments */
   for (i = 1; i < argc; ++i) /*lint !e850*/
   {
      if ( ! strcmp(argv[i], "-p") )
      {
         if ( i == argc-1 )
         {
            fprintf(stderr, "No seed supplied.\n");
            fprintf(stderr, "%s\n", usage);
            return SCIP_ERROR;
         }
         ++i;
         *seed = atoi(argv[i]);
         assert( i < argc );
      }
      else if ( ! strcmp(argv[i], "-s") )
      {
         if ( *settingsname != NULL )
         {
            fprintf(stderr, "%s\n", usage);
            return SCIP_ERROR;
         }
         if ( i == argc-1 )
         {
            fprintf(stderr, "No setting file name supplied.\n");
            fprintf(stderr, "%s\n", usage);
            return SCIP_ERROR;
         }
         ++i;
         *settingsname = argv[i];
         assert( i < argc );
      }
      /* check for time limit */
      else if ( ! strcmp(argv[i], "-t") )
      {
         if ( i == argc-1 )
         {
            fprintf(stderr, "No time limit supplied.\n");
            fprintf(stderr, "%s\n", usage);
            return SCIP_ERROR;
         }
         ++i;
         *timelimit = atof(argv[i]);
         assert( i < argc );
      }
      /* check for memory limit */
      else if ( ! strcmp(argv[i], "-m") )
      {
         if ( i == argc-1 )
         {
            fprintf(stderr, "No memory limit supplied.\n");
            fprintf(stderr, "%s\n", usage);
            return SCIP_ERROR;
         }
         ++i;
         *memlimit = atof(argv[i]);
         assert( i < argc );
      }
      /* check for node limit */
      else if ( ! strcmp(argv[i], "-n") )
      {
         if ( i == argc-1 )
         {
            fprintf(stderr, "No node limit supplied.\n");
            fprintf(stderr, "%s\n", usage);
            return SCIP_ERROR;
         }
         ++i;
         *nodelimit = atol(argv[i]);
         assert( i < argc );
      }
      /* check for display frequency */
      else if ( ! strcmp(argv[i], "-d") )
      {
         if ( i == argc-1 )
         {
            fprintf(stderr, "No display frequency supplied.\n");
            fprintf(stderr, "%s\n", usage);
            return SCIP_ERROR;
         }
         ++i;
         *dispfreq = atoi(argv[i]);
         assert( i < argc );
      }
      else
      {
         /* if problname is already specified */
         if ( *probname != NULL )
         {
            if ( *testname != NULL )
            {
               fprintf(stderr, "%s\n", usage);
               return SCIP_ERROR;
            }
            *testname = argv[i];
         }
         else
            *probname = argv[i];
      }
   }

   if ( *probname == NULL )
   {
      fprintf(stderr, "%s\n", usage);
      exit(1);
   }

   return SCIP_OKAY;
}
